/*
 * $Id: vanbrowse.vs,v 1.25 93/11/08 17:11:30 josh Stab $
 *
 * Copyright (c) 1993 by Viewlogic Systems, Inc.
 *
 * Vantage libraries are held in this directory structure:
 *
 * libname.lib
 *      vantage.lts (file contains symbolic name from character 16 on...)
 *      entity1.edr
 *          vantage.vdl (source code for entity)
 *          behv_name.adr
 *              vantage.vdl
 *              config_name.cdr
 *                  vantage.vdl
 *      package1.pdr
 *          vantage.vdl
 *          body1.bdr
 *              vantage.vdl
 *          
 *      
 * From this, we generalize that:
 *      1. An object corresponds to a directory.
 *      2. It can be viewed by looking at the vantage.vdl file inside,
 *         if there is one.
 *      3. It can be expanded by looping through all subdirectories
 *         of the form (*.edr, *.adr, and *.cdr), depending on whether
 *         the object is a (library, entity, or architecture), respectively.
 *         Libraries can also have .pdr subdirectories (packages), which
 *         can also have .bdr subdirectories (bodies).
 *
 * One special case exists: libraries are in a file "foo.lib", but "foo"
 * is not necessarily the name of the library -- that has to be retrieved
 * from the vantage.lts file inside.  This doesn't happen with any of the
 * others, but it needs we need to "virtualize" the function to get the
 * name from an object.
 *
 * So we make an class record for vantage objects, represented by a vector
 * [
 *      type_name,      (String identifying what object this thing is)
 *      name_func,      (Function mapping physical path to symbolic name)
 *      subdir_types,   (vector of types of subobjects under this type)
 *      ext             (file extension for this type (e.g. ".adr"))
 * ];
 *
 * It's easiest to create these class records bottom up because of the
 * top-down type reference.
 */

/*
 * All the good stuff in this file gets put into this package, and
 * all the exportable stuff is explicitly listed, so we know exactly
 * what the interface is to vansetup.vs.
 */
package browse(add_lib, toggle_expand, collapse, obj_filename, is_library,
               is_deletable, obj_pathname, reset_parent, obj_name,
               find_visual, view_obj, has_view, has_children, obj_time,
               select_pathname, set_fields, design_unit_string, obj_lib,
               obj_symname, update, set_working, configurations)
{
    load ("vanlibtms.vs");

    DO_INFO_BUTTON := FALSE;

    /*
     * For the time being, Viewscript vectors are much more efficient
     * than structures, so we're holding all the library data in them.
     * These symbolic names allow us structure-like access to the
     * heterogeneous vectors.
     */
    TYPE_NAME := 0;
    NAME_FUNC := 1;
    SUBDIR_TYPES := 2;
    EXT := 3;

    /*
     * All the data about the current library path is stored in a picklist.
     * The picklist entries may be libraries, entities, architectures, or
     * configurations.  The data for each entry in the picklist is a vector:
     *
     * [type, symbolic_name, pathname, child_pathname_list, is_expanded,
     *  has_view parent_obj nspaces view_button, time];
     */
    TYPE := 0;
    SYMBOLIC_NAME := 1;
    PATHNAME := 2;
    CHILDREN := 3;
    IS_EXPANDED := 4;
    VIEW_BUTTON := 5;
    PARENT_OBJ := 6;
    NSPACES := 7;
    VTEXT := 8;
    TIME := 9;

    NULL_VECTOR := make_vector (0);
    NSPACES_PER_LEVEL := 5;

    FIELDS := 10;           /* static text region for .tms data */
    TMS_DATA := 11;         /* an array of stuff from the .tms file */

    /*
     * These are the fields of the timestamp-data array.  The ones
     * beginning with LBR_FLAG_ are actually defined by Vantage, in
     * the source file include/lbrlibf.h, and appear as characters
     * in the timestamp file.  I find it convenient to use them as
     * indexes into an array holding all this stuff, and I might
     * as well use the 0th element for something too.
     */

    SOURCE_TIME             := 0;

    LBR_FLAG_SOURCE_FILE    := 1;
    LBR_FLAG_RCHK           := 2;
    LBR_FLAG_DBGLVL         := 3;
    LBR_FLAG_C_OPTLVL       := 4;
    LBR_FLAG_VSS_VER_MAJOR  := 5;
    LBR_FLAG_VSS_VER_MINOR  := 6;
    LBR_FLAG_VHDL_OPTLVL    := 7;
    LBR_FLAG_LANG           := 8;

    MAX_LBR_FLAG            := 8;

    VIF_TIME                := 9;
    BIN_TIME                := 10;

    NUM_TMS_FIELDS          := 11;

    /* Convert library pathname to symbolic name via the vantage.lts file */
    function lib2symbolic(pathname) {
        f := fopen (vsMerge ("", pathname, "vantage", "lts"), "r");
        for (i := 0; i < 16; ++i)
            fgetchar (f);
        symbolic_name := fgetline (f);
        fclose (f);
        if (symbolic_name == nil)
            error ("Unable to find symbolic name for library", pathname);
        symbolic_name;
    } /* function lib2symbolic */

    /* Convert entity/architecture/config pathname to symbolic name */
    function eac2symbolic(pathname)
        caddr (vsParse (pathname));

    /* Create class records */
    ConfigT       := ["Config",       eac2symbolic, NULL_VECTOR, "cdr"];
    ArchitectureT := ["Architecture", eac2symbolic, [ConfigT], "adr"];
    EntityT       := ["Entity",  eac2symbolic, [ArchitectureT], "edr"];

    BodyT         := ["Body",    eac2symbolic, NULL_VECTOR, "bdr"];
    PackageT      := ["Package", eac2symbolic, [BodyT], "pdr"];

    LibraryT      := ["Library", lib2symbolic, [EntityT, PackageT], "lib"];

    ViewT         := ["View",    nil, NULL_VECTOR, nil];

    /*
     * Generates a list of all the child elements of an object.
     * Each element of the list is a cons cell with the sub-object's
     * type and pathname.  This list is kept in the data for each
     * displayed object, so we know what to do when the user double-clicks
     * on it.
     */
    function find_children(pathname, type) {
        children := nil;
        foreach (subtype; vscript.vect_iter (type[SUBDIR_TYPES]))
            foreach (entity; TraverseDir (vsMerge ("", pathname, "*",
                                                   subtype[EXT]),
                                          vsAttr.Dir))
                children = cons (cons (subtype, entity), children);

        reverse_destructive (children);
    } /* function find_children */

    function set_working(files, row, is_working) {
        table := car (vcPicklist.Get (files, row, 0));
        name := vcTable.Get (table, 0, 0);
        type := vcTable.Get (table, 0, 1);
        vcRegion.Font (name, is_working ? vgFont.smallb : vgFont.small);
        vcRegion.Font (type, is_working ? vgFont.smallbi : vgFont.smalli);
        vcRegion.RequestSize (name);
        vcRegion.QueueForRefresh (name);
        vcRegion.RequestSize (type);
        vcRegion.QueueForRefresh (type);
    } /* function set_working */

    /*
     * Generate a string describing the object, including indentation
     * designed to give the appearance of a tree.  This could also
     * be done with some icons, using a table, as either a string or
     * a region fill work as the visual for a picklist.  However, 
     * find_visual is exported outside this package, for use in
     * a status window message, so we'd have to keep some form of this
     * function that generates strings.
     *
     * Note: if there is a "view", meaning a vantage.vdl file exists in
     * the subdirectory, then obj[VIEW_BUTTON] is set.
     *
     */
    function find_visual(files, obj) {
        type := obj[TYPE];
        has_view := fexists (obj_filename (obj));

        nchildren := length (obj[CHILDREN]);
        table := vcTable.Create ("", 1,
                                 3 + /*  name, type, fields */
                                 ((DO_INFO_BUTTON && (obj[TMS_DATA] == nil))
                                  ? 0 : 1) +
                                 ((nchildren > 0) ? 1 : 0) +
                                 (has_view ? 1 : 0));

        /* The visual for the object always starts with "ObjName: ObjType" */
        format := sprintf ("%%%ds", obj[NSPACES]) + " %s: ";
        visual := sprintf (format, "", obj[SYMBOLIC_NAME]);

        vcTable.Add (table, vcRegion.Create ("", vcLabel.fac, visual));

        type_label := vcRegion.Create ("", vcLabel.fac, type[TYPE_NAME]);
        vcRegion.Font (type_label, vgFont.smalli);
        vcTable.Add (table, type_label);

        function small_button(str, f) {
            label:= vcRegion.Create ("", vcLabel.fac, str);
            vcRegion.Font (label, vgFont.tiny);
            border := vcRegion.Create ("", vcBorder.fac, label);
            button := vcRegion.Create (str, vcButton.fac, f);
            vcRegion.FocusInhibit (button);
            vcRegion.AddChild (button, border);
            vcButton.HighlightRegion (button, border);
            button;
        } /* function small_button */

        fields_column := 0;
        obj[FIELDS] = func(str) {
            l := vcRegion.Create ("", vcLabel.fac, str);
            vcRegion.Font (l, vgFont.tiny);
            if (fields_column == 0) {
                vcTable.Add (table, l);
                fields_column = 2;
            } /* if */
            else
                vcTable.Set (table, 0, fields_column, l);
        };

        show_fields (obj[PATHNAME], obj[FIELDS], obj[TMS_DATA]);
        if (DO_INFO_BUTTON && (obj[TMS_DATA] != nil))
            vcTable.Add (table, small_button ("i", popup_info_cmd));

        if (nchildren > 0) {
            function do_expand() {
                row := obj_row (files, obj);
                vcPicklist.SetPick (files, row);
                toggle_expand (files, obj, row);
            } /* function do_expand */
            label := sprintf ("%d element%s", nchildren,
                              (nchildren == 1) ? "" : "s");
            vcTable.Add (table, small_button (label, do_expand));
        } /* if */

        if (has_view) {
            obj[VIEW_BUTTON] = small_button ("View",
                                             func() view_obj (files, obj););
            vcTable.Add (table, obj[VIEW_BUTTON]);
        } /* if */

        vcTable.Col (table, vcTable.All, vcTable.Border.Invisible, 5, 5, 5,
                     vgAlign.LowerLeft);
        vcTable.Col (table, 3, vcTable.Border.Invisible, 5, 5, 500,
                     vgAlign.LowerLeft);
        table;
    } /* function find_visual */

    function view_obj(files, obj) {
        vcPicklist.SetPick (files, obj_row (files, obj));

        if (obj[VTEXT] == nil) {
            obj[VTEXT] = vtx.CreateWindow (obj_filename (obj), TRUE);
            vtx.ReadOnly (obj[VTEXT], TRUE);
            vcRegion.DestroyNotify (vtx.Region (obj[VTEXT]), func(r) {
                obj[VTEXT] = nil;
            });
        } /* if */
        else
            vcWin.Close (vcRegion.Window (vtx.Region (obj[VTEXT])));
    } /* function view_obj */

    /*
     * Look for an object by its symbolic name in the picklist,
     * returning its row # or vcPicklist.NotSet.
     */
    function find_row(files, symbolic_name) {
        ret := vcPicklist.NotSet;
        vcPicklist.Dimensions (files, &n_rows, &n_cols);
        for (row := 0; (row < n_rows) && (ret == vcPicklist.NotSet); ++row) {
            obj := cdr (vcPicklist.Get (files, row, 0));
            if ((obj != nil) &&
                string_eq (obj[SYMBOLIC_NAME], symbolic_name) &&
                (obj[TYPE] == LibraryT))
                ret = row;
        } /* for */
        ret;
    } /* function find_obj */

    /*
     * Look for the specified library pathname in the picklist,
     * selecting it if found, or unselecting everything if not
     * found.  If a pathname doesn't contain a valid library,
     * then print a message to the status window.
     */
    function select_pathname(files, lib_pathname) {
        row := vcPicklist.NotSet;
        catch (vlEx.FAILURE)
            row = find_row (files, lib2symbolic (lib_pathname));
        exception {
            w := vcRegion.Window (files);
            vcWin.ClearStatus (w);
            vcWin.AddStatus (w, "Corrupt library: " +
                             vlErr.String (vlErr.CurrentError (),
                                           vlErrCntx.StatusLine));
        } /* exception */

        vcPicklist.SetPick (files, row);
    } /* function select_pathname */

    /*
     * Build an object, and add it to the picklist, either at the end
     * of the picklist, or at the current insertion point, depending on
     * boolean add_to_end.
     */
    function add_obj(files, type, pathname, symbolic_name, add_to_end, parent)
    {
        /*
         * If the symbolic_name was not supplied, use the object's name
         * function to figure it out.
         */
        if (symbolic_name == nil)
            symbolic_name = (type[NAME_FUNC]) (pathname);
        
        if (find_row (files, symbolic_name) != vcPicklist.NotSet) {
            w := vcRegion.Window (files);
            vcWin.ClearStatus (w);
            vcWin.AddStatus (w, sprintf ("%s `%s' is already in the list",
                                         type[TYPE_NAME], symbolic_name));
            return (nil);
        } /* if */

        /* Construct the vector to hold the data for this object. */
        obj := [type,
                symbolic_name,
                pathname,
                find_children (pathname, type),
                FALSE,          /* IS_EXPANDED */
                nil,            /* VIEW_BUTTON */
                parent,
                (parent == nil) ? 0 : parent[NSPACES] + NSPACES_PER_LEVEL,
                nil,
                nil,            /* time */
                nil,
                read_lib_tms (vsMerge ("", pathname, "vantage", "tms"))
                ];

        /* If this object is already in the picklist, gently tell the user. */
        visual := find_visual (files, obj);
        if (vcPicklist.SetByString (files, visual)) {
            w := vcRegion.Window (files);
            vcWin.ClearStatus (w);
            vcWin.AddStatus (w, sprintf ("%s `%s' is already in the list",
                                         type[TYPE_NAME], symbolic_name));
        } /* if */
        else {
            if (add_to_end) {
                declare n_rows, n_cols;
                vcPicklist.Dimensions (files, &n_rows, &n_cols);
                vcPicklist.Set (files, visual, obj, n_rows, 0);    
                vcPicklist.SetPick (files, n_rows);
                if (n_rows == 0)
                    set_working (files, 0, TRUE);
            } /* if */
            else
                vcPicklist.Add (files, visual, obj);
        } /* else */
        obj;
    } /* function add_obj */

    /*
     * This exported function adds a top-level library to the end of
     * the picklist.
     */
    function add_lib(files, symbolic_name, pathname)
        add_obj (files, LibraryT, pathname, symbolic_name, TRUE, nil);

    /* Hide any the objects displayed underneath this one. */
    function collapse(files, obj, row) {
        if (obj[IS_EXPANDED]) {
            nchildren := length (obj[CHILDREN]);

            for (r := 0; r < nchildren; ++r) {
                sub_obj := cdr (vcPicklist.Get (files, r + row + 1, 0));
                collapse (files, sub_obj, r + row + 1);
            } /* for */
            vcPicklist.DeleteRows (files, row + 1, nchildren);
            obj[IS_EXPANDED] = FALSE;
        } /* if */
    } /* function collapse */

    function expand(files, obj, row, nchildren) {
        vcPicklist.InsertRows (files, row + 1, nchildren);
        foreach (type_subdir; vscript.list_iter (obj[CHILDREN]))
            add_obj (files, car (type_subdir), cdr (type_subdir),
                     nil, FALSE, obj);
        obj[IS_EXPANDED] = TRUE;
    } /* function expand */

    /*
     * This is the heart of the browser UI.  It is called when someone
     * double-clicks on something int the picklist.  It shows the children
     * if they are hidden, and hides them if they are visible.
     *
     * There are 4 possible cases for an object, and each of them require
     * a different UI.  This routine handles everything.
     *
     * object situation         toggle_expand sequence
     * -----------------------------------------------
     * no children, no view     Nothing
     * view, no children        show text, hide text
     * no view, children        show children, hide children
     * view & children          show children, show text, hide children,
     *                                  hide text
     */
    function toggle_expand(files, obj, row) {
        vcWin.WithBusy (vcRegion.Window (files), func() {
            if (row == nil)
                row = obj_row (files, obj);
            nchildren := length (obj[CHILDREN]);

            if (nchildren > 0) {
                if (obj[IS_EXPANDED])
                    collapse (files, obj, row);
                else
                    expand (files, obj, row, nchildren);
            } /* if */
            else if (obj[VIEW_BUTTON] != nil)
                view_obj (files, obj); /* takes care of its own toggling */
        });
    } /* function toggle_expand */

    /* These predicate and observer functions are needed by vansetup.vs... */

    function is_library(obj)
        obj[TYPE] == LibraryT;

    function is_deletable(obj)
        TRUE;

    function obj_filename(obj)
        vsMerge ("", obj[PATHNAME], "vantage", "vdl");

    function obj_pathname(obj)
        obj[PATHNAME];

    function obj_time(obj) {
        if (obj[TIME] == nil)
            obj[TIME] = car (vsInfo (pathname));
        obj[TIME];
    } /* function obj_time */

    function obj_name(obj)
        obj[TYPE][TYPE_NAME] + " " + obj[SYMBOLIC_NAME];

    function obj_symname(obj)
        obj[SYMBOLIC_NAME];

    function has_view(obj)
        obj[VIEW_BUTTON] != nil;

    function has_children(obj)
        obj[CHILDREN] != nil;

    function obj_row(files, obj) {
        declare n_rows, n_cols;

        vcPicklist.Dimensions (files, &n_rows, &n_cols);
        row := n_rows - 1;
        while (cdr (vcPicklist.Get (files, row, 0)) != obj)
            --row;
        row;
    } /* function obj_row */

    /*
     * This is called by vansetup.vs when an object is deleted from the
     * disk.  It should also disappear from the browser, and the count
     * in the parent's visual should be decremented.
     */
    function reset_parent(files, obj, row) {
        obj = obj[PARENT_OBJ];
        row = obj_row (files, obj);
        collapse (files, obj, row);
        obj[CHILDREN] = find_children (obj[PATHNAME], obj[TYPE]);
        vcPicklist.Set (files, find_visual (files, obj), obj, row, 0);
        if (obj[CHILDREN] != nil)
            toggle_expand (files, obj, row);
    } /* function reset_parent */

    fields := vlPref.Val (prefs, "fields");

    function set_fields(files, f) {
        if (fields != f) {
            fields = f;
            foreach (r, c, i, is_set, vis, obj; vcPicklist.Items (files))
                show_fields (obj[PATHNAME], obj[FIELDS], obj[TMS_DATA]);
        } /* if */
    } /* function set_fields */

    /*
     * update is called after an analyze operation.  It records which
     * libraries are there, and which design elements are visible due to
     * expansion.  Then it wipes out the picklist and rebuilds it, so
     * that the timestamp info and number-of-element thingies are updated.
     *
     * We record what's visible by making a list of the objects in the
     * picklist.  So after we clear the picklist, we walk this list,
     * re-adding all the libraries to the picklist, and expanding all
     * the items whose obj[IS_EXPANDED] is true.
     */
    function update(files) {
        selected_obj := vcRegion.Get (files, vscmVT);
        selected_pathname := (selected_obj == nil) ? nil :
            selected_obj[PATHNAME];
        stuff := nil;
        foreach (r, c, i, is_set, vis, obj; vcPicklist.Items (files))
            stuff = cons (obj, stuff);
        declare n_rows, n_cols;
        vcPicklist.Dimensions (files, &n_rows, &n_cols);
        vcPicklist.DeleteRows (files, 0, n_rows);
        foreach (obj; vscript.list_iter (reverse_destructive (stuff))) {
            declare new_obj, new_obj_row;
            if (obj[TYPE] == LibraryT) {
                new_obj = add_lib (files, obj[SYMBOLIC_NAME], obj[PATHNAME]);
                vcPicklist.Dimensions (files, &new_obj_row, &n_cols);
                --new_obj_row;
            } /* if */
            else {
                new_obj_row = vcPicklist.NotSet;
                foreach (r, c, i, is_set, vis, o; vcPicklist.Items (files))
                    if ((o[TYPE] == obj[TYPE]) &&
                        string_eq (o[SYMBOLIC_NAME], obj[SYMBOLIC_NAME]))
                    {
                        new_obj_row = r;
                        new_obj = o;
                        break;
                    } /* if */
            } /* else */
            if (obj[IS_EXPANDED])
                expand (files, new_obj, new_obj_row,
                        length (new_obj[CHILDREN]));
        } /* foreach */
        if (selected_pathname != nil)
            foreach (r, c, i, is_set, vis, obj; vcPicklist.Items (files))
                if (string_eq (obj[PATHNAME], selected_pathname)) {
                    vcPicklist.SetPick (files, r);
                    break;
                } /* if */
    } /* function update */

    function is_set(field) fields + field == fields;

    function show_fields(path, text, tms_data) {
        str := "";

        if (tms_data != nil) {
            function add(field, index, xform) {
                data := tms_data[index];
                if (is_set (field) && (data != nil)) {
                    str += " " + convert (vlStringVT, field) + ": ";
                    if (istype (xform, vlTypespecVT))
                        data = convert (xform, tms_data[index]);
                    else if (xform != nil)
                        data = xform (data);
                    if (!istype (data, vlStringVT))
                        data = convert (vlStringVT, data);
                    str += data;
                } /* if */
            } /* function add */

            f := vanFields;

            function time_str(time) {
                if (time == 0)
                    "?";
                else {
                    declare mm, dd, yyyy, h, m, s;
                    if (vlCvtTime (time, &mm, &dd, &yyyy, &h, &m, &s) == 0)
                        error ("vlCvtTime failed", time);
                    sprintf ("%d/%d/%d %d:%02d:%02d",
                             mm, dd, yyyy, h, m, s);
                } /* else */
            } /* function time_str */

            function sub_2_1(val) (val == 2) ? "on" : "off";
            function sub_1_0(val) (val == 1) ? "on" : "off";

            function ver_hack(major)
                sprintf ("%d.%d", major, tms_data[LBR_FLAG_VSS_VER_MINOR]);

            add (f.SourceFile,         LBR_FLAG_SOURCE_FILE,   nil);
            add (f.SourceTime,         SOURCE_TIME,            time_str);
            if (tms_data[VIF_TIME] != 0) {
                add (f.VIFTime,            VIF_TIME,               time_str);
                add (f.BinTime,            BIN_TIME,               time_str);
                add (f.RangeCheck,         LBR_FLAG_RCHK,          sub_2_1);
                add (f.Debug,              LBR_FLAG_DBGLVL,        sub_2_1);
                add (f.C_Opt,              LBR_FLAG_C_OPTLVL,      sub_1_0);
                add (f.VHDL_Opt,           LBR_FLAG_VHDL_OPTLVL,   sub_1_0);
                add (f.VSSVersion,         LBR_FLAG_VSS_VER_MAJOR, ver_hack);
            } /* if */
        } /* if */

        text (str);
    } /* function show_fields */

    function popup_info_cmd()
        ;

    function design_unit_string(obj) {
        if ((obj[TYPE] == ConfigT) || (obj[TYPE] == ArchitectureT))
            substring (design_unit_string (obj[PARENT_OBJ]), 0, -1)
                + " " + obj[SYMBOLIC_NAME] + "\"";
        else if (obj[TYPE] == EntityT)
            "-eac \"" + obj[SYMBOLIC_NAME] + "\"";
        else if (obj[TYPE] == PackageT)
            "-pkg " + obj[SYMBOLIC_NAME];
        else if (obj[TYPE] == BodyT)
            "-pkgbody " + obj[SYMBOLIC_NAME];
        else
            error ("Can't make design unit string out of selected object",
                   obj[TYPE][TYPE_NAME]);
    } /* function design_unit_string */

    function obj_lib(obj) {
        if (obj[TYPE] == LibraryT)
            obj;
        else
            obj_lib (obj[PARENT_OBJ]);
    } /* function obj_lib */

    iterator configurations(obj) {
        function configs_in_arch(pathname)
            foreach (config; TraverseDir (vsMerge ("", pathname, "*", "cdr"),
                                          vsAttr.Dir))
                iterate (eac2symbolic (config), config);
            
        if (obj[TYPE] == ConfigT)
            iterate (obj[SYMBOLIC_NAME], obj[PATHNAME]);
        else if (obj[TYPE] == EntityT)
            foreach (arch; TraverseDir (vsMerge ("", obj[PATHNAME], "*",
                                                 "adr"), vsAttr.Dir))
                configs_in_arch (arch);
        else if (obj[TYPE] == ArchitectureT)
            configs_in_arch (obj[PATHNAME]);
    } /* iterator configurations */
} /* package browse */
