/*
 *  $Revision: 1.17 $
 *
 *  Facility to display information about active processes and so on, to allow
 *  the user to decide whether to quit. 
 *
 *  Copyright (c) 1992 by Viewlogic Systems, Inc.
 */

/*
 * Maintain a list of active subprocesses by hooking into
 * the create notification and exit notification of vcSys.
 */
active_subprocesses := nil;
vcSys.CallOnCreate (func(sys) {
    active_subprocesses = cons (sys, active_subprocesses);
    vcSys.CallOnExit (sys, func(mode, code)
        active_subprocesses =
                      vscript.delete_list_element (active_subprocesses, sys););
});

/*\Function vcQuit.IgnoreSubprocess -- Don't report info on a subprocess when quitting
 * 
 * DESCRIPTION
 *      Ordinarily, all active subprocesses are noted in the Quit
 *      dialog box.  vcQuit.IgnoreSubprocess prevents `sys' from being
 *      displayed in the quit dialog.
 */
function IgnoreSubprocess(sys)
    active_subprocesses =
        vscript.delete_list_element (active_subprocesses, sys);

function ModifiedWindowsInfo() {
    modified := "";
    foreach (w; vcWin.Loop ()) {
        if (vcWin.IsModified (w) && vcWin.IsMapped (w))
            modified = modified + "     " + vcWin.Title (w) + "\n";
    } /* foreach */
    if (!string_eq (modified, ""))
        modified = "The following windows have been modified:\n" +
            modified + "\n";
    modified;
} /* function ModifiedWindowsInfo */

function ActiveProcessInfo() {
    modified := "";
    if (active_subprocesses != nil) {
        modified = modified +
            "The following subprocesses are still alive:\n";
        for (a := active_subprocesses; a != nil; a = cdr (a))
            modified = modified + "     " + vcSys.Name (car (a)) + "\n";
        modified = modified + "\n";
    } /* if */

    modified;
} /* function ModifiedProcessInfo */

quit_functions := nil;
call_on_quit := nil;

function CallOnQuit(f)
    call_on_quit = cons (f, call_on_quit);

/*
 * Action routine for the Quit command, which closes all the windows.
 */
function Quit(w, e, arg)
{
    foreach (f; vscript.list_iter (call_on_quit))
        catch (vlEx.FAILURE)
            f ();
        exception
            vlErr.Report2 ("vcore", vlErrSev.Error,
                           "Caught exception during quit processing.");

    foreach (w; vcWin.Loop ())
        catch (vlEx.FAILURE)
            vcWin.Destroy (w);
        exception
            vlErr.Report2 ("vcore", vlErrSev.Error,
                           "Caught exception during quit processing.");
}

/*\vscmuserDocpage vcQuit.AddFunction -- add routine to be called on Quit dialog
 *
 * DESCRIPTION
 *      This will add a function to be called during the creation of
 *      the "OK to Quit?" dialog box.  Its purpose is to return either
 *      a text message describing a modified window, or a region
 *      containing controls to alter the behavior of the quit dialog.
 *      It is called without arguments:
 *
 *          f ();
 *
 * CAUTIONS
 *      This functionality is only available in the presence of the
 *      quit dialog box.  It is not available when replaying a macro
 *      which has recorded the quit command, and therefore should not
 *      be used to implement functionality which is always required
 *      prior to quitting; for that, use vcQuit.CallOnQuit.
 */
function AddFunction(f)
    quit_functions = cons (f, quit_functions);

function _Create(dflt, region) {
    strings := ModifiedWindowsInfo () + ActiveProcessInfo ();
    extra_strings := "";
    regions := nil;
    foreach (f; vscript.list_iter (quit_functions)) {
        /*print (f);*/
        val := f ();
        if (istype (val, vcRegionVT))
            regions = cons (val, regions);
        else if (val != nil) {
            if (!istype (val, string))
                val = convert (string, val);
            if (!string_eq (val, ""))
                extra_strings = val + "\n\n" + extra_strings;
        } /* else */
    } /* foreach */

    n_rows := 2 + length (regions);

    strings = strings + extra_strings;
    if (string_eq (strings, "")) {
        strings = nil;
        --n_rows;
    }

    t := vcTable.Create ("QuitDialog", n_rows, 1);
    if (strings != nil)
        vcTable.Add (t, vcRegion.Create ("", vcStaticText.fac, strings));

    foreach (r; vscript.list_iter (regions))
        vcTable.Add (t, r);

    if (vcWv.UI (wv) == vcUi.OpenLook)
        text := "Quit " + vcWin.ProgramName + "?";
    else
	text := "OK to quit " + vcWin.ProgramName + "?";
    vcTable.Add (t, vcRegion.Create ("", vcStaticText.fac, text));

    tbl := vcTable.Create ("quit_tbl", 1, 4);
    vcTable.Add (tbl, vcRegion.Create ("quit_icon", vcIcon.fac, "motif_head"));
    vcTable.Add (tbl, vcRubber.Create ("", 0, 0));
    vcTable.Add (tbl, t);
    vcTable.Add (tbl, vcRubber.Create ("", 0, 0));

    margin := vcRegion.Create ("quit_margin", vcMargin.fac, tbl,
                               left_margin   <- 0,
                               right_margin  <- 10,
                               top_margin    <- 10,
                               bottom_margin <- 5);

    vcRegion.AddChild (region, margin);
    nil;
} /* function _Create */

function _Get(arg1, arg2, arg3) {
    "yes";
} /* function _Get */
