/*
 *  $Revision: 1.202.1.2 $
 *
 *  Window manipulation functions.
 *
 *  Copyright (c) 1992,1994 by Viewlogic Systems, Inc.
 */


/*
 * Displays a dialog box prompting the user to choose whether to cancel the
 * close command or to discard any edits and carry out that command.  Note
 * the calls to vcRegion.PrefSet--by declaring the button regions macro
 * recordable we don't have to create complicated command objects in order to
 * record interactions in this particular dialog box.  This solution is not
 * advised for other dialog boxes as commands will remain truly portable,
 * whereas the region names can change from revision to revision.
 */
function close_or_cancel(w, str)
{
    table := vcTable.Create ("Quit?", 2, 1);
    vcTable.Add (table, vcRegion.Create ("msg", vcStaticText.fac, str));

    buttons := vcRegion.Create ("Quit?", vcButnRow.fac,
                                list (list ("cancel",
                                            "Cancel, do NOT Quit",
                                            func();,
                                            list (macro_on <- TRUE,
                                                  close_window <- TRUE)),
                                      list ("close",
                                            "Discard changes, then Quit",
                                            func() vcWin.Close (w);,
                                            list (macro_on <- TRUE,
                                                  close_window <- TRUE))));
    vcTable.Add (table, buttons);

    vcWin.Popup (w, table, vcButnRow.GetDefault (buttons),
                 vcWin.NoMenus + vcWin.NoCmdline + vcWin.NoStatus);
}

/*
 * Redefine the IsModified function, but only the first time that vcWin.vs
 * is loaded.  This avoids an infinite recursion problem that occurs when
 * vcWin.vs is reloaded for debugging.
 *
 * We redefine it here to check to see whether the owner facility has defined
 * a _Modified method, which is sometimes easier than calling vcWin.Modified.
 */
if (!istype (IsModified, closureVT)) {
    wired_IsModified := IsModified;

    IsModified = func(w) {
        if (!(ret := wired_IsModified (w))) {
            r := vcWin.OwnerRegion (w);
            if (!vcRegion.Destroyed (r)) {
               facenv := vcFac.Package (vcRegion.Fac (r));
               if (isbound (`_Modified, facenv))
                    ret = facenv._Modified (vcRegion.GetData (r), r);
            }
        }
        ret;
    };
}

/*
 * Action routine for the wclose command.  If the window hasn't been modifed,
 * just close it.  Otherwise, get the confirmation message and call the
 * close-or_cancel function to display a dialog box asking the user either to
 * cancel this command or really quit.
 */
function close_if_modified(w, e) {
    /*
     * If this is the last window, then this command is equivalent
     * to a quit command, but not when replaying macros (to save the
     * existing regression tests).  Windows that aren't decorated
     * (cached menu windows) and windows that aren't mapped don't
     * count toward this being the last window.
     */
    nwindows := 0;
    foreach (win; vcWin.Loop ())
        if (((vcWin.Parent (win) == nil) || (win == w)) &&
            vcWin.IsDecorated (win) &&
            vcWin.IsMapped (win))
            ++nwindows;

    if (!vcMac.IsReplayingOn () &&
        !vcMac.IsRecordingOn () &&
        (nwindows == 1) &&
        isbound (`args, vcWv))
        vcCmd.Run (QuitCmd, w, e, nil, FALSE);

    else if (!vcWin.IsModified (w))
        vcWin.Close (w);
    else
    {
        msg := vcWin.ConfirmMessage (w);
        if (msg == nil)
            msg = "
                 The text has been edited. 

 You may discard edits now and quit, or cancel 
 the request to Quit, and go back and either save 
 the contents or store the contents as a new file. 

";
        close_or_cancel (w, msg);
    }
}


wclose := vcWv.Cmd (fac, close_if_modified, "wclose");
vcCmd.SetHelp (wclose, "Close this window");

function DismissPopups(w, e) {
    ndismissed := 0;
    foreach (child; vcWin.Loop ())
        if (vcWin.Parent (child) == w) {
            ++ndismissed;
            vcWin.Close (child);
            vcWin.ClearStatus (w);
            vcWin.AddStatus (w, sprintf ("Dismissed %d popups.", ndismissed));
        }
}

DismissPopupsCmd := vcWv.Cmd (vcWin.fac, DismissPopups, "dismiss_popups");
vcCmd.SetHelp (DismissPopupsCmd,
               "Dismiss all popups (children) of this window");

/*
 * Quitting.
 */

vcFac.Create ("vcQuit", vscmVT);

QuitCmd := vcWv.Cmd (fac, vcQuit.Quit, "quit", "Quit",
                     vcWv.Param ("", vlStringVT, vcWv.Constant, "",
                                 vcQuit.fac));

function do_eval(w, ev, str) {
    str = convert (vlStringVT, str);
    last := substring (str, -1, 0);
    complete_str := str;
    if (!string_eq (last, ";") && !string_eq (last, "}"))
        complete_str = str + ";";
    vcWin.ClearStatus (w);
    vcWin.AddStatus (w, sprintf ("eval (%s)", str));
    vcWin.AddStatus (w, sprintf (" = %s", _Eval (w, ev, complete_str)));
}

eval_cmd := vcWv.Cmd (fac, do_eval, "eval",
                      "Evaluate ViewScript Expression",
                      vcWv.Param ("ViewScript", vlStringRestOfLine1VT,
                                  vcWv.Sticky, ""));

load_cmd := vcWv.Cmd (fac, func(w,e,s) _Eval (w,e,sprintf("load(\"%s\");",s));,
                      "load", "Load ViewScript File",
                      vcWv.Param ("", vlStringVT, vcWv.Sticky, "",
                                  vcFileSelect.fac, filter <- "*.vs"));

function build_vscript_window(language)
{
    func(win, event)
    {
        region := vcRegion.Create ("ViewScript Interpreter", vscmEval.fac, nil,
                                    macro_on <- TRUE,
                                    language <- language);
        win := vcWin.Create (region, vcWin.NoBindings);
        vcWin.SetIconName (win, "vscript");
    };
}

ViewScriptCmd := vcWv.Cmd (fac, build_vscript_window (vscmLanguage.VS),
                           "vscript");
vcCmd.SetHelp (ViewScriptCmd, "Invoke the ViewScript Interpreter");

SchemeCmd     := vcWv.Cmd (fac, build_vscript_window (vscmLanguage.SCM),
                           "scmint");
vcCmd.SetHelp (SchemeCmd, "Invoke the Scheme Interpreter");


/*
 * Turn on stdout for the ViewScript's printf statements.
 */
function stdout_on(win, event, flag)
    set_stdout (flag == vlOnOff.On);

default_val   := vlPref.Val (vlPref.Create ("vscript"), "stdout");
stdout_on_cmd := vcWv.Cmd (fac, stdout_on, "stdout_on", "ViewScript Output",
                           vcWv.Param ("stdout On", vlOnOffVT, vcWv.Sticky,
                                       default_val));
vcCmd.EnableMacro (stdout_on_cmd, vlOnOff.Off);


/*
 * This function is used mainly for debugging purposes.
 */
function turn_gc_on(win, event)
{
    gc_status (TRUE);
    purge ();
    memory ();
}

DefaultActionCmd := vcWv.Cmd (fac, DefaultAction, "DefaultAction");
vcCmd.EnableMacro (DefaultActionCmd, vlOnOff.Off);
vcBind.Set (vcWv.bindings, '\enter', DefaultActionCmd);
vcBind.Set (vcWv.bindings, '\leftdbl', DefaultActionCmd);
MenuBarCmd :=
    vcWv.Cmd (fac,
              func (win, ev) {
                  rgnMenus := vcRegion.FindFac (vcWin.Region (win),
                                                vcMenuPop.fac);
                  vcRegion.SetFocus (rgnMenus, rgnMenus);
                  vcMenuPop.MenuBar (rgnMenus, ev);
              });
vcCmd.EnableMacro (MenuBarCmd, vlOnOff.Off);
/*vcBind.Set (vcWv.bindings, '\F10', MenuBarCmd);*/

vcBind.Set (vcWv.bindings, '\^\F1',        vcWv.Cmd (fac, turn_gc_on));
vcBind.Set (vcWv.bindings, '\_\F4',        wclose);
vcBind.Set (vcWv.bindings, '\_\^C',        wclose);
vcBind.Set (vcWv.bindings, '\_\rightdown', wclose);
vcBind.Set (vcWv.bindings, '\_\^q',        QuitCmd);
vcBind.Set (vcWv.bindings, '\_\^V',        ViewScriptCmd);
vcBind.Set (vcWv.bindings, '\^]',          stdout_on_cmd);

/*
 * Entering the command line " " causes a completions window that
 * shows all the available named commands at that point.  This is
 * an excellent candidate for the help key.
 */
ShowCmds := vcWv.Cmd (fac, func(w, e) vcWin.RunCmdline (w, " ", e);,
                       "ShowCommands");
vcCmd.EnableMacro (ShowCmds, vlOnOff.Off);

vcBind.Set (vcWv.bindings, '\_\F1',         ShowCmds);
vcBind.Set (vcWv.bindings, '\_\help',       ShowCmds);

arrow_cmd := vcWv.Cmd (fac, Arrow);
vcBind.Set (vcWv.bindings, '\left',         arrow_cmd);
vcBind.Set (vcWv.bindings, '\right',        arrow_cmd);
vcBind.Set (vcWv.bindings, '\up',           arrow_cmd);
vcBind.Set (vcWv.bindings, '\down',         arrow_cmd);
vcBind.Set (vcWv.bindings, '\^\#\left',     arrow_cmd);
vcBind.Set (vcWv.bindings, '\^\#\right',    arrow_cmd);
vcBind.Set (vcWv.bindings, '\^\#\up',       arrow_cmd);
vcBind.Set (vcWv.bindings, '\^\#\down',     arrow_cmd);

focus_next_cmd := vcWv.Cmd (fac,
                            func(w, e) vcRegion.FocusNext (vcWin.Region (w)););
focus_prev_cmd := vcWv.Cmd (fac,
                            func(w, e) vcRegion.FocusPrev (vcWin.Region (w)););
vcBind.Set (vcWv.bindings, '\tab',   focus_next_cmd);
vcBind.Set (vcWv.bindings, '\^\tab', focus_next_cmd);
vcBind.Set (vcWv.bindings, '\#\tab', focus_prev_cmd);
vcBind.Set (vcWv.bindings, '\#\^\tab', focus_prev_cmd);
/*vcBind.Set (vcWv.bindings, '\enter', focus_next_cmd);*/

load ("vcTextEdit.vs");

vcBind.Set (vcWv.bindings, '\_\^]',
            vcWv.Cmd (fac,
                      func (w, event, str) SetTitle (w, str);,
                      "set_title", "Set Window Title",
                      vcWv.Param ("New window title", vlStringVT, vcWv.Sticky,
                                  "")));

/*****************************************************************************
 *
 * The basic macro commands are defined in the vcMac.vs file.  The new user
 * interface to the macro system is defined as the Macro System Panel facility
 * in the file vcMacPanel.vs.  The vcMac.vs file is *loaded* into this file,
 * i.e. all those commands are accessible through the vcWin package.  The new
 * commands for the Macro System Panel (record, stop, play and cancel) are
 * accessible through the vcMacPanel package.
 *
 *****************************************************************************/

load ("vcMac.vs");


function build_macro_panel(win, event)
{
    region := vcRegion.Create ("Macro System Panel", vcMacPanel.fac, event);
    vcWin.Create (region, vcWin.NoCmdline + vcWin.NoBindings);
}
MacPanelCmd := vcWv.Cmd (fac, build_macro_panel, "mpanel");
vcCmd.SetHelp (MacPanelCmd, "Graphical macro interface");

vcBind.Set (vcWv.bindings, '\_\^M', MacPanelCmd);


function build_macro_trans(win, event)
{
    vcWv.RequireFac (wv, "vcMacTrans");

    region := vcRegion.Create ("Macro Translator", vcMacTrans.fac, event);
    vcWin.Create (region, vcWin.NoCmdline + vcWin.NoBindings + vcWin.NoStatus
                  + vcWin.NoMacros);
}
MacTransCmd := vcWv.Cmd (fac, build_macro_trans, "mtrans");
vcCmd.SetHelp (MacTransCmd, "4.x -> Powerview macro translator");

/*****************************************************************************/

grab   := GrabData;         /* Old style, DON'T USE IT! */
ungrab := UngrabData;       /* Use vcRegion.SetGrab     */

load ("vcSystem.vs");


plotting_loaded := FALSE;

function plot_invoke_worksheet(win, event) {
    function do_load() {
        if (!plotting_loaded) {
            plotting_loaded = TRUE;
            vcFac.Create ("vspPlot", vscmVT);
        }
        vspPlot.PlotSetup (win, event);
    }

    if (vsp.vspPlotEnabled ())
        vcWin.WithBusy (win, do_load);
    else
        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                     "Plotting is not enabled.");

}

PLOT_ERROR_STR := "Problem invoking plotter, ";
err_strs := "";
reported_error := TRUE;

/* Enables/disables all plotting commands.  They should be disabled
 * when a plot is in process and should be enabled when no plot is
 * in process.
 */
function enable_plot_cmds (val) {
        vcCmd.Enable (plotgo_cmd, val);
        vcCmd.Enable (plotgodefs_cmd, val);
        vcCmd.Enable (setplot_cmd, val);
        vcCmd.Enable (PlotCmd, val);
}

/* 
 * Writes out the metafile, spawns the plotter, and attempts to spool the
 * plot.  This is called by facilities other than Viewfile from both the
 * plotting worksheet and from the command line.
 */
function plot_cont (w, e) {
    vcWin.ClearStatus (w);
    vcWin.AddStatus (w, "Creating Metafile..");
    if (plotting_loaded)
        if (vspPlot.plot_win_id != nil) {
            vcWin.ClearStatus (vspPlot.plot_win_id);
            vcWin.AddStatus (vspPlot.plot_win_id, "Creating Metafile..");
        }

    vcWin.Busy (w, TRUE);
    vsp.vspPlotMeta (w, e);
    dir_buf := vsp.vspGetDirBuf ();
    if (string_length (dir_buf) == 0)
        dir_buf = string_append (dir_buf, ".");

    shell_cmd := "splplt " + "-m " + vsp.vspGetMaybeDev () + " -d " +
        dir_buf + " -f " + vsp.vspGetFileBuf ();

    if (plotting_loaded) {
        vspPlot.CreateInfoBox ("Issuing the plot..");
        if (vspPlot.plot_win_id != nil)
            vcWin.ClearStatus (vspPlot.plot_win_id);
    }
    vcWin.ClearStatus (w);
    enable_plot_cmds (vlOnOff.Off);

    sys := vcSys.Command (shell_cmd, nil,
                func(output) printf ("%s\n", output);,
                func(output) {

                    /* Buffer the error messages until the prog. terms. */
                    if (reported_error) {
                        err_strs = string_copy (PLOT_ERROR_STR);
                        reported_error = FALSE;
                    }
                    else
                        err_strs = err_strs + "\n";
                    err_strs = err_strs + output;
                },
                func(mode, code) {
                    enable_plot_cmds (vlOnOff.On);
                    if ((mode == vcSysMode.Exited) && (code != 0)) {
                        err_strs = err_strs + "\n"
                            + "Plotter exited with status "
                                + number2string (code);
                        reported_error = TRUE;
                        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                                     err_strs);
                    }
                    else if (mode == vcSysMode.Signaled) {
                        err_strs = err_strs + "\n"
                            + "Plotter received signal "
                                + number2string (code);
                        reported_error = TRUE;
                        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                                     err_strs);
                    }
                    else {
                        if (vcWin.IsAlive (w)) {
                            vcWin.ClearStatus (w); 
                            vcWin.AddStatus(w,
                                            "Plotter finished.");
                            if (plotting_loaded)
                                vspPlot.UpdateInfoBox
                                  ("Plot complete.  Please Dismiss this box.");
                        }
                    }
                }, nil);

    /*
     * If a macro is being replayed, tell it to wait for the ploting command
     * to finish.  Otherwise, it will keep executed commands that might depend
     * on the plot command having finished.
     */
    if (vcMac.IsReplayingOn ())
        vcMac.SubprocessHold (sys, 0);

    vcWin.Busy (w, FALSE);
}

/* This needs to be called from within the plotting worksheets and
 * from the facility directly as well.  Writes out the spcurins.ins
 * and meta files, spawns the plotter, and attempts to spool the plot.
 * The plotting worksheet does not pop up.
 */
function plot_it (w, e) {
    vsp.vspWriteInsFile (w, e);
    fac_name := vsp.vspRetFacName (vcRegion.Fac (vcWin.OwnerRegion (w)));

    /*
     * !!NOTE the hardcoding of Cockpit's facility name!!  Viewframe should
     * only write out the dev.des file.  All other facilities need to do more.
     */
    if (string_eqi (fac_name, "vframe") == FALSE)
        plot_cont (w, e);
}

/* Issues a plot without having to bring up the plotting worksheets.
 * Goes with the default file name and scale based on information
 * retrieved from the individual facilities.  This is useful for the
 * macro that is created by Workview to plot projects.
 */
function plot_go_defs (win, event) {

    if (! vsp.vspPlotEnabled ())
        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                     "Plotting is not enabled.");

    /* Make sure there is an spcurins.ins file */
    vsp.vspCheckCurins ();

    /* Go with facility default file name */
    if (string_eqi (vsp.get_def_output_type (), "file"))
        vsp.vspGetFacFileName (win);

    /* Send the fac. window id to plotting software so that scale is ok */
    vsp.vspSetKludgeWindow (win);

    /* Have software calculate the scale factor */
    vsp.vspUpdateScale ();

    plot_it (win, event);
}

/* Issues a plot without having to bring up the plotting worksheets.
 * Goes with the default scale based on information retrieved from the 
 * individual facilities.  It is different from plot_go_defs only because
 * this routine uses the currently set file name and not the facility
 * default.  This is useful for macros run in plotting regression tests.
 */
function plot_go(win, event) {

    if (! vsp.vspPlotEnabled ())
        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                     "Plotting is not enabled.");

    /* Make sure there is an spcurins.ins file */
    vsp.vspCheckCurins ();

    /* Send the fac. window id to plotting software so that scale is ok */
    vsp.vspSetKludgeWindow (win);

    vsp.get_def_output_name ();

    /* Have software calculate the scale factor */
    if (!vsp.vspUserSetScale ())
        vsp.vspUpdateScale ();

    /* Issue the plot */
    plot_it (win, event);
}

PlotCmd := vcWv.Cmd (fac, plot_invoke_worksheet, "plotsetup");
vcCmd.SetHelp (PlotCmd, "Setup plotting worksheet");

/*
 * Define the 'plotgodefs' command.  It issues a plot directly, given the
 * current plot settings.  Uses default file name and scale.  Creates a new
 * spcurins.ins. 
 */
plotgodefs_cmd := vcWv.Cmd (fac, plot_go_defs, "plotgodefs");
vcCmd.SetHelp (plotgodefs_cmd,
               "Issue a plot.  Uses default file name and scale.");

/*
 * Define the 'plotgo' command.  It issues a plot directly, given the current
 * plot settings.  Uses current file name and def. scale.  Creates a new 
 * spcurins.ins file.
 */
plotgo_cmd := vcWv.Cmd (fac, plot_go, "plotgo");
vcCmd.SetHelp (plotgo_cmd,
               "Issue a plot.  Uses current file name and default scale.");

/*
 * Define the 'setplot' command.
 */
function setplot(w, e, comm, val)
{
    if (! vsp.vspPlotEnabled ())
        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                     "Plotting is not enabled.");
    if (string_eq (comm, ""))
        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                     "Please specify a valid setplot command");
    if (string_eq (val, ""))
        vlEx.Throw2 (vlEx.FAILURE, "vcWin", vlErrSev.Error,
                     "Please specify a valid setplot command modifier");

    vsp.vspSetKludgeWindow (w);
    if (vsp.sp_sapopt (string_append (comm, " ", val)))
        vsp.vspWriteInsFile (w, e);
}
setplot_cmd := vcWv.Cmd (fac, setplot, "setplot", "Set Plot", 
                         vcWv.Param ("Setplot command", vlStringVT,
                                     vcWv.Constant, ""),
                         vcWv.Param ("Command modifier(s)", vlStringVT,
                                     vcWv.Constant, ""));

vcCmd.EnableMacro (setplot_cmd, vlOnOff.On);

/*
 * Action routine for the `Prefs' command to show and edit all the preferences
 * for a given facility.
 */
function Prefs(w, e)
    vcWin.WithBusy (w, func() {
        vcWv.RequireFac (wv, "vcPref");
        vcPref.CreateWindow (w, vcRegion.Fac (vcWin.OwnerRegion (w)));
    });


PrefsCmd := vcWv.Cmd (fac, Prefs, "preferences");
vcCmd.SetHelp (PrefsCmd, "Edit preferences for the current facility");
vcBind.Set (vcWv.bindings, '\F13', PrefsCmd);

/*
 * Function to flash the screen when an unbound key is pressed.
 */
function UnboundKey(w, e) {
    vcWin.ClearStatus (w);
    vcWin.AddStatus (w, sprintf ("Key %c is unbound", vcEvent.Ch (e)));
    vcWin.VisualBell (w);
}

UnboundKeyCmd := vcWv.Cmd (vcWin.fac, UnboundKey, "_");
vcCmd.EnableMacro (UnboundKeyCmd, FALSE);

/*
 * Bind all unbound keys in vcWv.bindings to UnboundKeyCmd.
 */
foreach (c; vcWv.AllChars ())
    if (vcBind.Get (vcWv.bindings, c) == nil)
        vcBind.Set (vcWv.bindings, c, UnboundKeyCmd);

/*
 * Go back and remove any UnboundKeyCmd bindings for keys that tend to get
 * pressed on their way to (or back from) other operations.  It's more
 * efficient to do it this way than to check for 12 different keys in
 * the vcWv.AllChars loop above.
 */
foreach (c; vscript.list_iter (list ('\move',
                                     '\altdown', '\altup',
                                     '\ctrldown', '\ctrlup',
                                     '\shiftdown', '\shiftup',
                                     '\leftup', '\middleup', '\rightup',
                                     '\leftdbl', '\middledbl', '\rightdbl')))
    if (vcBind.Get (vcWv.bindings, c) == UnboundKeyCmd)
        vcBind.Set (vcWv.bindings, c, nil);


/*
 * Auxillary routine for the ShowBindings function.  For each character in the
 * scan table check if it has a corresponding binding.  If it does, add the
 * name of the binding and the command to the ViewText window.
 */
function add_bindings(tx, bindings) {
    declare cmd, c;

    for (i := 0; i < 1023; ++i) {
        c = int2char (i);
        cmd = vcBind.Get (bindings, c);
        if ((cmd != nil) && (cmd != UnboundKeyCmd))
            vtx.FastAppend (tx, sprintf ("%-20s %-20s %s\n",
                                         vcWv.CharToString (c),
                                         vcCmd.Name (cmd),
                                         (((str := vcCmd.Help (cmd)) == nil) ?
                                         "" : str)));
    }
}

/*
 * Action routine for the `ShowBindings' command.  It creates a ViewText
 * window in which all the current bindings will be displayed.  The actual
 * displaying is done by the add_bindigns function defined above.
 */
function ShowBindings(window, event)
    vcWin.WithBusy (window, func() {
        tx := vcRegion.Create ("log", vtx.fac, "bindings.log",
                               scratch <- TRUE);

        facs := vlSym.ObjCreate (101);
        foreach (r; vcRegion.Children (vcWin.OwnerRegion (window), TRUE)) {
            fac := vcRegion.Fac (r);
            if (facs[fac] == nil) {
                facs[fac] = TRUE;

                if (vcBind.NumBindings (vcFac.Bindings (fac)) > 0) {
                    vtx.AppendString (tx, sprintf ("%s Bindings\n\n",
                                                   vcFac.Name (fac)));
                    add_bindings (tx, vcFac.Bindings (fac));
                    vtx.AppendString (tx, "\n\n");
                } /* if */
            } /* if */
        } /* foreach */

        vtx.AppendString (tx, "Global Bindings\n\n");

        add_bindings (tx, vcWv.bindings);

        vtx.FastAppendFinished (tx);
        vtx.ReadOnly (tx, TRUE);
        vcWin.Create (tx, 0);
    });

ShowBindCmd := vcWv.Cmd (vcWin.fac, ShowBindings, "ShowBindings");
vcCmd.SetHelp (ShowBindCmd, "Show all the bindings currently in effect");


/*
 * This function is provided for upward compatibility only.  Don't use the
 * vcWin.create (lowercase) function!!!  Use the vcWin.Create (uppercase)
 * function to create windows.
 */
function create(region, flags)
    _Create (wv, nil, region, nil, flags);

/*\vscmuserDocpage vcWin.Create -- Create a window for a region hierarchy.

DESCRIPTION
     Given a region hierarchy, create a window to fit it.  The region's ideal
     size is used to create the window, given the limitations of the
     workstation screen.  Other limitations on a reasonable size for any given
     window should be set up by the facility associated with the top level
     region.

NOTE
     This function takes a variable number of arguments.  The only required
     argument is the region.  If the icon_name argument is specified, it
     should be the filename of a bitmap file (no ext) searched for in WDIR.

     The flags can be added together and can be any of the following:

         vcWin.NoFlags          Default
         vcWin.NoCmdline        No command line window area
         vcWin.NoStatus         No status strip on the bottom
         vcWin.NoMenus          No menu strip at the top
         vcWin.NoDecoration     No window manager borders (used for menus)
         vcWin.NoBindings       No strip showing mouse bindings
         vcWin.Bindings         Display current mouse bindings 
         vcWin.NoMacros         No macros recorded here
         vcWin.BackingStore     Use backing store when window is obscured
         vcWin.Hidden           Don't display this window initially;
                                it can be made viewable with vcWin.Show

RETURNS
     A window handle.

SYNOPSIS
     vcWindowVT vcWin.Create([parent,] region [,icon_name] [,flags])
        [vcWindowVT parent;]
         vcRegionVT region;
        [vlStringVT icon_name;]
        [vlMaskVT   flags;]
*/
function Create(args...) {
    declare parent, region, icon_name, flags;

    if (istype (car (args), vcWindowVT)) {
        parent = car (args);
        args = cdr (args);
    }

    region = car (args);
    args = cdr (args);

    if (istype (car (args), vlStringVT)) {
        icon_name = car (args);
        args = cdr (args);
    }

    if (args != nil)
        flags = car (args);
    else
        flags = vcWin.NoFlags;
    _Create (wv, parent, region, icon_name, flags);
}

/*
 * Functions related to starting up viewhelp.
 */
function HelpCmdline(data, event, rest_of_line) {
    win  := vcEvent.Window (event);
    name := vcFac.Name (vcRegion.Fac (vcWin.OwnerRegion (win)));
    parm := convert (vlStringVT, rest_of_line);
    vcWin.WithBusy (win, func() vcAlert.SpawnHelp (win, name, parm););
}
function HelpInteractive(data, event)
    HelpCmdline (data, event, "");
HelpCmdlineCmd := vcWv.Cmd (fac, HelpCmdline, "help", "Help",
                            vcWv.Param ("Help for subject",
                                        vlStringRestOfLineVT,
                                        vcWv.Constant, ""));
HelpInteractiveCmd := vcWv.Cmd (fac, HelpInteractive, "_help");
vcBind.Set (vcWv.bindings, '\F1',           HelpInteractiveCmd);
vcBind.Set (vcWv.bindings, '\help',         HelpInteractiveCmd);


/*
 * These are abbreviated for ease of typing.
 */
if (!isbound (`Flag, current_package ())) /* Upward compatibility... */
    Flag := vcWinFlag;

NoFlags         := Flag.NoFlags;
NoCmdline       := Flag.NoCmdline;
NoStatus        := Flag.NoStatus;
NoMenus         := Flag.NoMenus;
NoDecoration    := Flag.NoDecoration;
Pinup           := Flag.Pinup;
NoBindings      := Flag.NoBindings;
Bindings        := Flag.Bindings;
NoMacros        := Flag.NoMacros;
BackingStore    := Flag.BackingStore;
Hidden          := Flag.Hidden;

function menu_xlate(win, event, old_menu_file, new_menu_file)
{
    /*
     * First, find the facility data object of the facility that is in charge
     * of the window where the menu translation occurs.  Note that that value
     * is used in the nested functions below.  This can happen because of the
     * lexical scoping rules inherited from Scheme.
     */
    the_fac := vcRegion.Fac (vcWin.OwnerRegion (win));

    function do_preview(w, event)
        vcWin.Create (vcRegion.Create ("Menu Previewer", vcMenuPop.fac,
                                       load (new_menu_file)));
    preview_cmd := vcWv.Cmd (fac, do_preview, "preview");
    vcCmd.SetHelp (preview_cmd,
                   "Bring up the translated menus in a temporary window");

    function do_install(w, event) {
        buttons := vcRegion.Create ("Confirm Installation", vcButnRow.fac,
                    list (list ("yes", "Yes, install new menus",
                                func()
                                {
                                    m := load (new_menu_file);
                                    vcFac.SetMenus (the_fac, m);
                                    vcFrame.ResetMenus (vcWin.Region (win), m);
                                }, list (close_window <- TRUE)),
                          list ("no", "Don't install",
                                func ();, list (close_window <- TRUE))));
        vcWin.Popup (vcEvent.Window (event), buttons,
                     vcButnRow.GetDefault (buttons),
                     vcWin.NoMenus + vcWin.NoCmdline + vcWin.NoStatus);
    }
    install_cmd := vcWv.Cmd (fac, do_install, "install");
    vcCmd.SetHelp (install_cmd,
                   "Install the menus in window " + vcWin.Title (win) +
                   " and use them from now on in facility " +
                   vcFac.Name (the_fac));

    cmd := nil;
    function reset_cmd() {
        cmd = "menu4to5 " + old_menu_file + " " + new_menu_file;
        vcCmd.Enable (preview_cmd, vlOnOff.Off);
        vcCmd.Enable (install_cmd, vlOnOff.Off);
    }

    function exit_handler(mode, code) {
        if ((mode == vcSysMode.Exited) && (code == 0))
        {
            vcCmd.Enable (preview_cmd, vlOnOff.On);
            vcCmd.Enable (install_cmd, vlOnOff.On);
            vcWin.UpdateMenus (xlate_window);
        }
    }

    function retranslate(w, e, old, new) {
        old_menu_file = old;
        new_menu_file = new;
        reset_cmd ();
        vcShell.Restart (xlate_shell, cmd, nil);
        vcSys.CallOnExit (vcShell.Sys (xlate_shell), exit_handler);
    }

    retranslate_cmd := vcWv.Cmd (fac, retranslate, "retranslate",
                            "4.x to Powerview Menu Translator",
                            vcWv.Param ("4.x Menu File", vlStringVT,
                                        vcWv.Sticky, old_menu_file),
                            vcWv.Param ("Powerview Menu File", vlStringVT,
                                        vcWv.Sticky, new_menu_file));

    xlate_menus := vcWv.MakeMenu ("Xlate menus",
                                  vcWin.RedSquareEntry,
                                  list ("Preview", preview_cmd),
                                  list ("Install", install_cmd),
                                  list ("Translate another", retranslate_cmd));

    reset_cmd ();
    xlate_shell := vcRegion.Create ("shell", vcShell.fac, cmd);

    t := vcTable.Make ("Menu Translator", 2, 1,
                       vcRegion.Create ("menus", vcMenuPop.fac, xlate_menus),
                       xlate_shell);
    xlate_window := vcWin.Create (t);

    vcSys.CallOnExit (vcShell.Sys (xlate_shell), exit_handler);
}

function old_menu_default(w)
    vcFac.Name (vcRegion.Fac (vcWin.OwnerRegion (w))) + ".mn1";
function new_menu_default(w)
    vcFac.Name (vcRegion.Fac (vcWin.OwnerRegion (w))) + "mn.vs";

MenuTransCmd := vcWv.Cmd (fac, menu_xlate, "menu4to5",
                            "4.x to Powerview Menu Translator",
                            vcWv.Param ("4.x Menu File", vlStringVT,
                                        vcWv.Dynamic, old_menu_default),
                            vcWv.Param ("Powerview Menu File", vlStringVT,
                                        vcWv.Dynamic, new_menu_default));


/*
 * This function scans a string for an output redirection >file or >>file.
 * If it finds one, it modifies its input arguments fname and mode to determine
 * how to open that file.  Those variables should be declared and then
 * passed by &.  This function always returns the string with any I/O
 * redirection stripped.
 */
function ParseRedirect(str, fname, mode) {
    *fname = nil;
    *mode = "w";

    /* Look for redirection in the string */
    if (lindex (str, '>') != nil) {
        len := string_length (str);
        for (i := 1; (i < len) && (*fname == nil); ++i)
            if ((str[i] == '>') && (str[i - 1] != '\\')) {
                j := i + 1;
                if ((j < len) && (str[j] == '>')) {
                    *mode = "a";
                    ++j;
                }
                printf ("mode=%s, j=%s\n", mode, j);
                rest_of_string := string_token (substring (str, j, 0),
                                                fname, " \t");
                str = substring (str, 0, i) + rest_of_string;
            }
    }
    str;
}

function echo(w, e, str) {
    declare fname, mode;
    str = ParseRedirect (str, &fname, &mode);
    if (fname == nil) {
        vcWin.ClearStatus (w);
        vcWin.AddStatus (w, str);
    }
    else {
        f := fopen (fname, mode);
        fputline (f, str + "\n");
        fclose (f);
    }
}

EchoCmd := vcWv.Cmd (fac, echo, "echo_status", "Echo",
                     vcWv.Param ("", vlStringRestOfLineVT, vcWv.Sticky, ""));

CustomMenu := vcWv.MakeMenu ("Customize",
                             vcWv.Pin (),
                             list ("Menu translator",   MenuTransCmd),
                             list ("Macro translator",  MacTransCmd),
                             list ("Edit Preferences",  PrefsCmd));


function viewdoc(win, event)
    vcWin.RunCmdline (win, "sys -nowindow online", event);
ViewDocCmd := vcWv.Cmd (fac, viewdoc);
vcCmd.SetHelp (ViewDocCmd, "On-line documentation");

/*\vscmuserDocpage vcWin.ReadErrorFileCmd -- command to read a file of errors
 *
 * DESCRIPTION
 *      This is a Core command to read in a file of error messages
 *      and display the results in an alert box.  From the alert
 *      box, the user can do the standard things:  Visit(Next) or
 *      Help.
 *
 * SEE ALSO
 *      Powerview Error Navigation User's Guide.
 */
ReadErrorFileCmd :=
    vcWv.Cmd (fac, func(data, event, filename)
                       vcWv.ReadErrorFile (vcEvent.Window (event), filename);,
              "TrackErrors", "Track (visit) errors from file",
              vcWv.Param("Error file", vlStringVT, vcWv.Sticky, "",
                         vcFilePop.fac, filter <- "*.err", read_now <- TRUE));
vcCmd.SetHelp (ReadErrorFileCmd, "Track (visit) file of error messages");

HelpOnErrorMsgsCmd :=
    vcWv.Cmd (fac, func(win, event) vhelp.ListErrMsgTopics (win, event););
vcCmd.SetHelp (HelpOnErrorMsgsCmd,
    "View list of error message references available through ViewHelp");
/*
 * Even though this is listed in the menu as "Help on Commands", it
 * really is just the list of all ViewHelp documentation, excluding
 * error message documentation.
 */
HelpOnCommandsCmd :=
    vcWv.Cmd (fac, func(win, event) vhelp.ListNonErrMsgTopics (win, event););
vcCmd.SetHelp (HelpOnCommandsCmd,
               "View list of command references available through ViewHelp");

HelpMenu := vcWv.MakeMenu ("Help",
                           vcWv.Pin (),
                           list ("ViewDoc",             ViewDocCmd),
                           list ("Help on Error Messages", HelpOnErrorMsgsCmd),
                           list ("Help on Commands",       HelpOnCommandsCmd),
                           list ("Track Errors from file", ReadErrorFileCmd),
                           list ("Show Bindings",       ShowBindCmd),
                           list ("Show Commands",       ShowCmds));

/*
 * This is implemented as a menu function rather than a menu string,
 * so that ProgramName can be modified by a vload loader file, via
 * the function SetProgramName.
 */
ProgramName := caddr (vscript.vsParse (car (vcWv.args)));
function QuitEntry()
    list (func() vcRegion.Create ("quit", vcLabel.fac, "Quit " + ProgramName);,
          QuitCmd);

/*\vscmuserDocpage vcWin.SetProgramName -- Set program name for quit menu & dialog box
 * 
 * SYNOPSIS
 *      vcWin.SetProgramName (vlStringVT program_name)
 *
 * DESCRIPTION
 *      An application written using vload will have a Red Square
 *      menu entry "Quit vload" and a quit dialog box that says,
 *      "OK to quit vload?"  This function will change "vload" to
 *      `program_name'.
 */
function SetProgramName(name) ProgramName = name;

RedSquareMenu := vcWv.MakeMenu ("Red Square",
                                vcWv.Pin (),
                                list ("Dismiss Window", wclose),
                                list ("Dismiss Children", DismissPopupsCmd),
                                list ("System",         system_cmd),
                                list ("Plot Setup...",  PlotCmd),
                                list ("Plot Go",        plotgo_cmd),
                                list ("Macros",         MacPanelCmd),
                                list ("ViewScript/VS",  ViewScriptCmd),
                                list ("ViewScript/SCM", SchemeCmd),
                                list ("Edit text file", ViewTextCmd),
                                list ("Customize",      CustomMenu),
                                list ("Help",           HelpMenu),
                                vcWv.Separator (),
                                QuitEntry ());

RedSquareMenuNoPlot := vcWv.MakeMenu ("Red Square",
                                vcWv.Pin (),
                                list ("Dismiss Window", wclose),
                                list ("Dismiss Children", DismissPopupsCmd),
                                list ("System",         system_cmd),
                                list ("Macros",         MacPanelCmd),
                                list ("ViewScript/VS",  ViewScriptCmd),
                                list ("ViewScript/SCM", SchemeCmd),
                                list ("Edit text file", ViewTextCmd),
                                list ("Customize",      CustomMenu),
                                list ("Help",           HelpMenu),
                                vcWv.Separator (),
                                QuitEntry ());

RedSquareMenuNoPlotNoQuit := vcWv.MakeMenu ("Red Square",
                                vcWv.Pin (),
                                list ("Dismiss Window", wclose),
                                list ("Dismiss Children", DismissPopupsCmd),
                                list ("System",         system_cmd),
                                list ("Macros",         MacPanelCmd),
                                list ("ViewScript/VS",  ViewScriptCmd),
                                list ("ViewScript/SCM", SchemeCmd),
                                list ("Edit text file", ViewTextCmd),
                                list ("Customize",      CustomMenu),
                                list ("Help",           HelpMenu),
                                vcWv.Separator (),
                                list ("Dismiss Window", wclose));

RedSquareMenuMinimal := vcWv.MakeMenu ("Red Square",
                                vcWv.Pin (),
                                list ("Dismiss Window", wclose),
                                list ("Dismiss Children", DismissPopupsCmd),
                                list ("System",         system_cmd),
                                vcWv.Separator (),
                                QuitEntry ());

red_square_size := vgSize.Create (16, 16, 16, 16, 16, 16);

RedSquareEntry  := list (func() vcRegion.Create ("RedSquare", vcColor.fac,
                     vgColor.Red, size <- red_square_size);, RedSquareMenu);

RedSquareEntryNoPlot := list (func() vcRegion.Create ("RedSquare", vcColor.fac,
                              vgColor.Red, size <- red_square_size);,
                              RedSquareMenuNoPlot);

RedSquareEntryNoPlotNoQuit := list (func() vcRegion.Create ("RedSquare",
                                vcColor.fac, vgColor.Red,
                                size <- red_square_size);,
                                RedSquareMenuNoPlotNoQuit);

RedSquareEntryMinimal  := list (func() vcRegion.Create ("RedSquare", 
                            vcColor.fac, vgColor.Red,
                            size <- red_square_size);, RedSquareMenuMinimal);
/*
 * Code to handle the pop-up command line
 */
if (!isbound (`cmdline, current_package ())) {
    cmdline         := nil;
    owner_window    := nil;
    cmdline_window  := nil;
    the_event       := nil;
}

function cmdline_win_destroyed(w) {
    vcRegion.RemoveChild (cmdline);
    cmdline_window = nil;
}

function get_cmdline()
    if (cmdline == nil) {
        cmdline = vcWv.GetCmdline (wv);
        vcCmdline.CallOnReturn (cmdline, func(ev, str) {
            vcWin.Destroy (cmdline_window);
            vcWin.RunCmdline (owner_window, str, the_event);
        });
    }


function show_cmdline(win, event) {
    ow := vcEvent.Window (event);
    the_event = event;

    if ((ow == nil) || ((ow != cmdline_window) && vcWin.HasCmdline (ow))) {
        if (cmdline_window != nil)
            vcWin.Destroy (cmdline_window);

        owner_window = ow;
        get_cmdline ();
        cmdline_window =
            vcWin.Popup (owner_window, cmdline, cmdline,
                         vcWin.NoStatus + vcWin.NoBindings + vcWin.NoMacros);
        vcWin.Forward (owner_window, cmdline_window);
        vcWin.DestroyNotify (cmdline_window, cmdline_win_destroyed);
        vcRegion.SetFocus (cmdline, cmdline);
    }
}

run_cmd := vcWv.Cmd (vcWin.fac, show_cmdline);

/*
 * Bind these in vcWv.bindings instead of vcWin.bindings because the
 * space-key should be given to facilities, if they want it, and then
 * checked in the global bindings afterward.
 */
vcBind.Set (vcWv.bindings, ' ',   run_cmd);
vcBind.Set (vcWv.bindings, '\_x', run_cmd);
vcBind.Set (vcWv.bindings, '\_X', run_cmd);

/*
 * The call to register the vcMacPanel facility has to be after a definition
 * of the vcWin.RedSquareMenu, which is used inside the vcMacPanel.vs file.
 */
vcFac.Create ("vcMacPanel", vscmVT);
