/*
 *  $Revision: 1.4.1.1 $  $Date: 1994/11/07 21:12:29 $
 *
 *  Routines for parsing up a version 4 or 5 RCS file. For more info
 *  read rcsfile(5). This should be commented more.
 *
 *  Copyright (c) 1994 by Viewlogic Systems, Inc.
 */



/* The structure which represents the entire RCS file. */
struct
rcs_text_s
{
    vlStringVT  head;
    vscmVT      symbols;
    vscmVT      locks;
    vscmVT      deltas;
};



/* The structure which represents a single version delta. */
struct
rcs_delta_s
{
    vlStringVT  version;
    vlStringVT  date;
    vlStringVT  author;
    vlStringVT  state;
    vscmVT      branches;
    /* Yeah, I ignore some other stuff. */
};



/*\intFunction rcs_alloc_text -- Allocate a new rcs_text_s struct.
 */
function rcs_alloc_text()
{
    rcs_text        := struct rcs_text_s { "", nil, nil, nil };


    /* All constants are arbitrary. */
    rcs_text.symbols = vlSym.StrCreate (71);
    rcs_text.locks   = vlSym.StrCreate (5);
    rcs_text.deltas  = vlSym.StrCreate (71);

    vlSym.Dynamic (rcs_text.symbols, 2.0, 2.0);
    vlSym.Dynamic (rcs_text.locks, 2.0, 2.0);
    vlSym.Dynamic (rcs_text.deltas, 2.0, 2.0);

    rcs_text;
}



/*\intFunction rcs_alloc_delta -- Allocate a new rcs_delta_s struct.
 */
function rcs_alloc_delta()
{
    rcs_delta := struct rcs_delta_s { "", "", "", "", nil };
}



/*\intFunction rcs_parse_admin -- Parse up the administration section.
 * 
 * SYNOPSIS
 *      rcs_parse_admin(fp, rcs)
 *          vlFileVT            fp;
 *          struct rcs_text_s   rcs;
 *
 * DESCRIPTION
 *      This routine parses up the admin section of an RCS file. For our
 *      purposes, the administration section consists only of those fields
 *      which are required to be present by RCS. These fields are the head,
 *      the symbols list, and the locks list. The remainder of the admin
 *      section is ignored.
 */
function rcs_parse_admin(fp, rcs)
{
    declare line;
    declare done_with_admin;

    done_with_admin = FALSE;

    while (!null (line = fgetline (fp)) && !done_with_admin)
        if (string_search (line, "head") == 0)
            rcs_parse_head (line, rcs);

        else if (string_search (line, "symbols") == 0)
            rcs_parse_symbols (line, rcs, fp);

        else if (string_search (line, "locks") == 0)
        {
            rcs_parse_locks (line, rcs, fp);
            done_with_admin = TRUE;
        }
}



/*\srcFunction rcs_parse_head -- Parse up the head.
 */
function rcs_parse_head(line, rcs)
{
    declare token;

    line = string_token (line, &token, " \t");
    line = string_token (line, &token, " \t\n;");

    rcs.head = token;
}



/*\srcFunction rcs_parse_symbols -- Parse up the symbols list.
 *
 * DESCRIPTION
 *      The symbol list has the following syntax:
 *
 *          symbols    {id : num}*;
 */
function rcs_parse_symbols(line, rcs, fp)
{
    declare token;
    declare symbol;
    declare version;

    done       := FALSE;
    str        := "";

    while (!done)
    {
        str  = str + line;
        done = rindex (line, ';');
        if (!done)
            line = fgetline (fp);
    }

    str = substring (str, 8, 0);
    while (!null (str = string_token (str, &symbol, " \t\n:;")))
    {
        str = string_token (str, &version, " \t\n:;");
        vlSym.Store (rcs.symbols, symbol, version);
    }
}        



/*\srcFunction rcs_parse_locks -- Parse up the locks list.
 *
 * DESCRIPTION
 *      The lock list has the following syntax:
 *
 *          locks      {id : num}*;  {strict  ;}
 */
function rcs_parse_locks(line, rcs, fp)
{
    declare user;
    declare version;
    declare index;

    done       := FALSE;
    str        := "";

    while (!done)
    {
        str  = str + line;
        done = rindex (line, ';');
        if (!done)
            line = fgetline (fp);
    }

    str = substring (str, 5, 0);

    index = string_search (str, "strict;");
    if (!null (index))
        str = substring (str, 0, index);

    while (!null (str = string_token (str, &user, " \t\n:;")))
    {
        str = string_token (str, &version, " \t\n:;");
        vlSym.Store (rcs.locks, user, version);
    }

}



/*\intFunction rcs_parse_delta -- Parse up the deltas.
 * 
 * SYNOPSIS
 *      rcs_parse_delta(fp, rcs)
 *          vlFileVT            fp;
 *          struct rcs_text_s   rcs;
 *
 * DESCRIPTION
 *      This routine parses up all the delta sections found in an RCS file.
 *      It allocates new delta records and adds them to the rcs structure
 *      which is being built.
 */
function rcs_parse_delta(fp, rcs)
{
    declare line;
    declare token;
    declare i;


    first_time := TRUE;


    while (!null (line = fgetline (fp)))
    {
        /* The first time we get here, there may be some junk left over
         * from the admin section which we don't want. So, we just throw
         * away everything until we encounter a numeric digit in the first
         * column which indicates that the delta section has really begun.
         */
        while (first_time)
        {
            i = char2int (line[0]);
            if ((i >= 48) && (i <= 57))
                first_time = FALSE;
            else
            {
                if (null (line = fgetline (fp)))
                    return nil;
            }
        }


        i = char2int (line[0]);
        if ((i >= 48) && (i <= 57))
        {
            rcs_delta := rcs_alloc_delta ();

            rcs_delta.version = string_trim (line);
            /* printf ("version: %s\n", rcs_delta.version); */

            line = fgetline (fp);

            line = substring (line, 4, 0);
            line = string_token (line, &token, " \t;");            
            rcs_delta.date = token;
            /*printf ("date: %s\n", rcs_delta.date);*/

            line = string_token (line, &token, " \t;");
            line = string_token (line, &token, " \t;");            
            rcs_delta.author = token;
            /* printf ("author: %s\n", rcs_delta.author); */

            line = string_token (line, &token, " \t;");
            line = string_token (line, &token, " \t;");            
            rcs_delta.state = token;
            /* printf ("state: %s\n", rcs_delta.state); */

            line = fgetline (fp);

        /* !!! What's this? Its branch stuff that I can't get to work:
         *
         *          WE`LL PUNT ...
         *
         *     ... because code freeze happens in three hours. And I
         *     don't think I need it anyway. Well, not right now at least.
         */    
/*
            line = substring (line, 8, 0);
            line = string_trim (line);
            while (!null (line = string_token (line, &token, " \t\n;")))
                rcs_delta.branches = cons (token, rcs_delta.branches);

*/
            vlSym.Store (rcs.deltas, rcs_delta.version, rcs_delta);
        }

        /* The delta section is over. */
        if (string_search (line, "desc") == 0)
            break;
    }
}



/*\Function rcs_parse -- Parse the specified RCS file.
 * 
 * SYNOPSIS
 *      struct rcs_file_s
 *      rcs_parse(filename)
 *          vlStringVT      filename;
 *          
 * DESCRIPTION
 *      This routine opens the file specified by ``filename'' and parses
 *      it and loads some of the interesting info from this structure into
 *      the structure which it allocates.
 *
 * RETURNS
 *      A data structure containing interesting bits from a parsed up RCS
 *      file.
 */
function rcs_parse(filename)
{
    fp  := fopen (filename, "r");
    rcs := rcs_alloc_text();

    rcs_parse_admin(fp, rcs);
    rcs_parse_delta(fp, rcs);

    fclose (fp);
    rcs;
}




/*\intFunction rcs_rcstime_to_longtime -- Convert RCS time format to long.
 * 
 * DESCRIPTION
 *      RCS stores its date in the following format:
 *
 *          yy.mm.dd.h.m.s
 *
 *      This routine converts it from this format to the standard long format
 *      used by most time based routines. This routine does not perform any
 *      adjustment based on time zones.
 */
function
rcs_rcstime_to_longtime(rcstime)
{
    declare longtime;

    yy  := string2number ("19" + substring (rcstime, 0, 2));
    mm  := string2number (substring (rcstime, 3, 5));
    dd  := string2number (substring (rcstime, 6, 8));
    h   := string2number (substring (rcstime, 9, 11));
    m   := string2number (substring (rcstime, 12, 14));
    s   := string2number (substring (rcstime, 15, 17));

    longtime = 0;

    /* Add the years. */
    for (i := 1970; i < yy; ++i)
    {
        if (rcs_leapyear_p (i))
            dd += 366;
        else
            dd += 365;
    }

    /* Add the months. */
    for (i = 0; i < (mm - 1); ++i)
        dd += rcs_rcstime_months[i];

    /* Add the days, hours, minutes, seconds. */
    longtime += ((dd - 1) * 86400) + (h * 3600) + (m * 60) + s;

    longtime;
}


rcs_rcstime_months := [ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ];



/*\intFunction rcs_leapyear_p -- Is the given year a leapyear?
 * 
 * DESCRIPTION
 *      Return TRUE ``year'' is a leapyear.
 */
function
rcs_leapyear_p(year)
{
    (((((year) % 4) == 0) && ((year) % 100 != 0)) || (((year) % 400) == 0));
}


