/*
 *  $Revision: 1.23.1.3 $  $Date: 1994/11/23 19:03:12 $
 *
 *  VDM encapsulation file for the file-based storage oparations.
 *
 *
 *  Note on the endProc function
 *  ----------------------------
 *      Because I/O integrators can use asynchronous functions in their
 *      implementations of the SM functions, VDM cannot presume that when a
 *      _vdm[...] method returns, the SM function has been completed.  You
 *      must call the provided endProc function to signal to the VDM backplane
 *      that the SM function has indeed terminated.
 *
 *      In addition, the SM backplane has to be completely reentrant and allow
 *      many requests from different applications to be executed at the same
 *      time without tramping over each other's data.  To this end, each
 *      method is passed in an `endArg' which uniquely identifies the thread
 *      of execution for this particular method invocation.  When a method
 *      terminates and wants to notify VDM, you must call the endProc function
 *      with two arguments: the data object and the endArg:
 *
 *          endProc (obj, endArg, user_arg);
 *
 *      If you don't call the endProc function, VDM will keep waiting for this
 *      event to occur and will eventually time out.
 *
 *  Returning a value from a _vdm method
 *  ------------------------------------
 *      Note that the endProc function takes three arguments.  The `user_arg'
 *      argument allows the integrator to pass back a return value from the
 *      integration method.  Remember, the _vdm[...] methods return
 *      immediately, but that doesn't mean that the method has finished
 *      execution.  That's why any actual return value from a _vdm function
 *      will be ignored.  But the, what if you really want to return
 *      something?  Well, it's not as straightforward, but it works.  What you
 *      do, is you specify your return value as the third argument to the
 *      endProc function.  That's the `user_arg' argument.  That argument will
 *      be converted into an attribute, which will be added to the data
 *      object.  The name of the attribute is "_vdmReturnValue".  Because
 *      attributes are generic, no restrictions are placed on the type of the
 *      "return" value.
 *
 *  Example of a _vdm method return value
 *  -------------------------------------
 *      Suppose that for a given SM encapsulation, the integrator wants the
 *      _vdmRegister method to return the name of the data object as it is
 *      stored by the I/O.  We are assuming that the name of the
 *      data object and the name of its representation in the SM are different
 *      (like RCS, which creates ,v files).
 *
 *          function _vdmRegister(obj, endProc, endArg, smCntx)
 *          {
 *              // Do the registration
 *              ...
 *              // Get the new name
 *              new_name := some_extract_func (vdm.ObjGetPhysicalName (obj)); 
 *
 *              // Signal the end of the method and return the new name.
 *              endProc (obj, endArg, new_name);
 *          }
 *
 *
 *          Now, in your application, the code would look something like this:
 *
 *          foo := vdm.ObjCreate ("YourObjName", "YourClass", location,
 *                                your_vdm_cntx, nil);
 *          vdm.ObjRegister (foo);
 *          new_foo_name := vdm.ObjGetAttr (foo, "_vdmReturnValue");
 *
 *
 *      You might wander why all these complications, but the answer is quite
 *      obvious once you consider the fact that VDM is asynchrounous and that
 *      the endProc function might be called long after the method's function
 *      has returned.  It's the user that must tell VDM when the backplane
 *      methods terminate and when that happens, there's very little context
 *      information left to work with.  The `obj' and `endArg' arguments help
 *      to alleviate this problem and allow VDM to support multiple execution
 *      threads at the same time.
 *
 *  What's a storage manager context?
 *  ---------------------------------
 *      Some storage manager want to communicate state information between
 *      methods.  The _vdmCreate method returns as its value the storage
 *      manager context.  It's whatever you wish it to be.  That value is
 *      passed in to all the methods.  You make the use of it: VDM serves only
 *      as a conduit to pass it along to all the methods.
 *
 *  Currently supported methods
 *  ---------------------------
 *      _vdmInit                    - general
 *      _vdmCreate                  - general
 *      _vdmGeneric                 - general
 *
 *      _vdmCheckin                 - SM
 *      _vdmCheckoutLock            - SM
 *      _vdmCheckoutNoLock          - SM
 *      _vdmDelete                  - SM
 *      _vdmFind                    - SM
 *      _vdmGetPhysicalDescriptor   - SM
 *      _vdmLock                    - SM
 *      _vdmLoop                    - SM
 *      _vdmMetaDataName            - SM [optional]
 *      _vdmQuery                   - SM
 *      _vdmRegister                - SM
 *      _vdmStatus                  - SM
 *      _vdmUnlock                  - SM
 *
 *      _vdmOpen                    - I/O
 *      _vdmClose                   - I/O
 *      _vdmRead                    - I/O
 *      _vdmReadChar                - I/O
 *      _vdmReadLine                - I/O
 *      _vdmWrite                   - I/O
 *      _vdmWriteChar               - I/O
 *      _vdmWriteLine               - I/O
 *
 *  Copyright (c) 1994 by Viewlogic Systems, Inc.
 */



/*\Docpage _vdmInit -- Initialize SM encapsulation.
 * 
 * DESCRIPTION
 *      This function is called only once by the VDM.  It's purpose is to
 *      initialize this particular storage manager encapsulation.  Note that a
 *      second initialization method in this file, _vdmCreate, is used to bind
 *      the storage manager encapsulation to a particular location.
 *      
 *      In addition, this function must return a structure which will be used
 *      to prompt and to parse for the location's string.  For more
 *      information about the ViewScript structures, see struct(man).
 *
 * NOTE
 *      The difference between this method and the _vdmCreate method is that
 *      the _vdmInit method is called only once per storage encapsulation.
 *      The _vdmCreate method is called for each location managed by this
 *      storage manager.
 *
 * SEE ALSO
 *      struct
 */
function _vdmInit(obj, endProc, endArg, ignore)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmInit.\n"));

    endProc (obj, endArg, struct vdmFileVT {vlStringVT pathname});
}



/*\Docpage _vdmCreate -- Instantiate the SM encapsulation.
 *
 * DESCRIPTION
 *      The _vdmCreate method will be called when the VDM requests the first
 *      storage operation performed on a data object in the given location.
 *      Any SM request can trigger instantiation.  The instantiation of the
 *      storage manager encapsulation is postponed until the VDM actually
 *      wants to use it in order not to slow down the initialization of
 *      applications built on top of the VDM.
 *
 *      This method returns the storage manager's context.  The context
 *      represents the particular instantiation of this storage manager
 *      encapsulation for a particular location.  The same storage manager can
 *      be used to manage another location, but VDM will create two separate
 *      instances of the storage manager encapsulation for the two lcoations.
 *
 * NOTE
 *      The difference between this method and the _vdmInit method is that the
 *      _vdmInit method is called only once per storage encapsulation.  The
 *      _vdmCreate method is called for each location managed by the given
 *      storage manager.
 *
 * CAUTION
 *      The object passed in to this function is a generic data object managed
 *      by the VDM library.  It has no relation whatsoever to any design data
 *      objects and it is used only by VDM.
 */
function _vdmCreate(obj, endProc, endArg, ignore, struct_type, location_string)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmCreate: %s\n", location_string));

    new_string := vscript.vsAbsPath (string_trim (location_string));

    if (location_string == nil)
        vlEx.Throw2 (vlEx.FAILURE, "vdmFile", vlErrSev.Error,
                     sprintf ("Illegal pathname specification--`%s'",
                              location_string));

    /*
     * The context for the File encapsulation is the pathname where the files
     * reside.
     */
    endProc (obj, endArg, new_string);
}



/*\Docpage _vdmCheckoutLock -- Checkout with a lock.
 *
 * DESCRIPTION
 *      Data object is checked out from the storage manager and a "lock" is
 *      requested.
 *
 * NOTE
 *      The term "lock" is understood by VDM as an exclusive right to modify
 *      the data object without any other user being able to _vdmCheckin any
 *      modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmCheckoutLock(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmCheckoutLock.\n");

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmCheckoutNoLock -- Check out without a lock.
 * 
 * DESCRIPTION
 *      Retrieve a revision of the data object from the storage manager, but
 *      don't lock it.  This operation is commonly referred to as a "fetch".
 *
 *      With no other information provided by the user, the checked out
 *      revision should be the latest one.
 *
 *      Because there's no lock requested, other users can checkin new
 *      versions of this data object.  For this reason, the "fetch" operation
 *      should only be done for read-only purposes.  If the user intends to
 *      modify the object, a lock should be requested via the _vdmCheckoutLock
 *      method.
 *
 * NOTE
 *      The term "lock" is understood by VDM as an exclusive right to modify
 *      the data object without any other user being able to _vdmCheckin any
 *      modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmCheckoutNoLock(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmCheckoutNoLock.\n");

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmCheckin -- Check in a data object.
 *
 * DESCRIPTION
 *      Store a new revision of the data object in the storage manager.  If
 *      the user had a lock on the object, it should be relinquished.
 *
 * NOTE
 *      The term "lock" is understood by VDM as an exclusive right to modify
 *      the data object without any other user being able to _vdmCheckin any
 *      modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmCheckin(obj, endProc, endArg, FileCntx, log_message)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmCheckin.\n");

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmLock -- Get a lock for the object.
 *
 * DESCRIPTION
 *      The storage manager is requested to establish a lock for the data
 *      object for this user.  No other user is then allowed to check out this
 *      object with a lock.
 *
 * NOTE
 *      The term "lock" is understood by VDM as an exclusive right to modify
 *      the data object without any other user being able to _vdmCheckin any
 *      modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmLock(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmLock.\n");

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmUnlock -- Cancel a lock.
 *
 * DESCRIPTION
 *      Relinquish the lock on the data object.
 *
 * NOTE
 *      The term "lock" is understood by VDM as an exclusive right to modify
 *      the data object without any other user being able to _vdmCheckin any
 *      modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmUnlock(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmUnlock.\n");

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmRegister -- Register a new data object.
 *
 * DESCRIPTION
 *      Tell the storage manager about a brand new data object that is being
 *      added under its care and management.  Note that this operation is
 *      different from checking in, which creates a new revision for an
 *      already existing design object.
 *
 * NOTE
 *      Some storage managers don't require that you specify an initialization
 *      message.  Just ignore it in that case.
 */
function _vdmRegister(obj, endProc, endArg, FileCntx, init_message)
{
    name     := vdm.ObjGetPhysicalName (obj);
    pathname := vscript.vsMerge ("", FileCntx, name, "");

    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmRegister: %s\n", pathname));

    fclose (fopen (pathname, "w"));

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmDelete -- Delete the data object.
 *
 * DESCRIPTION
 *      Delete the data object from the storage manager.  No assumptions are
 *      made whether the delete operation is permanent or can be undone.
 */
function _vdmDelete(obj, endProc, endArg, FileCntx)
{
    cntx     := vdm.ObjGetCntx (obj);
    name     := vdm.ObjGetPhysicalName (obj);
    pathname := vscript.vsMerge ("", FileCntx, name, "");

    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmDelete: %s\n", pathname));

    if (! vscript.vsDelete (pathname, vsAttr.File))
    {
        msg := sprintf ("Unable to delete the %s file.", pathname);
        vlErr.Out2 ("vdmFile", vlErrSev.Error, msg);

        vdm.CntxCallErrorHandlers (cntx, msg);
    }

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmQuery -- Get information on the data object.
 *
 * DESCRIPTION
 *      This method returns a list of object informations.  The list has the
 *      following format:
 *
 *          ("log" (("user1" . "rev1") ("user2" . "rev32"))
 *                 ("version1" "version2"))
 *
 *      The first element is a string representing the log.  More on the log
 *      below.
 *
 *      The second element is a list of the lock information.  Some storage
 *      managers that allow multiple users to have locks on multiple branches
 *      will find this format useful.  Other storage managers will return a
 *      single element list.  The lock information is a list of pairs: each
 *      pair is composed of the user name and the revision.  The pair contains
 *      information about a user which holds a lock for a particular revision.
 *
 *      The third element is a list of all the versions of this object.
 *
 * CAUTION
 *      Gathering the log information might be an expensive operation.  For
 *      this reason, this method allows the user to specify whether to get the
 *      log information or not.  The `get_log' argument will tell you that.
 *
 *      Secondly, because of the potential amount of the log information, the
 *      user might not choose to have it returned as a string, but will pass
 *      in a function pointer to be used to display the log information.  If
 *      the `log_proc' argument is not nil, you must pass the log information
 *      to this function.  It's declared to take a single argument, a string,
 *      and can be called multiple times.
 *
 *      The user will know when the log information is over, when this
 *      function returns.
 */
function _vdmQuery(obj, endProc, endArg, FileCntx, get_log, log_proc)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmQuery.\n");

    msg := "";

    if (get_log)
    {
        declare     mm, dd, yy, h, m, s;
        name     := vdm.ObjGetPhysicalName (obj);
        pathname := vscript.vsMerge ("", FileCntx, name, "");
        info     := vscript.vsInfo (pathname);

        if (info != nil)
        {
            vscript.vlCvtTime (car (info), &mm, &dd, &yy, &h, &m, &s);
            msg  = sprintf ("Timestamp: %d/%d/%d\n", mm, dd, yy);
            info = cdr (info);
            msg  = msg + sprintf ("Byte size: %d\n", car (info));
            info = cdr (info);

            if (car (info) == vsAttr.File)
                msg = msg + sprintf ("%s is a file.", pathname);
            else
                msg = msg + sprintf (" %s is a directory.", pathname);

            if (log_proc != nil)
                log_proc (msg);
        }
    }
            
    endProc (obj, endArg, list (msg, nil, nil));
}



/*\Docpage _vdmFind -- Find the object.
 *
 * DESCRIPTION
 *      Is this object under the control of this storage manager in the given
 *      location?
 */
function _vdmFind(obj, endProc, endArg, FileCntx)
{
    pathname := vscript.vsMerge ("", FileCntx, vdm.ObjGetPhysicalName (obj),
                                 "");

    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmFind: %s.\n", pathname));

    endProc (obj, endArg, fexists (pathname));
}



/*\Docpage _vdmGetPhysicalDescriptor -- Get the physical descriptor.
 * 
 * DESCRIPTION
 *      Each persistent data object is stored somewhere on disk, and thus has
 *      a physical descriptor.  The form of the descriptor depends on the
 *      storage manager.  For file-based storage managers, it might be a
 *      filename,  For databases, it might be a combination of tokens
 *      describing the name and the location of the database plus some
 *      information about where the given data is located in that database.
 */
function _vdmGetPhysicalDescriptor(obj, endProc, endArg, FileCntx)
{
    pathname := vscript.vsMerge ("", FileCntx, vdm.ObjGetPhysicalName (obj),
                                 "");

    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmGPD: %s\n", pathname));

    endProc (obj, endArg, pathname);
}



/*\Docpage _vdmLoop -- Build a list of objects for a VDM loop.
 * 
 * DESCRIPTION
 *      Returns a list of data objects of the given class in this location.
 *      If there are no objects of that type, it returns nil.  This function
 *      is called by the vdm.ClassLoopObj function.
 */
function _vdmLoop(obj, endProc, endArg, FileCntx, class, loc, cntx, usr_filter)
{
    file_template   := vdm.ClassBuildFileTemplate (class);
    regexp_template := vdm.ClassBuildRegexpTemplate (class, usr_filter);
    path            := vscript.vsMerge ("", FileCntx, file_template, "");
    path_length     := string_length (FileCntx) + 1;
    regexp          := vscript.re_compile (regexp_template);
    obj_list        := nil;
    matched_attrs   := nil;
    match_str       := nil;

    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmLoop: %s\n", path));

    /*
     * The algorithm for the file encapsulation is as follows:
     *
     * 1. Substitute all the attribute references in the template with '*'.
     * 2. Append the "wildcarded" template to the location's path.
     * 3. Loop through all the files that match the template.  Then apply the
     *    regular expression template to narrow down the match.
     * 4. If the regular expression template matches, extract the attribute
     *    values and build an object handle.  Then, add it to the object list.
     */
    foreach (pathname; vscript.TraverseDir (path, vsAttr.Both))
    {
        match_str = substring (pathname, path_length, 0);
        if (vscript.re_exec (regexp, match_str))
        {
            matched_attrs = vdm.ClassGetMatchedAttributes (class, regexp,
                                                           match_str);
            obj_list = cons (vdm.ObjCreate (match_str, vdm.VersLatest, class,
                                            loc, cntx, matched_attrs),
                             obj_list);
        }
    }

    endProc (obj, endArg, obj_list);
}



/*\Docpage _vdmStatus -- What's the storage status of this object.
 * 
 * DESCRIPTION
 *      The sole purpose of this method is to determine the status of the
 *      given object with respect to the storage manager.  This method must
 *      return two values, representing the possible storage manager state
 *      (vdm.SMStatusVT) and the possible working area state (vdm.WorkAreaVT).
 *
 *      If the object is under the storage manager's control, the object's
 *      status can be one of the following:
 *
 *          vdm.SMStatus.Unlocked       - object is not locked
 *          vdm.SMStatus.LockedByYou    - object is locked by you
 *          vdm.SMStatus.LockedByOther  - object is locked by another user
 *          vdm.SMStatus.Uncontrolled   - object is not yet in SM
 *
 *      If the storage manager allows the data object to persist outside of
 *      the storage manager's storage area, then such an object is said to
 *      reside in the user's working area.  Its status can then be one of the
 *      following:
 *
 *          vdm.WorkStatus.New       - object is brand new.
 *          vdm.WorkStatus.Modified  - object is modified
 *          vdm.WorkStatus.Outdated  - object is not the same as object in SM
 *          vdm.WorkStatus.UpToDate  - object is the same as that in SM
 *          vdm.WorkStatus.Empty     - object is not in the work area
 *
 * SEE ALSO
 *      vdm.SMStatusVT, vdm.WorkStatusVT
 */
function _vdmStatus(obj, endProc, endArg, FileCntx)
{
    name     := vdm.ObjGetPhysicalName (obj);
    pathname := vscript.vsMerge ("", FileCntx, name, "");

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmFile _vdmStatus: %s\n", pathname));

    if (null (info := vscript.vsInfo (pathname)))
        endProc (obj, endArg, list (vdm.WorkStatus.Empty,
                                    vdm.SMStatus.Uncontrolled));
    else
    {
        /*
         * Although a directory can be considered a compound object, there's
         * no way to expand it without making it into a new VDM location.  So,
         * just for kicks, a temporary parameter is added to indicate whether
         * the object is a director or not.
         */
        if (caddr (info) == vsAttr.File)
            vdm.ObjSetAttr (obj, "DIRECTORY", "no", FALSE);
        else
            vdm.ObjSetAttr (obj, "DIRECTORY", "yes", FALSE);

        endProc (obj, endArg, list (vdm.WorkStatus.UpToDate,
                                    vdm.SMStatus.Unlocked));
    }
}



/*\Docpage _vdmMetaDataName -- Create a data name for a meta-data object.
 * 
 * DESCRIPTION
 *      This is an optional method.  It should be provided only if the given
 *      storage manager does not handle the meta-data itself.
 *
 *      The sole purpose of this method is to return a string that will
 *      represent the data name of the meta-data object.  If the storage
 *      manager does not handle meta-data, VDM will create a meta-data object
 *      which shadows the data object.  But, given a data object, VDM has to
 *      be able to find its corresponding meta-data object.  That's where this
 *      method enters the picture.
 *
 *      The object passed into this method is the data object.  Use its data
 *      name to create a unique meta-data name.  For file-based storage
 *      managers, it's recommended that you prepend the name of a
 *      subdirectory, like `vdmmeta' to the data object's name and return it.
 *      In this way, the name of the meta-data object is the same as that of
 *      the data object's that it's supposed to shadow, but in a special
 *      subdirectory.  You can make that subdirectory anything you want, make
 *      it visible or not to the user, just as long as it's valid for this
 *      storage manager.
 */
function _vdmMetaDataName(obj, endProc, endArg, rcsCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmMetaDataName.\n");

    endProc (obj, endArg, vscript.vsMerge ("", "vdmmeta",
                                           vdm.ObjGetPhysicalName (obj), ""));
}



/****************************************************************************/
/*                              I/O methods                                 */
/****************************************************************************/


/*
 * The file I/O encapsulation needs to hang the file pointer somewhere: the
 * object is the most obvious and convenient object.  The file pointer will be
 * stored as the value of the _vdmFp attribute.
 */
FILE_ATTR_STR := "_vdmFp";



/*\Docpage _vdmOpen -- Open a design object for I/O operations.
 *
 * DESCRIPTION
 *      Opens the the object for a particular mode of I/O operations.  If the
 *      object has already been opened and not closed, an exception will be
 *      raised.
 */
function _vdmOpen(obj, endProc, endArg, FileCntx, mode)
{
    path      := vscript.vsMerge ("", FileCntx, vdm.ObjGetPhysicalName (obj),
                                  "");
    fp        := nil;
    file_mode := nil;

    if (vdmDebug)
        debug_out (obj,
                   sprintf ("File _vdmOpen: %s, mode: %s.\n", path, mode));

    /*
     * Verify that the _vdmFp attribute is not already taken.  If so, somebody
     * is trying to reopen this object.
     */
    if (! (null (vdm.ObjGetAttr (obj, FILE_ATTR_STR))))
        vlEx.Throw2
            (vlEx.FAILURE, "vdmFile", vlErrSev.Err,
             sprintf
             ("Attempted to re-open %s when it hasn't been closed yet.",
              vdm.ObjGetName (obj)));

    
    if (mode == vdm.IOMode.Read)
        file_mode = "r";
    else
    {
        /*
         * Make sure any sub-directories exist when we are performing "write"
         * operations.  Ideally, we would wait for an operation to fail before
         * doing this ... but that's more involved.
         */
        <device, subdir, filename> := vscript.vsParse2 (path);

        vscript.vsCreateWithPath (subdir, vsAttr.Dir);
    
        if (mode == vdm.IOMode.Write)
            file_mode = "w";
        else if (mode == vdm.IOMode.Append)
            file_mode = "a";
        else
            file_mode = "r+";
    }

    fp = fopen (path, file_mode);
    vdm.ObjSetAttr (obj, FILE_ATTR_STR, fp, FALSE);

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmClose -- Close a design object for I/O operations.
 *
 * DESCRIPTION
 *      If the object has been closed already, just ignore it.  Otherwise,
 *      close the object in such a way that any subsequent I/O request will
 *      fail, with the exception of open.
 */
function _vdmClose(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
    {
        path := vscript.vsMerge ("", FileCntx, vdm.ObjGetPhysicalName (obj),
                                 "");
        debug_out (obj, sprintf ("File _vdmClose: %s.\n", path));
    }

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    /*
     * If the object has already been closed, ignore it.  Otherwise, close the
     * file pointer and set the attribute value to nil.
     */
    if (! null (fp))
    {
        fclose (fp);
        vdm.ObjSetAttr (obj, FILE_ATTR_STR, nil, FALSE);
    }

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmReadChar -- Read a single character.
 */
function _vdmReadChar(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmReadChar.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    endProc (obj, endArg, fgetchar (fp));
}



/*\Docpage _vdmReadLiine -- Read a single line from the design.
 */
function _vdmReadLine(obj, endProc, endArg, FileCntx)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmReadLine.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    endProc (obj, endArg, fgetline (fp));
}



/*\Docpage _vdmRead -- Read a number of characters from a design.
 */
function _vdmRead(obj, endProc, endArg, FileCntx, byte_no)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmRead.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    /*
     * This is a horribly inefficient implementation, but we can't do anything
     * better in ViewScript right now: it has no concept of reading a block of
     * the specified size.
     */
    str := "";
    ch  := fgetchar (fp);

    for (i := 0; ((i < byte_no) && (ch != nil)); ++i)
    {
        str = str + string_var (ch);
        ch  = fgetchar (fp);
    }

    endProc (obj, endArg, str);
}



/*\Docpage _vdmWriteChar -- Write a single character into a design object.
 */
function _vdmWriteChar(obj, endProc, endArg, FileCntx, char)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmWriteChar.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    fputchar (fp, char);

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmWriteChar -- Write a single line into a design object.
 */
function _vdmWriteLine(obj, endProc, endArg, FileCntx, line)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmWriteLine.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    fputline (fp, line);

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmWrite -- Write a number of character into a design object.
 */
function _vdmWrite(obj, endProc, endArg, FileCntx, buf, byte_no)
{
    if (vdmDebug)
        debug_out (obj, "File _vdmWrite.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    fputline (fp, buf);

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmGeneric -- Generic method.
 *
 * DESCRIPTION
 *      The VDM backplane has a fixed number of methods.  The generic method
 *      is provided as a means of extending the backplane to allow custom
 *      integrations that require more storage management capabilities than
 *      the predefined methods allow.  This function will be called with a
 *      keyword, which should serve as a dispatch key to call other functions.
 *      This method will be called as a result of a call to the
 *      vdm.ObjCallSmGeneric(man) API function.  That function accepts the
 *      dispatch key, represented as a string, and an addition argument of any
 *      type.
 *
 * EXAMPLE
 *      arg := nil;
 *      vdm.ObjCallGeneric (obj, "action1", arg);
 */
function _vdmGeneric(obj, endProc, endArg, storage_mgr_cntx, dispatch_key, arg)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("File _vdmGeneric: %s.\n", dispatch_key));

    endProc (obj, endArg, nil);
}



/*\Docpage debug_out -- Handy output function.
 * 
 * DESCRIPTION
 *      Takes an object and a string and if the vdmDebug variable is set, it
 *      calls the VDM's context output handlers with the string.
 *
 *      When running ViewData, the output will show up in the session log
 *      window.
 */
function debug_out(obj, str)
    vdm.CntxCallOutputHandlers (vdm.ObjGetCntx (obj), str);
