/*
 *  $Revision: 1.61.1.1 $   $Date    : 94/02/25 01:16:28 $
 *
 *  Viewscript UI definitions for the Core vcPref facility.
 *
 *  Copyright (c) 1992,1994 by Viewlogic Systems, Inc.
 */

/*\vscmuserSection vcPref. -- Preference Editor

DESCRIPTION
     The preference editor provides a graphical interface to
     manage user options.  It can be instantiated as a region
     in another window, or it can be created in its own window.
     When in its own window, menu picks are available for users
     to apply their changes temporarily, install the changes
     into the appropriate ini file, etc.

     The filename used for permanently installing preferences is,
     by default, "powerview.ini".  It will look for this file
     in the WDIR path, and write it into the current directory
     if it fails to find one.  Facilities can override the filename
     by setting a variable "IniFileName" in their Viewscript
     environment.  For example:
            IniFileName := "gtest.ini"
     This means that all updated preferences will go into that
     file.

     When instantiating the preference editor as a region inside
     another window, the menu buttons are not visible.  Typically,
     the programmer will access the edited value of the preferences
     using vcRegion.Get on the vcPref region.

FACILITY DATA
     list (window, facility, preference_id)

PREFERENCES
     all_prefs          -- Show the programmer preferences as well as the user
                           preferences.  By default only user prefs are shown
     scroll_threshold   -- Number of options below which we don't add a
                           scrollbar
     show_default       -- Show the default value of each preference
                           (goes in status when selected if FALSE)
     show_help          -- Show the help string for each preference
                           (goes in status when selected if FALSE)
     n_visible_options  -- Number of visible options to show by default
     width              -- Pixel width of pref editor.  0 means expand
                           to ideal size, non-zero means use a horizontal
                           scrollbar if necessary
*/


function fac_has_pref(fac)
{
    foreach (name, help, type, val, deflt, changed;
             vlPref.UserOptions ((vcFac.Package (fac)).prefs))
        return (TRUE);
    FALSE;
}


package vcFacSelector(_Create, _Get)
{
    function build_facility_list()
    {
        foreach (fac; vcFac.Loop ())
            if (fac_has_pref (fac))
                fac_list = cons (fac, fac_list);
        fac_list = reverse_destructive (fac_list);
        n_facs = length (fac_list);
    }


    fac_list := nil;
    n_facs   := 0;


    function _Create(deflt, region)
    {
        if (fac_list == nil)
        {
            build_facility_list ();
            build_facility_list = nil;
        }

        p := vcRegion.Create ("", vcPicklist.fac, nil,
                              columns   <- 5,
                              rows      <- ceil ((n_facs + 4) / 5),
                              exclusive <- TRUE);

        i := 0;
        foreach (fac; vscript.list_iter (fac_list))
        {
            vcPicklist.Add (p, vcFac.Title (fac), fac);

            if (fac == deflt)
                vcPicklist.SetPick (p, i);
            ++i;
        }

        vcRegion.AddChild (region, p);
        p;
    }


    function _Get(p, region, ts)
        vcRegion.Get (p, ts);
}

/* If that facility type is RegionVT, we end up not freeing it */
vcFac.CreateFromPackage ("vcFacSelector", vscmVT, vcFacSelector);


function do_facility(p, e, the_fac)
{
    if (p.modified == nil)
    {
        set_fac (p, the_fac, (vcFac.Package (the_fac)).prefs);
        vcWin.SetTitle (vcRegion.Window (p.table),
                        "Preference Editor: " + vcFac.Title (p.fac));
    }
    else
        CreateWindow (p.window, the_fac);
}


fac_cmd := vcWv.Cmd (fac, do_facility, "facility", "Set Facility",
                     vcWv.Param ("", vcFacilityVT, vcWv.Dynamic,
                                 func(p) p.fac;, vcFacSelector.fac));

/*
 * Enable/disable commands at different times throughout execution.
 */
function enable_cmds(p, window)
{
    if (window == nil)
        window = vcRegion.Window (p.region);

    vcCmd.Enable (apply_cmd,        p.enable_apply);
    vcCmd.Enable (apply_all_cmd,    p.enable_apply_all);
    vcCmd.Enable (install_cmd,      p.enable_install);
    vcCmd.Enable (defaults_cmd,     p.enable_defaults);
    vcCmd.Enable (restore_cmd,      p.enable_restore);

    if (window != nil)
        vcWin.UpdateMenus (window);
}


function EnableCmds(pref_region, window)
    enable_cmds (vcRegion.GetData (pref_region), window);

/*
 * Every time we get FocusIn, update the command list.  This should
 * happen here and not _Initial because if the there's more than once
 * preference window up, they each may require a different set of
 * enabled commands.
 */
function _FocusIn(p, event)
    enable_cmds (p, (event == nil) ? nil : vcEvent.Window (event));

_Initial := _FocusIn;

                           
function make_pref_list(p, modified) {
    pref_list := nil;
    foreach (name; vscript.list_iter (modified))
        pref_list = cons (cons (string2symbol (name),
                                vlPref.Val (p.prefs, name)),
                          pref_list);
    pref_list;
} /* function make_pref_list */

function get_editor(p, option_num)
    vcTable.Get (p.table, option_num, p.editor_column);

function get_editor_value(p, option_num) {
    ;
} /* function get_editor_value */

function set_editor_value(p, option_num, val) {
    ;
} /* function set_editor_value */

function do_apply(p, event)
{
    option_num        := 0;
    something_changed := FALSE;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
    {
        if (is_modified (p, name))
        {
            something_changed = TRUE;
            vlPref.Setval (p.prefs, name,
                           vcRegion.Get (get_editor (p, option_num), type));
        }
        else if (changed)
            something_changed = TRUE;

        ++option_num;
    }

    p.enable_apply    = FALSE;
    p.enable_restore  = FALSE;
    p.enable_defaults = something_changed;

    _FocusIn (p, event);

    if (something_changed && (p.reset != nil))
        update_fac (p, p.window, make_pref_list (p, p.modified));

    p.modified = nil;
    something_changed;
}


apply_cmd := vcWv.Cmd (fac, do_apply, "apply");
vcCmd.SetHelp (apply_cmd, "Try the preference changes in the current window");


function _Get(p, pref_region, type)
{
    option_num        := 0;
    something_changed := FALSE;
    modified_prefs    := nil;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
    {
        if (changed || is_modified (p, name))
            modified_prefs =
                cons (cons (string2symbol (name),
                            vcRegion.Get (get_editor (p, option_num), type)),
                      modified_prefs);
        ++option_num;
    }

    modified_prefs;
}


function _Set(p, option_list)
{
    option_num   := 0;
    new_modified := nil;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
    {
        option := assq (string2symbol (name), option_list);

        if (option)
        {
            vcRegion.Set (get_editor (p, option_num),
                          convert (type, cdr (option)));
            new_modified = cons (name, new_modified);
        }
        else if (changed || is_modified (p, name))
            vcRegion.Set (get_editor (p, option_num), val);

        ++option_num;
    }

    p.modified = new_modified;
    enable_cmds (p, vcRegion.Window (p.table));
}


function update_fac(p, w, pref_list) {
    foreach (r; vcRegion.Children (vcWin.Region (w), TRUE)) {
        if (vcRegion.Fac (r) == p.fac) {

            /*
             * If these are the standard preferences for this facility,
             * use vcRegion.PrefSet so the facility can ask has changed
             * by calling vcRegion.PrefGet, etc.  Otherwise, we have to
             * call the reset function directly, passing in the unique
             * preferences database.  This is done in ViewDRC, which has
             * two distinct preference database.  One for error severities,
             * and one for other settings.
             */
            if ((vcFac.Package (p.fac)).prefs == p.prefs)
                vcRegion.PrefSet (r, pref_list);
            else {
                data := vcRegion.Data (r, vcFac.Typespec (p.fac));
                p.reset (data, r, p.prefs, pref_list);
            } /* else */
        } /* if */
    } /* foreach */
} /* function update_fac */


function do_apply_all(p, event)
{
    save_modified := p.modified;
    if (do_apply (p, event) && (p.reset != nil))
    {
        pref_list := make_pref_list (p, save_modified);
        foreach (w; vcWin.Loop ())
            if (w != p.window)
                update_fac (p, w, pref_list);
    }

    p.enable_apply_all = FALSE;
    _FocusIn (p, event);
}


apply_all_cmd := vcWv.Cmd (fac, do_apply_all, "apply_all");
vcCmd.SetHelp (apply_all_cmd, "Try the preference changes in all windows");


function do_defaults(p, event)
{
    option_num        := 0;
    something_changed := FALSE;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
    {
        if (changed)
        {
            vlPref.Revert (p.prefs, name);
            something_changed = TRUE;
            p.modified = cons (name, p.modified);
        }

        save := p.modified; p.modified = nil;

        vcRegion.Set (get_editor (p, option_num), deflt);
        p.modified = save;

        ++option_num;
    }

    p.enable_defaults   = FALSE;
    p.enable_restore    = FALSE;
    p.enable_install    = TRUE;
    p.enable_apply      = TRUE;
    p.enable_apply_all  = TRUE;

    _FocusIn (p, event);

    if (something_changed && (p.reset != nil))
        update_fac (p, p.window, make_pref_list (p, p.modified));

    p.modified = nil;
}


defaults_cmd := vcWv.Cmd (fac, do_defaults, "defaults");
vcCmd.SetHelp (defaults_cmd,
               "Restore the preferences to their default values");


function do_restore(p, event)
{
    option_num        := 0;
    something_changed := FALSE;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
    {
        if (changed)
            something_changed = TRUE;

        if (is_modified (p, name))
            vcRegion.Set (get_editor (p, option_num), val);

        ++option_num;
    }

    p.enable_apply      = FALSE;
    p.enable_apply_all  = FALSE;
    p.enable_restore    = FALSE;
    p.enable_defaults   = something_changed;

    _FocusIn (p, event);

    p.modified = nil;
}


restore_cmd := vcWv.Cmd (fac, do_restore, "restore");
vcCmd.SetHelp (restore_cmd,
               "Restore the preferences to their previous values");


function do_install(p, event)
{
    do_apply (p, event);

    fname := "powerview.ini";
    private_ini_file := FALSE;

    if (isbound (`IniFileName, vcFac.Package (p.fac)))
    {
        private_ini_file = TRUE;
        fname = (vcFac.Package (p.fac)).IniFileName;
    }

    path := vscript.vsGetini (fname);

    if (path == nil)
        path = fname;

    if (!vlPref.WriteIni (p.prefs, path, !private_ini_file))
    {
        path = fname;

        include_package := !private_ini_file ||
            (p.prefs != (vcFac.Package (p.fac)).prefs);

        if (!vlPref.WriteIni (p.prefs, fname, include_package))
            vlEx.Throw2 (vlEx.FAILURE, "vcPref", vlErrSev.Error,
                         sprintf ("Could not write preferences to %s", fname));
    }

    p.enable_install = FALSE;
    _FocusIn (p, event);
    p.is_modified = FALSE;

    w := vcEvent.Window (event);
    vcWin.ClearStatus (w);
    vcWin.AddStatus (w, "Installed preferences in file " + path);
}

install_cmd := vcWv.Cmd (fac, do_install, "install");
vcCmd.SetHelp (install_cmd,
               "Permanently install the changes by writing a new version of the initialization file");


function is_modified(p, name)
{
    ret := FALSE;
    for (m := p.modified; (m != nil) && !ret; m = cdr (m))
        if (string_eq (car (m), name))
            ret = TRUE;
    ret;
}


function _Modified(p, region)
    p.is_modified;

/*
 * Constrain the default size of a region to a specific pixel width.
 * If the region is happy to be smaller, then let it be smaller.  If
 * it wants to be bigger, set its ideal size as specified in the
 * parameter to this size, but allow it to expand on a resize.
 * This is sort of painful to implement due to the lack of callbacks
 * for manipulating size structures.
 */
function constrain(r, width)
{
    size := vgSize.Alloc ();
    vcRegion.Size (r, size);

    ideal_x := vgSize.GetIdealWidth (size);
    min_x := vgSize.GetMinWidth (size);
    if (ideal_x > width) 
       vgSize.SetIdealWidth (size, width > min_x ? width : min_x);

    vcRegion.PrefSet (r, list (size <- size));

    r;
}

function label(str, font)
{
    r := vcRegion.Create (str, vcStaticText.fac, str);
    vcRegion.Font (r, font);
    r;
}

/*\vscmuserFunction vcPref.SetFac -- Set the current facility for the pref editor
*/
function SetFac(p_region, fac) {
    p := vcRegion.GetData (p_region);
    if (p.fac != fac)
        set_fac (p, fac, (vcFac.Package (fac)).prefs);
} /* function SetFac */

function set_fac(p, fac, prefs)
{
    p.fac   = fac;
    p.prefs = prefs;

    if (isbound (`_ResetPrefs, vcFac.Package (p.fac)))
        p.reset = (vcFac.Package (p.fac))._ResetPrefs;
    else if (isbound (`_Reset, vcFac.Package (p.fac)))
    {
        pkg     := vcFac.Package(p.fac);
        funk    := pkg._Reset;
        p.reset  = func(d, r, p, mod) funk(d, r);;
    }
    else
        p.reset = nil;

    n_options := 0;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
        ++n_options;
    
    p.table = vcTable.Create ("", n_options, p.n_columns);
/*    p.editors = make_vector (n_options); */

    if (p.show_help && p.show_default)
        p.editor_to_help_dflt = nil;
    else if (p.editor_to_help_dflt != nil)
        vlSym.Clear (p.editor_to_help_dflt);
    else {
        p.editor_to_help_dflt = vlSym.ObjCreate (11);
        vlSym.Dynamic (p.editor_to_help_dflt, 2.0, 2.0);
    } /* else */

    p.modified        = nil;
    p.enable_apply    = FALSE;
    p.enable_apply_all= FALSE;
    p.enable_install  = FALSE;
    p.enable_defaults = FALSE;
    p.enable_restore  = FALSE;

    function add_modified(name)
    {
        if (p.modified == nil)
        {
            p.enable_apply      = TRUE;
            p.enable_apply_all  = TRUE;
            p.enable_install    = TRUE;
            p.enable_defaults   = TRUE;
            p.enable_restore    = TRUE;
            _FocusIn (p, nil);
            p.is_modified = TRUE;
        }

        if (!is_modified (p, name))
            p.modified = cons (name, p.modified);
    }

    groups := nil;
    n_options = 0;

    foreach (name, help, type, val, deflt, changed;
             p.option_iterator (p.prefs))
    {
        if (changed)
            p.enable_defaults = TRUE;

        /* If this option is not part of a group that's already been created */
        group := option_group (p, name);
        if ((group == nil) || memq (group, groups)) {
            if (group != nil) {
                name = car (group);
                help = cadr (group);
                groups = cons (group, groups);
            } /* if */

            vcTable.Add (p.table, label (name, vgFont.small));
            if (p.show_default)
                vcTable.Add (p.table, constrain (label (deflt, vgFont.smallb),
                                                 150));

            ts_editor := create_editor (p.prefs, name, type, val,
                                        add_modified);
/*            p.editors[n_options] = ts_editor; */
            constrained_editor := constrain (ts_editor, 300);

            /*
             * If we are supressing the display of either the default value
             * or help string columns, we would like to at least show that
             * information in the status window, whenever the corresponding
             * editor has focus.  This is accomplished via
             * vcRegion.CallOnFocusChange, which takes the new focus region
             * as an argument.  We want to quickly determine the help and
             * default strings corresponding to each region in the hierarchy
             * of the editor.  We do that through a hash table mapping regions
             * into a cons cell holding the help and default value.
             */
            if (p.editor_to_help_dflt != nil) {
                help_deflt := cons (help, deflt);
                foreach (r; vcRegion.Children (ts_editor, TRUE))
                    if (vcBind.NumBindings (vcRegion.Bindings (r)) > 0)
                        p.editor_to_help_dflt[r] = help_deflt;
                if (vcBind.NumBindings (vcRegion.Bindings (ts_editor)) > 0)
                    p.editor_to_help_dflt[ts_editor] = help_deflt;
            } /* if */

            vcTable.Add (p.table, constrained_editor);
            if (p.show_help)
                vcTable.Add (p.table, label (help, vgFont.smalli));
        } /* if */
        ++n_options;
    } /* foreach */

    if ((n_options >= p.scroll_threshold) || (p.width != 0)) {
/*
        flags := list (vertical <- vcScroll.Presence.Always);
        if (p.width != 0)
            flags = cons (horizontal <- vcScroll.Presence.Sometimes, flags);

        scr := vcTable.Scroll3 ("scr", p.table, p.table,
                                3, p.n_visible_options, n_options,
                                p.width / 2, p.width, p.width,
                                flags);
*/

        scr := vcTable.Scroll ("scr", p.table, p.table,
                               3, p.n_visible_options, n_options);
    } /* if */
    else
        scr := vcRegion.Create ("", vcBorder.fac, p.table, rounded <- FALSE);

    vcTable.Set (p.big_table, 1, 0, scr);
    vcTable.AlignColumns (list (p.headers, p.table));
}

iterator all_prefs(p)
{
    foreach (name, help, type, val, deflt, changed; vlPref.Options (p))
    {
        if ((help[0] == '~') && (help[1] == '~'))
            help = substring (help, 2, 0);
        iterate (name, help, type, val, deflt, changed);
    }
}


function _ResetPrefs(p, region, prefs, modified) {
    function l(str) label (str, vgFont.mediumb);

    p.scroll_threshold  = vcRegion.PrefGet (region, "scroll_threshold");
    p.width             = vcRegion.PrefGet (region, "width");
    p.n_visible_options = vcRegion.PrefGet (region, "n_visible_options");
    p.show_help         = vcRegion.PrefGet (region, "show_help");
    p.show_default      = vcRegion.PrefGet (region, "show_default");

    p.n_columns = 2;
    p.editor_column = 1;
    if (p.show_default) {
        p.n_columns = p.n_columns + 1;
        p.editor_column = 2;
    } /* if */
    if (p.show_help)
        p.n_columns = p.n_columns + 1;
    p.headers = vcTable.Create ("", 1, p.n_columns);
    vcTable.Add (p.headers, l ("Name"));
    if (p.show_default)
        vcTable.Add (p.headers, l ("Default"));
    vcTable.Add (p.headers, l ("Value"));
    if (p.show_help)
        vcTable.Add (p.headers, l ("Help"));

    if (vcRegion.PrefGet (region, "all_prefs")) 
        p.option_iterator = all_prefs;
        
    vcTable.Set (p.big_table, 0, 0, p.headers);
    set_fac (p, p.fac, p.prefs);
} /* function _ResetPrefs */

function _Create(win_fac_prefs, region) {
    p := struct {
        prefs       = caddr (win_fac_prefs);
        table       = nil;
        modified    = nil;
        window      = car (win_fac_prefs);
        region      = region;
        fac         = cadr (win_fac_prefs);
        reset       = nil;
        big_table   = vcTable.Create ("", 2, 1);
        is_modified = FALSE;
        headers     = nil;
        option_iterator = vlPref.UserOptions;
        scroll_threshold = nil;         /* set in _ResetPrefs */
        width       = nil;              /* set in _ResetPrefs */
        n_visible_options = nil;        /* set in _ResetPrefs */
        show_help       = nil;          /* set in _ResetPrefs */
        show_default    = nil;          /* set in _ResetPrefs */
        n_columns       = 2;            /* fixed in _ResetPrefs */
        editor_column   = 1;            /* fixed in _ResetPrefs */
/*        editors         = nil;      */    /* fixed in _ResetPrefs */
        editor_to_help_dflt = nil;      /* fixed in set_fac */
        has_focus       = FALSE;    
        enable_apply    = FALSE;
        enable_apply_all= FALSE;
        enable_install  = FALSE;
        enable_defaults = FALSE;
        enable_restore  = FALSE};

    _ResetPrefs (p, region, prefs, nil);
    vcRegion.AddChild (region, p.big_table);

    /*
     * We allow the user or the programmer to supress the default field
     * and/or the help field.  If either of these is suppressed, then
     * we should put that info in the status line whenever the user focuses
     * on the corresponding editor.
     */
    vcRegion.CallOnFocusChange (region, func(new_focus_region) {
        /*
         * The editor may have been inserted in a region hierarchy and
         * then destroyed, in which case there's no way, unfortunately,
         * to cancel the FocusChange notify until the window is destroyed.
         * At least don't raise an exception.  Also, don't bother with this
         * if the user is showing the help and default fields.
         */
        if (!vcRegion.Destroyed (region) && (p.editor_to_help_dflt != nil)) {
            help_dflt := p.editor_to_help_dflt[new_focus_region];
            if (help_dflt != nil) {
                w := vcRegion.Window (region);
                vcWin.ClearStatus (w);
                if (!p.show_help)
                    vcWin.AddStatus (w, car (help_dflt));
                if (!p.show_help && !p.show_default)
                    vcWin.AddStatus (w, " ");
                if (!p.show_default) {
                    deflt_str := convert (vlStringVT, cdr (help_dflt));
                    vcWin.AddStatus (w, sprintf ("[%s]", deflt_str));
                } /* if */
                p.has_focus = TRUE;
            } /* if */
            else if (p.has_focus) {
                vcWin.ClearStatus (vcRegion.Window (region));
                p.has_focus = FALSE;
            } /* else if */
        } /* if */
        else
            p.has_focus = FALSE;
    });
    p;
}


vlPref.Replace (prefs, "all_prefs", vlTrueFalseVT,
                "~~Show the programmer preferences in the Preference Editor,
as well as the user preferencess",
                FALSE);
vlPref.Replace (prefs, "scroll_threshold", vlUnsignedIntVT,
                "~~Number of options below which we don't add a scrollbar",
                8);
vlPref.Replace (prefs, "show_default", vlTrueFalseVT,
                "Show the default value of each preference\n(goes in status when selected if FALSE)",
                TRUE);
vlPref.Replace (prefs, "show_help", vlTrueFalseVT,
                "Show the help string for each preference\n(goes in status when selected if FALSE)",
                TRUE);
vlPref.Replace (prefs, "n_visible_options", vlUnsignedIntVT,
                "Number of visible options to show by default",
                8);
vlPref.Replace (prefs, "width", vlUnsignedIntVT,
                "Pixel width of pref editor\n0 means expand to ideal size\nnon-zero means use a horizontal scrollbar if necessary",
                0);

vlPref.Parse ("powerview.ini", wv);

menu := vcWv.MakeMenu ("Preference",
                        vcWin.RedSquareEntryNoPlotNoQuit,
                        list ("Set &Facility", fac_cmd),
                        list ("&Apply", apply_cmd),
                        list ("Apply A&ll", apply_all_cmd),
                        list ("&Defaults", defaults_cmd),
                        list ("&Restore", restore_cmd),
                        list ("&Install", install_cmd));
vcFac.SetMenus (fac, menu);


function create_window(w, fac, prefs, vcpref_prefs) {
    name := vcFac.Title (fac);
    r    := vcRegion.Create1 (name, vcPref.fac, list (w, fac, prefs),
                              vcpref_prefs);
    w     = vcWin.Create (w, r);

    vcWin.SetTitle (w, "Preference Editor: " + name);

    vcWin.SetConfirmMessage (w, "
 You have changed the preferences without saving 
 them permanently to the appropriate initialization
 file. You may continue and discard
 your changes, or you may cancel this dialog box, press 
 the Install button, and then dismiss the window.
");
    w;
} /* function create_window */

/*\vscmuserFunction vcPref.CreateSmallWindow -- Creates pref window without help or default fields

DESCRIPTION
     Like CreateWindowAux, but disables the help and defaults fields.
     This should be used when the preference editor would otherwise
     come up too large, due to intricate editing facilities.

RETURNS
     The window ID.
*/
function CreateSmallWindow(w, fac, prefs) {
    create_window (w, fac, prefs,
                   list (show_help <- FALSE,
                         show_default <- FALSE));
} /* function CreateSmallWindow */


/*\vscmuserFunction vcPref.CreateWindowAux -- Creates pref window with auxiliary database

DESCRIPTION
     Like CreateWindow, but takes an auxiliary preference database
     instead of using the facility's default one.

RETURNS
     The window ID.
*/
function CreateWindowAux(w, fac, prefs) {
    create_window (w, fac, prefs, nil);
} /* function CreateWindowAux */

/*\vscmuserFunction vcPref.CreateWindow -- Creates preference window for facility

RETURNS
     The window ID.
*/
function CreateWindow(w, fac) {
    if (!fac_has_pref (fac))
        fac = vcWv.fac;
    CreateWindowAux (w, fac, (vcFac.Package (fac)).prefs);
} /* function CreateWindow */


if (!isbound (`pref_editor_tables, current_package ()))
{
    pref_editor_tables := vlSym.ObjCreate (11);
    vlSym.Dynamic (pref_editor_tables, 2, 2);
}


struct vcPrefGroupInfoVT {
    vcFacilityVT fac;
    vscmVT fac_prefs;
    vscmVT option_names;
    vlStringVT group_name;
    vlStringVT help;
};

/*\vscmuserFunction vcPref.SetEditor -- Change the editor used for a particular preference option

DESCRIPTION
     This is used to set up a custom editor for a preference option,
     or for grouping together a list of options.

     `pref_database' and `option_name' define the option you want to
     change the editor for.

     `option_name' can be either a single name string, or a list of names:

          list ("Group Title", "Group Help", "option1", "option2"...)

     If a list, the first element is the title of the group, which
     does not correspond to a real preference.  The remaining elements
     of the list are the option names included in the group.  A single
     editor will be created for all the members of the group.  The
     data passed to the editor's _Create routine, and expected from
     the editor's _Get routine, take the form:

          list ((name1 . value1), (name2 . value2)...)

     `facility' defines the editor for the option name.

     `fac_prefs' are passed to that facility with vcRegion.Create.

NOTE
      The "list of option names" feature is not supported yet.  We
      must work with a single option name.
*/
function SetEditor(pref_database, option_name, facility, fac_prefs...) {
    /* Now args contains (fac pref1 pref2 ...) */

    /*
     * Get the table for this preference database.
     */
    table := vlSym.Fetch (pref_editor_tables, pref_database);

    /*
     * If there is none, set one up.
     */
    if (table == nil) {
        table = vlSym.StrCreate (11);
        vlSym.Dynamic (pref_editor_tables, 2, 2);
        vlSym.Store (pref_editor_tables, pref_database, table);
    } /* if */

    if (istype (option_name, consVT))
        vlSym.Store (table, name, struct vcPrefGroupInfoVT {
            facility, fac_prefs,
            cddr (option_name), /* group members */
            car (option_name),  /* group title */
            cadr (option_name)  /* group help */
        });
    else
        vlSym.Store (table, option_name, cons (facility, fac_prefs));
} /* function SetEditor */

function option_group(p, option_name) {
    group := nil;
    table := vlSym.Fetch (pref_editor_tables, prefs);
    if (table != nil) {
        fac_and_args := vlSym.Fetch (table, option_name);
        if (istype (fac_and_args, vcPrefGroupInfoVT))
            group = fac_and_args;
    } /* if */
    group;
} /* function option_is_part_of_group */

function create_editor(prefs, option_name, type, val, add_modified) {
    table := vlSym.Fetch (pref_editor_tables, prefs);
    option_list := list (option_name);
    declare fac, pref_list, env;

    if (table != nil) {
        fac_prefs := vlSym.Fetch (table, option_name);
        if (istype (fac_prefs, vcPrefGroupInfoVT)) {
            option_list = fac_prefs.option_list;
            fac_prefs = fac_prefs.fac_prefs;
            val := nil;
            foreach (option; vscript.list_iter (option_list))
                val = cons (vlPref.Val (prefs, option), val);
            val = reverse_destructive (val);
        } /* if */

        if (fac_prefs != nil) {
            fac = car (fac_prefs);
            pref_list = cdr (fac_prefs);
            env = vcFac.Package (fac);
        } /* if */
    } /* if */

    if (fac == nil) {
        env = vcCmd.TsToFac (type);
        fac = env.fac;
        pref_list = nil;
    } /* if */

    editor := vcRegion.Create1 (option_name, fac, val, pref_list);
    env.CallOnChange (editor, func(args...) {
        foreach (option_name; vscript.list_iter (option_list))
            add_modified (option_name);
    });
    editor;
} /* function create_editor */
