/*
 *  $Revision: 2.1.1.1 $  $Date    : 94/06/24 23:36:15 $
 *
 *  Facility for browsing VDM classes.
 *
 *  Copyright (c) 1995 by Viewlogic Systems, Inc.
 */



/*\vscmuserSection vdmClassSelect. -- Facilty for browsing VDM classes.
 *
 * DESCRIPTION
 *      This facility is provides a browser for classes stored in a VDM
 *      context.  It's based on the vcFilePop Core facility for browsing files
 *      and its layout is identical to vcFilePop. 
 *
 *      You can request from a vdmClassBrowse region data of either the
 *      vpmLocationVT type, or vlStringVT.  Use the vcRegion.Get(man) function
 *      for requesting the current selection from the browser.
 *
 * FACILITY DATA
 *      consVT  - a pair consisting of a VDM context and a default class.
 *                If class cannot be procurred, use nil.  The class can
 *                be specified either as an actual vdmClassVT object or as
 *                an alias.
 *
 *                Alternately, the initialization value may be just the
 *                class, and the context may be specified via the
 *                "cntx" preference.
 *
 * PREFERENCES
 *      cntx - vdmCntxVT  -       See FACILITY DATA above.
 *
 * SEE ALSO
 *      vdmLocSelect.
 */



vcWv.RequireFac (wv, "vcMultOpts");



/*\Docpage MakeClassSelect -- Build the browser's popup window.
 * 
 * DESCRIPTION
 *      If the user presses the Browse button, this function will build the
 *      popup window with a picklist listing all the avaiable classes.
 */
function MakeClassSelect(this, event)
{
    /*
     * When the class browse picklist is destroyed, we need to enable the
     * Browse button again.
     */
    function destroy_pick(region)
        if (this.popup != nil)
        {
            vcRegion.Enable (this.button, TRUE, TRUE);
            this.popup = nil;
        }


    /*
     * When the user makes a picklist selection, set the command line as well.
     */
    function set_cmdline(class, on_or_off, index)
        if (on_or_off)
            vcCmd.Run (ResetValueCmd, this, event,
                       list (vdm.ClassGetName (class)), TRUE);


    /*
     * Build a scrollable picklist listing all the classes currently visible
     * in the VDM context.  The picklist data is an actual class structure.
     * When the user selects an entry, the command line is set to the
     * class's alias.
     */
    pick := vcRegion.Create ("", vcPicklist.fac, this.classes,
                             exclusive          <- TRUE,
                             will_scroll        <- TRUE,
                             enhanced_notify    <- TRUE);
    pick := vcPicklist.Scroll ("", pick, 5, 10, 0);

    vcPicklist.Notify (pick, set_cmdline);


    /*
     * Give the picklist some kind of a title.
     */
    pick_msg := vcRegion.Create ("msg", vcStaticText.fac,
                                 "\002Available classes",
                                 align          <- vgAlign.MiddleCenter,
                                 formatted      <- TRUE,
                                 recessed_text  <- TRUE);
    vcRegion.FocusInhibit (pick_msg);


    /*
     * Dissmis button to get rid of the popup.
     */
    dismiss_button :=
        vcRegion.Create ("buttons", vcButnRow.fac,
                         list (list ("accept", "Accept", func();,
                                     list (close_window <- TRUE))));

    table := vcTable.Create ("", 3, 1);
    vcTable.Add (table, pick_msg);
    vcTable.Add (table, pick);
    vcTable.Add (table, dismiss_button);

    this.popup = vcWin.Popup (vcEvent.Window (event), table, pick,
                              vcWin.NoFlags);
    vcRegion.Enable (this.button, FALSE, TRUE);

    /*
     * When the popup is closed, we need to enable the Browse button again.
     */
    vcRegion.DestroyNotify (pick, destroy_pick);
}


BrowseCmd := vcWv.Cmd (fac, MakeClassSelect, "browse");

vcCmd.EnableMacro (BrowseCmd, FALSE);



/*\Docpage _Create -- Required Core method.
 * 
 * DESCRIPTION
 *      Defines the GUI layout of the facility.  Note that the facility's data
 *      structure stores the individual layout components (the browse button,
 *      command line) so that other methods don't have to hard code the exact
 *      positions of those elements within the main table.
 */
function _Create(default_data, region)
{
    vdm_cntx := vcRegion.PrefGet (region, "cntx");

    if (vdm_cntx == nil)
    {
        vdm_cntx       = car (default_data);
        default_class := cdr (default_data);
    }
    else
        default_class := default_data;

    parent_name := vcRegion.Name (region);
    cmdline     := vcRegion.Create (parent_name + "_cmdline", vcCmdline.fac,
                                    "", macro_on <- vcRegion.MacroOn (region));
    button      := vcWv.LabelButtonCreate (parent_name + "_browse",
                                           "Browse...", nil, BrowseCmd, nil);

    vcRegion.FocusInhibit (cmdline);
    vcRegion.FocusInhibit (button);

    table := vcTable.Create ("", 1, 2);

    vcTable.Add (table, cmdline);
    vcTable.Add (table, button);

    vcRegion.AddChild (region, table);

    /*
     * Process the default class.  Note that it can be either a real class
     * object, or a string representing a class name.
     */
    if (istype (default_class, vdmClassVT))
        vcCmdline.Set (cmdline, vdm.ClassGetName (default_class));

    else if (istype (default_class, vlStringVT) &&
             (! string_eq (default_class, ""))  &&
             vdm.CntxFindClass (vdm_cntx, default_class))
        vcCmdline.Set (cmdline, default_class);


    /*
     * Predicate function used to sort the class list.
     */
    function class_name_eq(elem1, elem2)
        string_le (car (elem1), car (elem2));


    /*
     * Gather the list of classes and a list of completions.  Note that the
     * latter is a flat list of class names.
     */
    completion_list := nil;
    class_list      := nil;

    foreach (bunch, bunch_name; vdm.CntxLoopClassFiles (vdm_cntx))
        foreach (class, class_name; vdm.ClassFileLoop (bunch))
        {
            completion_list = cons (class_name, completion_list);
            class_list      = cons (cons (class_name, class), class_list);
        }

    class_list = vscript.Sort (class_list, class_name_eq);


    this := struct
    {
        table    = table;           /* The entire vdmLocSelect layout       */
        cmdline  = cmdline;         /* Command line, for easy access        */
        button   = button;          /* Browse button                        */
        popup    = nil;             /* Popup brows window                   */
        complete = completion_list; /* Completion list                      */
        classes  = class_list;      /* List of class name/class pairs       */
        cntx     = vdm_cntx         /* VDM context for the current instance */
    };

    this;
}



/*\Docpage _Destroy -- How to clean up.
 * 
 * DESCRIPTION
 *      In case external forces destroy the facility, don't forget to get rid
 *      of the popup browse window.
 */
function _Destroy(this)
    if (this.popup != nil)
    {
        win        := this.popup;
        this.popup  = nil;
        vcWin.Destroy (win);
    }



/*\Docpage _Get -- Extract data.
 *
 * DESCRIPTION
 *      Returns a class object.  If it is an illegal class name, and we have
 *      requested a string, we don't want "nil", we want the name so we can
 *      report an error properly.
 */
function _Get(this, region, ts)
    if (ts == vlStringVT)
        vcCmdline.Get (this.cmdline);
    else
        vdm.ClassFindClass (this.cntx, vcCmdline.Get (this.cmdline));



/*\Docpage complete -- Complete a partial selection.
 */
function complete(this, event)
{
    old := vcCmdline.Get (this.cmdline);
    new := vcMultOpts.Complete (old, this.complete);

    if (new == nil)
        vcWin.VisualBell (vcEvent.Window (event));
    else
        vcCmdline.Set (this.cmdline, new);
}

CompleteCmd := vcWv.Cmd (fac, complete, "complete");



/*\Docpage ResetValue -- Action routine for the ResetValueCmd.
 *
 * DESCRIPTION
 *      This command is run whenever class browse picklist changes values,
 *      so that the new value is recorded in the macro file in a coherent
 *      manner.  We do this in lieu of making the browse picklist popup macro
 *      recordable -- those don't record so well.
 */
function ResetValue(this, event, value)
    vcCmdline.Set (this.cmdline, value);

ResetValueCmd := vcWv.Cmd (fac, ResetValue, "reset", "Reset Value",
                           vcWv.Param ("", vlStringVT, vcWv.Constant, ""));


/*\Docpage vdmClassSelect.Set - set the value of the region
 *
 * SYNOPSIS
 *      vdmLocSelect.Set(region, str)
 *          vcRegionVT  region;
 *          vlStringVT  str;
 */
function Set(r, str)
{
    this := vcRegion.GetData (r);
    vcCmdline.Set (this.cmdline, str);
}



/****************************************************************************/
/*                          Key bindings.                                   */
/****************************************************************************/


vcBind.Set (bindings, '?',          BrowseCmd);
vcBind.Set (bindings, '\rightdown', BrowseCmd);
vcBind.Set (bindings, ' ',          CompleteCmd);



/****************************************************************************/
/*                          Preferences.                                    */
/****************************************************************************/

vlPref.Replace (prefs, "cntx", vscmVT, "The vdm context", nil);
