/*
 *  $Revision: 1.5.1.8 $  $Date: 1995/05/04 21:26:51 $
 *
 *  Useful editor functions for the VDMF class definitions.
 *
 *  Copyright (c) 1994,1995 by Viewlogic Systems, Inc.
 */



/*
 * NOTE: This file is loaded into the top level package.  It's loaded from the
 * vdmMaster.scm file.
 */



/*
 * Declare ITC source contexts used by the EditViewDraw function.
 */
vdraw_context := vitc.CntxCreate (vipc.GidDFLT, vipc.UidDFLT,
                                  vipc.FidDFLT, vitc.TidVDRAW,
                                  vitc.NameDFLT, vipc.HnmANY,
                                  vipc.PidANY, vipc.CidANY);

my_context    := vitc.SrcCntxCreate ("vdmEdit", "VDMF Editors");


vbubble_context := vitc.CntxCreate (vipc.GidDFLT, vipc.UidDFLT,
                                    vipc.FidDFLT, "ViewFSM-Capture",
                                    nil, vipc.HnmANY, vipc.PidANY,
                                    vipc.CidANY);


/*\Docpage vdmEditViewDraw -- ViewDraw editor.
 *
 * DESCRIPTION
 *      It handles the following situations:
 *
 *      1. ViewData is loaded into the vdraw package.
 *
 *      2. We are not in ViewDraw, but ViewDraw is already up an running and
 *         just needs to open a new window, instead of invoking yet another
 *         process.
 *
 *      3. ViewDraw is not running, and we need another process.
 */
function vdmEditViewDraw(obj, outProc, errProc, exitProc)
{
    declare cmd;

    location     := vdm.ObjGetLocation (obj);
    name         := vdm.ObjGetName (obj);
    class_name   := vdm.ClassGetName (vdm.ObjGetClass (obj));
    is_schematic := TRUE;

    if (string_eq (class_name, "Schematic"))
        cmd = "locate - sch " + name;
    else
    {
        cmd = "locate - sym " + name;
        is_schematic = FALSE;
    }


    /*
     * If this call is being made from ViewDraw, just call it.
     */
    if (isbound (`vdraw, current_package ()))
    {
        if (is_schematic)
            vdraw.schematic (name);
        else
            vdraw.symbol (name);
    }

    /*
     * Else assume that ViewDraw is up and running and send the command to it.
     * If that fails, spawn a new ViewDraw, register to receive a tool start
     * message from it and then send the command to it.
     */
    else if (vitc.Send (vitc.MsgCMD, my_context, vdraw_context, cmd,
                        vipc.FlgDIRECT) != vitc.Success)
    {
        if (is_schematic)
            vcSys.Command ("viewdraw -sch " + name, nil, outProc, errProc,
                           exitProc, nil);
        else
            vcSys.Command ("viewdraw -sym " + name, nil, outProc, errProc,
                           exitProc, nil);
    }
    else if (! null (exitProc))
        exitProc (vcSysMode.Exited, 0);
}



/*\Docpage vdmEditDefault -- Default VDM editor.
 * 
 * DESCRIPTION
 *      The vcWin.TextEdit function does most of the work in figuring out
 *      which editor to use.  If it's unavailable, we try to pick off the
 *      value of the EDITOR environment variable and spawn a subprocess.  If
 *      that fails, an error is signaled to the user and the function exits.
 */
function vdmEditDefault(dataObj, outProc, errProc, exitProc)
{
    path := vdm.ObjGetPhysicalDescriptor (dataObj);

    if (isbound (`vcWin, current_package ()))
        vcWin.TextEdit (nil, path, nil, exitProc, FALSE, FALSE);
    else if ((editor := vscript.getenv ("EDITOR")) != nil)
        vcSys.Command (editor + " " + path, nil, outProc, errProc, exitProc,
                       nil);
    else
    {
        vlErr.Out2 ("vdmEdit", vlErrSev.Error,
                    "No default editor available to edit " + path);
        if (! null (exitProc))
            exitProc (vcSysMode.Exited, 1);
    }
}



/*\Docpage vdmEditMegafile -- Megafile editor.
 * 
 * DESCRIPTION
 *   Pops up a new vcBrowse window to display the contents of a Megafile.
 *   Requires the "vdmMega" storage manager.
 */
function vdmEditMegafile(dataObj, outProc, errProc, exitProc)
{
    declare device, path, fname, ext;
    path = vdm.ObjGetPhysicalDescriptor (dataObj);
    <device, path, fname, ext> = vscript.vsParse (path);
    path = vscript.vsMerge (device, path, "", "");

    if (isbound (`vdata, current_package ()))
    {
        /* function to call when editor is destroyed */
        function edit_complete(reg)
            if (!null (exitProc))
                exitProc (vcSysMode.Exited, 0);

        /* create a location managed by the vdmMega storage manager */
        location := vpm.LocationCreate (vdm.ObjGetName (dataObj), 
                                        "vdmMega", nil, path, nil);

        /* set the foreign data to location megafile lives in so that 
         * the Mega SM encapsulation can do SM operations using the 
         * SM associated with the actual Megafile
         */
        vpm.LocationSetForeignData (location, vdm.ObjGetLocation (dataObj));

        /* create a new context so we can have a map that contains only our 
         * megafile location.
         */
        context := vdm.CntxCreate (sprintf ("MegafileEdit%s", vcWv.Time()));
        map := vdm.CntxGetMap (context);

        vpm.MapClearLocations (map);
        vpm.MapAddLocation (map, location);

        vd := vcRegion.Create1 ("Megafile " + path + fname, 
                                vdata.fac, 
                                context, 
                                list (show_info   <- FALSE,
                                      show_legend <- FALSE,
                                      hide_root   <- TRUE));

        vcRegion.DestroyNotify (vd, edit_complete);
        win := vcWin.Create (vd);
    }
    else
    {
        vlErr.Out2 ("vdmEditMegafile", vlErrSev.Error,
                    "Megafile " + path + 
                    " can only be edited from within ViewData.");
        if (! null (exitProc))
            exitProc (vcSysMode.Exited, 1);
    }
}



/*\Docpage vflFlow_edit - viewflow editor
 *
 * DESCRIPTION
 *    Edits a vflFlow object (a flow for ViewFlow). 
 *    It first searches for a vflow region in the current process;
 *    if none is found, a new ViewFlow process is kicked off.
 */
function vflFlow_edit(obj, output, error, exit)
{
    /* first look for a vflow region in our executable */
    declare vflow_rgn;

    if (isbound (`vflow, nil))
        foreach (w; vcWin.Loop ())
            if (vflow_rgn == nil)
                vflow_rgn = vcRegion.FindFac (vcWin.OwnerRegion (w), 
                                              vflow.fac);

    if (vflow_rgn == nil)
        vcSys.Command ("viewflow -mode edit " + vdm.ObjGetName (obj),
                        nil, output, error, exit, nil);
    else
    {
        /* this part here assumes that the vflow facility is defined */
        unwind_protect
        {
            ev := vcEvent.Create (0, 0, ' ', nil, 
                        vcRegion.Window (vflow_rgn), vflow_rgn, nil);
            vcCmd.Run (vflow.OpenCmd, vcRegion.GetData (vflow_rgn),
                          ev, list (vdm.ObjGetName (obj),
                     vpm.LocationGetAlias (vdm.ObjGetLocation (obj)), 
                                 vflowMode.Edit), FALSE);
        }
        cleanup
            exit (vcSysMode.Exited, 0);
    }
}



/*\Docpage vdmEditViewFSM -- ViewFSM editor.
 */
function vdmEditViewFSM(obj, outProc, errProc, exitProc)
{
    declare cmd;

    location     := vdm.ObjGetLocation (obj);
    name         := vdm.ObjGetName (obj);
    class_name   := vdm.ClassGetName (vdm.ObjGetClass (obj));

    cmd = "locate " + name;


    /*
     * Else assume that ViewFSM is up and running and send the command to it.
     * If that fails, spawn a new ViewFSM, register to receive a tool start
     * message from it and then send the command to it.
     */
    if (vitc.Send ("Command Line", my_context, vbubble_context,
                   "sch " + name, vipc.FlgDIRECT) != vitc.Success)
    {
        vcSys.Command ("vfsmcap " + name, nil, outProc, errProc,
                       exitProc, nil);
    }
    else if (! null (exitProc))
        exitProc (vcSysMode.Exited, 0);
}
