/*
 *  $Revision: 2.1.1.9 $  $Date    : 94/06/24 23:36:15 $
 *
 *  Facility for browsing VPM locations.
 *
 *  Copyright (c) 1995 by Viewlogic Systems, Inc.
 */



/*\vscmuserSection vdmLocSelect. -- Facilty for browsing VDM locations.
 *
 * DESCRIPTION
 *      This facility is provides a browser for locations stored in a VDM
 *      context.  It preserves the search order of the locations as they are
 *      stored in the context.  It's based on the vcFilePop Core facility for
 *      browsing files and its layout is identical to vcFilePop.
 *
 *      You can request from a vdmLocBrowse region data of either the
 *      vpmLocationVT type, or vlStringVT.  Use the vcRegion.Get(man) function
 *      for requesting the current selection from the browser.
 *
 * FACILITY DATA
 *      consVT  - a pair consisting of a VDM context and a default location.
 *                If location cannot be procurred, use nil.  The location can
 *                be specified either as an actual vpmLocationVT object or as
 *                a location alias.
 *
 *                Alternately, the initialization value may be just the
 *                location, and the context may be specified via the
 *                "cntx" preference.
 *
 * EXAMPLE
 *      This example is taken from the ViewData's vdata.vs file.  It shows how
 *      to use the vdmLocSelect facility as a parameter editor facility:
 *
 *      function get_default_browse_data(dv)
 *      {
 *          default_location := nil;
 *
 *          foreach (obj, class; vlSym.Loop (dv.selected_table))
 *              if (match_type_or_class (vpmLocationVT, obj, class))
 *              {
 *                  default_location = obj;
 *                  break;
 *              }
 *
 *          cons (dv.cntx, default_location);
 *      }
 *
 *
 *      CreateCollectionCmd :=
 *          vcWv.Cmd (fac, CreateCollection, "create_collection",
 *                    "Create Collection",
 *              vcWv.Param ("Name", vlStringVT, vcWv.Sticky, "",
 *                          "Collection's name"),
 *              vcWv.Param ("Class", vlStringVT, vcWv.Sticky, "vdmCollection",
 *                          config_select_fac,
 *                          "Type of collection to be built"),
 *              vcWv.Param ("Location", vscmVT, vcWv.Dynamic,
 *                          get_default_browse_data, vdmLocSelect.fac,
 *                          "Location where the collection will be placed"));
 *
 * PREFERENCES
 *      cntx - vdmCntxVT  -       See FACILITY DATA above.
 *
 * SEE ALSO
 *      vdmClassSelect.
 */



vcWv.RequireFac (wv, "vcMultOpts");



/*\Docpage MakeLocationSelect -- Build the browser's popup window.
 * 
 * DESCRIPTION
 *      If the user presses the Browse button, this function will build the
 *      popup window with a picklist listing all the avaiable locations.
 */
function MakeLocationSelect(this, event)
{
    /*
     * When the location browse picklist is destroyed, we need to enable the
     * Browse button again.
     */
    function destroy_pick(region)
        if (this.popup != nil)
        {
            vcRegion.Enable (this.button, TRUE, TRUE);
            this.popup = nil;
        }


    /*
     * When the user makes a picklist selection, set the command line as well.
     */
    function set_cmdline(location, on_or_off, index)
        if (on_or_off)
            vcCmd.Run (ResetValueCmd, this, event,
                       list (vpm.LocationGetAlias (location)), TRUE);


    /*
     * Build a scrollable picklist listing all the locations currently visible
     * in the VDM context.  The picklist data is an actual location structure.
     * When the user selects an entry, the command line is set to the
     * location's alias.
     */
    pick := vcRegion.Create ("", vcPicklist.fac, this.loc_list,
                             exclusive          <- TRUE,
                             will_scroll        <- TRUE,
                             enhanced_notify    <- TRUE);
    pick := vcPicklist.Scroll ("", pick, 5, 10, 0);

    vcPicklist.Notify (pick, set_cmdline);


    /*
     * Give the picklist some kind of a title.
     */
    pick_msg := vcRegion.Create ("msg", vcStaticText.fac,
                                 "\002Available locations",
                                 align          <- vgAlign.MiddleCenter,
                                 formatted      <- TRUE,
                                 recessed_text  <- TRUE);
    vcRegion.FocusInhibit (pick_msg);


    /*
     * Dissmis button to get rid of the popup.
     */
    dismiss_button :=
        vcRegion.Create ("buttons", vcButnRow.fac,
                         list (list ("accept", "Accept", func();,
                                     list (close_window <- TRUE))));

    table := vcTable.Create ("", 3, 1);
    vcTable.Add (table, pick_msg);
    vcTable.Add (table, pick);
    vcTable.Add (table, dismiss_button);

    this.popup = vcWin.Popup (vcEvent.Window (event), table, pick,
                              vcWin.NoFlags);
    vcRegion.Enable (this.button, FALSE, TRUE);

    /*
     * When the popup is closed, we need to enable the Browse button again.
     */
    vcRegion.DestroyNotify (pick, destroy_pick);
}


BrowseCmd := vcWv.Cmd (fac, MakeLocationSelect, "browse");

vcCmd.EnableMacro (BrowseCmd, FALSE);



/*\Docpage _Create -- Required Core method.
 * 
 * DESCRIPTION
 *      Defines the GUI layout of the facility.  Note that the facility's data
 *      structure stores the individual layout components (the browse button,
 *      command line) so that other methods don't have to hard code the exact
 *      positions of those elements within the main table.
 */
function _Create(default_data, region)
{
    vdm_cntx := vcRegion.PrefGet (region, "cntx");

    if (vdm_cntx == nil)
    {
        vdm_cntx          = car (default_data);
        default_location := cdr (default_data);
    }
    else
        default_location := default_data;

    parent_name := vcRegion.Name (region);
    cmdline     := vcRegion.Create (parent_name + "_cmdline", vcCmdline.fac,
                                    "", macro_on <- vcRegion.MacroOn (region));
    button      := vcWv.LabelButtonCreate (parent_name + "_browse",
                                           "Browse...", nil, BrowseCmd, nil);

    vcRegion.FocusInhibit (cmdline);
    vcRegion.FocusInhibit (button);

    table := vcTable.Create ("", 1, 2);

    vcTable.Add (table, cmdline);
    vcTable.Add (table, button);

    vcRegion.AddChild (region, table);

    /*
     * Process the default location.  Note that it can be either a real
     * location object, or a string representing a location alias.
     */
    if (istype (default_location, vpmLocationVT))
        vcCmdline.Set (cmdline, vpm.LocationGetAlias (default_location));

    else if (istype (default_location, vlStringVT) &&
             (! string_eq (default_location, "")))
        foreach (loc; vpm.MapLoopLocation (vdm.CntxGetMap (vdm_cntx)))
            if (string_eq (default_location, vpm.LocationGetAlias (loc)))
            {
                vcCmdline.Set (cmdline, vpm.LocationGetAlias (loc));
                break;
            }
            
    /*
     * Gather the list of locations.  Note that I need to reverse it, because
     * the list is built in the reverse order using the cons function.
     */
    loc_list      := nil;
    complete_list := nil;

    foreach (loc; vpm.MapLoopLocation (vdm.CntxGetMap (vdm_cntx)))
    {
        loc_list = cons (cons (vpm.LocationGetAlias (loc), loc), loc_list);
        complete_list = cons (vpm.LocationGetAlias (loc), complete_list);
    }

    loc_list = reverse_destructive (loc_list);


    this := struct
    {
        table    = table;           /* The entire vdmLocSelect layout       */
        cmdline  = cmdline;         /* Command line, for easy access        */
        button   = button;          /* Browse button                        */
        popup    = nil;             /* Popup brows window                   */
        complete = complete_list;   /* Completion list                      */
        loc_list = loc_list;        /* List of alias/location pairs         */
        cntx     = vdm_cntx         /* VDM context for the current instance */
    };

    this;
}



/*\Docpage _Destroy -- How to clean up.
 * 
 * DESCRIPTION
 *      In case external forces destroy the facility, don't forget to get rid
 *      of the popup browse window.
 */
function _Destroy(this)
    if (this.popup != nil)
    {
        win        := this.popup;
        this.popup  = nil;
        vcWin.Destroy (win);
    }



/*\Docpage _Get -- Extract data.
 *
 * DESCRIPTION
 *      Returns a location object.  If it is an illegal location name, and we
 *      have requested a string, we don't want "nil", we want the name so we
 *      can report an error properly.
 */
function _Get(this, region, ts)
{
    str := vcCmdline.Get (this.cmdline);

    if (((str == nil) || (string_eq (str, ""))) &&
           !vcRegion.PrefGet (region, "allow_empty"))
        vlEx.Throw2 (vlEx.FAILURE, "vdmLocSelect", vlErrSev.Error,
                     "You must specify a location alias.");

    if (ts == vlStringVT)
        str;
    else
    {
        loc := vpm.MapFindLocation (vdm.CntxGetMap (this.cntx), str);

        if (loc == nil)
            vlEx.Throw2 (vlEx.FAILURE, "vdmLocSelect", vlErrSev.Error,
                         sprintf ("Unknown location `%s'.", str));
        loc;
    }
}



/*\Docpage complete -- Complete location selection.
 * 
 * DESCRIPTION
 *      Attempts to complete a partially completed location alias entry.
 */
function complete(this, event)
{
    old := vcCmdline.Get (this.cmdline);
    new := vcMultOpts.Complete (old, this.complete);

    if (new == nil)
        vcWin.VisualBell (vcEvent.Window (event));
    else
        vcCmdline.Set (this.cmdline, new);
}

CompleteCmd := vcWv.Cmd (fac, complete, "complete");



/*\Docpage ResetValue -- Action routine for the ResetValueCmd.
 *
 * DESCRIPTION
 *      This command is run whenever location browse picklist changes values,
 *      so that the new value is recorded in the macro file in a coherent
 *      manner.  We do this in lieu of making the browse picklist popup macro
 *      recordable -- those don't record so well.
 */
function ResetValue(this, event, value)
    vcCmdline.Set (this.cmdline, value);

ResetValueCmd := vcWv.Cmd (fac, ResetValue, "reset", "Reset Value",
                           vcWv.Param ("", vlStringVT, vcWv.Constant, ""));



/*\Docpage vdmLocSelect.Set - set the value of the region
 *
 * SYNOPSIS
 *      vdmLocSelect.Set(region, str)
 *          vcRegionVT  region;
 *          vlStringVT  str;
 */
function Set(r, str)
{
    this := vcRegion.GetData (r);
    vcCmdline.Set (this.cmdline, str);
}


/****************************************************************************/
/*                          Key bindings.                                   */
/****************************************************************************/


vcBind.Set (bindings, '?',          BrowseCmd);
vcBind.Set (bindings, '\rightdown', BrowseCmd);
vcBind.Set (bindings, ' ',          CompleteCmd);



/****************************************************************************/
/*                          Preferences.                                    */
/****************************************************************************/

vlPref.Replace (prefs, "cntx", vscmVT, "~~The VDM context", nil);
vlPref.Replace (prefs, "allow_empty", vlYesNoVT, 
                     "~~Whether to allow empty string", FALSE);
