/*
 *  $Revision: 1.1.1.7 $
 *
 *  VDM encapsulation file for megafiles.
 *
 *  This is a first attempt for providing support for megafile. This
 *  encapsulation is very sparse. The primary feature provided by this
 *  encapsulation is the ability to browse the elements of a megafile.
 *  By "browse" we mean that the user can get a listing of the elements
 *  which reside in a megafile.
 *
 *  For more information on how to write storage manager encapsulations,
 *  please see the vdmRCS encapsulation.
 */



/*\Docpage _vdmInit -- Initialize SM encapsulation.
 * 
 * DESCRIPTION
 *      This function is called only once by the VDM.  It's purpose is to
 *      initialize this particular storage manager encapsulation.  Note that a
 *      second initialization method in this file, _vdmCreate, is used to bind
 *      the storage manager encapsulation to a particular location.
 *      
 *      In addition, this function must return a structure which will be used
 *      to prompt and to parse for the location's string.  For more
 *      information about the ViewScript structures, see struct(man).
 *
 * NOTE
 *      The difference between this method and the _vdmCreate method is that
 *      the _vdmInit method is called only once per storage encapsulation.
 *      The _vdmCreate method is called for each location managed by this
 *      storage manager.
 *
 * SEE ALSO
 *      struct
 */
function _vdmInit(obj, endProc, endArg, ignore)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("Mega _vdmInit.\n"));

    x := struct vdmMegaVT { vlStringVT work_area };
    endProc (obj, endArg, x);
}



/*\Docpage _vdmCreate -- Instantiate the SM encapsulation.
 *
 * DESCRIPTION
 *      The _vdmCreate method will be called when the VDM requests the first
 *      storage operation performed on a data object in the given location.
 *      Any SM request can trigger instantiation.  The instantiation of the
 *      storage manager encapsulation is postponed until the VDM actually
 *      wants to use it in order not to slow down the initialization of
 *      applications built on top of the VDM.
 *
 *      This method returns the storage manager's context.  The context
 *      represents the particular instantiation of this storage manager
 *      encapsulation for a particular location.  The same storage manager can
 *      be used to manage another location, but VDM will create two separate
 *      instances of the storage manager encapsulation for the two lcoations.
 *
 * NOTE
 *      The difference between this method and the _vdmInit method is that the
 *      _vdmInit method is called only once per storage encapsulation.  The
 *      _vdmCreate method is called for each location managed by the given
 *      storage manager.
 *
 * CAUTION
 *      The object passed in to this function is a generic data object managed
 *      by the VDM library.  It has no relation whatsoever to any design data
 *      objects and it is used only by VDM.
 */
function _vdmCreate(obj, endProc, endArg, ignore, struct_type, location_string)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("Mega _vdmCreate: %s\n", location_string));
  
    struct vdmMegaObjVT
    {
        vscmVT tbl; /* the .tbl file */
        vscmVT lib  /* the .lib file */
    };
  
    struct vdmMegaContextVT
    {
        vlStringVT    work_area;      /* path to where megafile lives */
        vscmVT        location;       /* vdm location of megafile */
        vlTrueFalseVT initialized;    /* TRUE = this struct filled in */
        vdmMegaObjVT  sch, sym, wir   /* vdmObjVTs for know megafile types */
    };

    /* get working area */
    work_area := vscript.vsAbsPath (string_trim(location_string) + "/");

    context := struct vdmMegaContextVT {
        work_area, 
        nil,
        FALSE,
        struct vdmMegaObjVT {nil, nil}, 
        struct vdmMegaObjVT {nil, nil}, 
        struct vdmMegaObjVT {nil, nil}
    };


    endProc (obj, endArg, context);
}



/* Utility function initializing encapsulation context.
 */
function init_cntx (context, vdmCntx, loc)
{
    if (!context.initialized) 
    {
        function get_class (name)
        {
            class := vdm.CntxFindClass(vdmCntx, name);
            if (null(class))
            vlEx.Throw2 (vlEx.FAILURE,
                         "vdmMega",
                         vlErrSev.Err,
                         sprintf ("Use of vdmMega requires class \"%s\".",
                                  name));
            class;
        }

        function find_object ( name, location, class )
        {
            /* create an obj ref and see if we can find it */
            object := vdm.ObjCreate(name,
                                    vdm.VersLatest,
                                    class,
                                    location, 
                                    vdmCntx,
                                    nil);
            if (vdm.ObjFind(object, nil, location))
                return object;
            else
                return nil;
        }

        declare class;

        context.location = vpm.LocationGetForeignData (loc);

        class = get_class ("AnyFileOrDir");
        context.sym.tbl = find_object ("sym.tbl", context.location, class);
        context.sch.tbl = find_object ("sch.tbl", context.location, class);
        context.wir.tbl = find_object ("wir.tbl", context.location, class);

        context.sym.lib = find_object ("sym.lib", context.location, class);
        context.sch.lib = find_object ("sch.lib", context.location, class);
        context.wir.lib = find_object ("wir.lib", context.location, class);

        context.initialized = TRUE;
    }
}



/* get a vdmObjVT for the megafile that owns the given obj 
*/
function get_megafile ( context, obj )
{
    class_name := vdm.ClassGetName (vdm.ObjGetClass (obj));
    declare mega;

    if (equal(class_name, "Schematic"))
        mega = list(context.sch.tbl, context.sch.lib);
    else if (equal(class_name, "Symbol"))
        mega = list(context.sym.tbl, context.sym.lib);
    else if (equal(class_name, "Wir"))
        mega = list(context.wir.tbl, context.wir.lib);

    return mega;
}



/*\Docpage _vdmLoop -- Build a list of objects for a VDM loop.
 * 
 * DESCRIPTION
 *      Returns a list of data objects of the given class in this location.
 *      If there are no objects of that type, it returns nil.  This function
 *      is called by the vdm.ClassLoopObj function.
 *
 * For vdmMega, this method only recognizes the classes "Schematic", 
 * "Symbol", and "Wir".
 */
function _vdmLoop(obj, endProc, endArg, megaCntx, class, loc, cntx, usr_filter)
{
  if (vdmDebug)
    debug_out (obj, sprintf ("Mega _vdmLoop: %s\n", vpm.LocationGetAlias(loc)));

  init_cntx (megaCntx, cntx, loc);  

  obj_list   := nil;
  file_name  := nil;
  class_name := vdm.ClassGetName(class);

  if (equal(class_name, "Schematic"))
    file_name = "sch";
  else if (equal(class_name, "Symbol"))
    file_name = "sym";
  else if (equal(class_name, "Wir"))
    file_name = "wir";

  if (file_name != nil)
  {
    declare base_path;
    file_template   := vdm.ClassBuildFileTemplate (class);
    regexp_template := vdm.ClassBuildRegexpTemplate (class, usr_filter);
    
    path            := vscript.vsMerge ("", megaCntx.work_area, file_name, "");
    
    if (vdm.MegaExists(path))
    { 
      mega := vdm.MegaOpen(path, "r");

      path_length     := string_length (megaCntx.work_area);
      regexp          := vscript.re_compile (regexp_template);
    
      matched_attrs   := nil;
      match_str       := nil;
      match_path      := nil;

      /*
       * The algorithm for the file encapsulation is as follows:
       *
       * 1. Substitute all the attribute references in the template with '*'.
       * 2. Append the "wildcarded" template to the location's path.
       * 3. Loop through all the files that match the template.  Then apply the
       *    regular expression template to narrow down the match.
       * 4. If the regular expression template matches, extract the attribute
       *    values and build an object handle.  Then, add it to the object 
       *    list. 
       */
      foreach (key; vdm.MegaLoopRecordKeys(mega))
      {
        match_path = vscript.vsMerge("", path, key, "");
        match_str = substring (match_path, path_length, 0);
        
        if (vscript.re_exec (regexp, match_str))
        {
          matched_attrs = vdm.ClassGetMatchedAttributes (class, regexp,
                                                         match_str);
          obj_list = cons (vdm.ObjCreate (match_str, vdm.VersLatest, class,
                                          loc, cntx, matched_attrs),
                           obj_list);
        }
      }
      vdm.MegaClose(mega);
    }
  }
  endProc (obj, endArg, obj_list);
}



/*\Docpage _vdmStatus -- What's the storage status of this object.
 * 
 * DESCRIPTION
 *      The sole purpose of this method is to determine the status of the
 *      given object with respect to the storage manager.  This method must
 *      return two values, representing the possible storage manager state
 *      (vdm.SMStatusVT) and the possible working area state (vdm.WorkAreaVT).
 *
 *      If the object is under the storage manager's control, the object's
 *      status can be one of the following:
 *
 *          vdm.SMStatus.Unlocked       - object is not locked
 *          vdm.SMStatus.LockedByYou    - object is locked by you
 *          vdm.SMStatus.LockedByOther  - object is locked by another user
 *          vdm.SMStatus.Uncontrolled   - object is not yet in SM
 *
 *      If the storage manager allows the data object to persist outside of
 *      the storage manager's storage area, then such an object is said to
 *      reside in the user's working area.  Its status can then be one of the
 *      following:
 *
 *          vdm.WorkStatus.New       - object is brand new.
 *          vdm.WorkStatus.Modified  - object is modified
 *          vdm.WorkStatus.Outdated  - object is not the same as object in SM
 *          vdm.WorkStatus.UpToDate  - object is the same as that in SM
 *          vdm.WorkStatus.Empty     - object is not in the work area
 *
 * SEE ALSO
 *      vdm.SMStatusVT, vdm.WorkStatusVT
 */
function _vdmStatus(obj, endProc, endArg, megaCntx)
{
    declare tbl, lib;

    <tbl, lib> = get_megafile (megaCntx, obj);
    endProc (obj, endArg, vdm.ObjGetStatus (tbl));
}



/*\Docpage debug_out -- Handy output function.
 * 
 * DESCRIPTION
 *      Takes an object and a string and if the vdmDebug variable is set, it
 *      calls the VDM's context output handlers with the string.
 *
 *      When running ViewData, the output will show up in the session log
 *      window.
 */
function debug_out(obj, str)
    vdm.CntxCallOutputHandlers (vdm.ObjGetCntx (obj), str);

