/* C S A M P L E . C

Module:  CSAMPLE

Functional Description: 
    Routines to demonstrate C interface for Viewsim - this is a very
    simplistic example using a 2 input NAND gate.  The attributes on 
    the symbol denoting that the NAND is a C model are:

         SAMPLE=NAND
         SIMMODEL=C
         LEVEL=SAMPLE
         PINORDER=A B QN

    The LEVEL attribute allows vsm netlisting down to the SAMPLE
    level.  SIMMODEL=C indicates that this is a C model.  The pin
    labels can be anything, but PINORDER must indicate that the
    input pins come before the output pin, because that is the 
    arbitrary order in which the code below defines them.

    Note that the PINORDER used here is different from the PINORDER
    of the Viewsim NAND2 builtin primitive.  You can not test this
    C model using the Viewsim NAND2 builtin symbol unless you change 
    the pin order either in this C model or on the NAND2 symbol.
    
Author:  
    rl, Nov. 1, 1990
    
 *
 * $Log: csample.c,v $
 * Revision 1.7.2.2  1995/03/03  21:50:04  dahan
 * Add fcs_vsim531 label
 *
 * Revision 1.7  1994/08/25  21:12:30  dahan
 * merged version of 521(cosim+) with 53 baseline
 *
 * Revision 1.6.1.1.1.1  94/05/17  13:02:28  dahan
 * Initial revision for branch: rel52
 * 
 * Revision 1.6.1.1  94/05/17  12:26:40  dahan
 * Initial revision for branch: rel52
 * 
 * Revision 1.6  94/01/19  09:47:57  dahan
 * Add the Log macro for messages
 *
 *
 *

Modified by: 
Who  Ver  When       What
thc  1.0  mar 04 92  changed calls to vscu_define_pin.
thc  1.0  feb 05 92  moved vscu_set_keyword call from vsc_initialization
                     routine to sample_proc routine.
*/

#include <stdio.h>
/* #define PROTO 1 */
#include "viewsim.h"

/* Structure definition to support NAND instances */

#define NAND_INPUTS 2
#define VERSION "NAND C Model Version 1.0"
#define SAMPLE_CMD "CSTATS"

typedef struct temp_nand {
    struct temp_nand *next_inst;
    unsigned long     delay;
    vsc_pin_4values   output_pin;
} nand_struct;


/* GLOBAL */

    vsc_entries_struct *Nand_entries = NULL;
    nand_struct *Head_nand_inst      = NULL;
   
/*************************************************************************
*   VSC_INITIALIZATION
*
*  This is the main routine which must be named vsc_initialization
**/

void vsc_initialization()
{
   extern vsc_entries_struct *sample_proc();

   vscu_printf("Initializing C Models ...\n");

   /* Register the symbol attribute SAMPLE=NAND with Viewsim so
      it can identify the C model */
	
   if (!vscu_set_type("SAMPLE", sample_proc))
      vscu_printf("vscu_set_type registration of \"SAMPLE\" failed.\n");
}

/*************************************************************************
*
*   SAMPLE_PROC
*
*  This routine defines the entry points of all models of type SAMPLE
**/

vsc_entries_struct *sample_proc(module_name)
   char *module_name;
{
   extern void *nand_init();
   extern void nand_eval();
   extern void nand_restart();
   extern void nand_info();
   extern void nand_close();
   extern void nand_save();
   extern void nand_restore();
   extern void cmd_handler();
   extern int  strcmp();

   /* Since we know that the only C module is a NAND gate, make
      sure that the module name is NAND */

   if (strcmp("NAND", module_name))
      {
      vscu_printf("SAMPLE model of value \"%s\" not supported.\n",
                  module_name);
      return(NULL);
      }
	
   /* Remember this routine will be called multiple times (for each
      instance) but only one entries structure should be allocated */
	
   if (Nand_entries == NULL)
      {
	
   /* Allocate the entries structure and assign the entry points  */
	
      Nand_entries = (vsc_entries_struct *)
                     vscu_allocate(sizeof(vsc_entries_struct));
				
      Nand_entries -> inst_init       = nand_init;
      Nand_entries -> inst_eval       = nand_eval;
      Nand_entries -> inst_restart    = nand_restart;
      Nand_entries -> inst_info       = nand_info;
      Nand_entries -> inst_close      = nand_close;
      Nand_entries -> inst_save       = nand_save;
      Nand_entries -> inst_restore    = nand_restore;

   /* Register the command SAMPLE_CMD with Viewsim */

      if (!vscu_set_keyword(SAMPLE_CMD, cmd_handler))
         vscu_printf("vscu_set_keyword registration of \"%s\" failed.\n",
                     SAMPLE_CMD);
      }
	
   return(Nand_entries);
}

/**************************************************************************
*   NAND_INIT
*
* This routine handles the intialization for each instance of a nand gate
**/

void *nand_init()
{
   nand_struct *nand_inst;
   char        **aargv;
   int         i, aargc; 
   long delay;
   extern char *strchr();
   extern int strncmp();

   /* register the pins with Viewsim to ensure that the model and
      Viewsim are synchronized */

   if (!vscu_define_pin(0, INPUT))
      {
      vscu_printf("Model NAND pin A registration failed.\n");
      return(NULL);
      }
   if (!vscu_define_pin(1, INPUT))
      {
      vscu_printf("Model NAND pin B registration failed.\n");
      return(NULL);
      }
   if (!vscu_define_pin(2, OUTPUT))
      {
      vscu_printf("Model NAND pin QN registration failed.\n");
      return(NULL);
      }
	
   /* Allocate the nand structure */
	
   nand_inst = (nand_struct *) vscu_allocate(sizeof(nand_struct));

   /* Link this instance into the list of nand instances */
   /* for use by the cmd_handler routine. */
	
   nand_inst -> next_inst = Head_nand_inst;
   Head_nand_inst         = nand_inst;
	
   /* Assign initial values */

   nand_inst -> output_pin    = X4;
   nand_inst -> delay         = 0;

   /* Get CDELAY attribute specified via schematic */
   aargc = vscu_get_inst_attr(&aargv);

   for (i = 0; i < aargc; ++i)
      if (!strncmp(aargv[i], "CDELAY=", 7))
         {
         if (vscu_string_time(aargv[i] + 7, &delay))
            nand_inst->delay = (unsigned long) delay;
         else vscu_printf("Invalid CDELAY attribute \"%s\" specified on instance \"%s\"\n",
                          aargv[i], vscu_get_inst_name(nand_inst));
         }

   return(nand_inst);
}

/*************************************************************************
*   NAND_EVAL     
*
*   This routine is called by Viewsim if any input pin changes or a wakeup
*   is scheduled on a NAND gate.
**/

void nand_eval(inst_id, pin_array)
void *inst_id;
vsc_pin_values *pin_array;
{
   extern vsc_pin_4values vsc_31to3_state[];
   extern vsc_pin_4values vsc_4state_not[];

   int             i;
   vsc_pin_4values qn = HIGH4;
   vsc_pin_values  qnout;

   /* NAND inputs */
   for (i = 0; i < NAND_INPUTS && qn != LOW4; ++i)
      qn &= vsc_31to3_state[pin_array[i]];
   qn = vsc_4state_not[qn];

   /* if necessary, schedule output */
   if (qn != ((nand_struct *) inst_id)->output_pin)
      {
      if (qn == HIGH4)
         qnout = D1;
      else if (qn == LOW4)
         qnout = D0;
      else
         qnout = D1D0;

      vscu_sched_output(inst_id, 2, qnout, ((nand_struct *) inst_id)->delay,
                        INERTIAL_DELAY);
      ((nand_struct *) inst_id)->output_pin = qn;
      }
}

/****************************************************************************
*
*   NAND_RESTART
*
* This routine is called by Viewsim if the "restart" command was issued 
**/

void nand_restart(inst_id)
void *inst_id;
{
   ((nand_struct *) inst_id) -> output_pin = X4;
}

/****************************************************************************
*  
*    NAND_SAVE
*
*  Saving / restoring is not yet supported.
***/

void nand_save(inst_id)
void *inst_id;
{
   return;	
}

/***************************************************************************
*
*   NAND_RESTORE
*
*  Saving / restoring is not yet supported.
**/

void nand_restore(inst_id)
void *inst_id;
{
   return;	
}

/****************************************************************************
*
*   NAND_INFO
*
*  This routine is called by Viewsim to retrieve information about an
*  instance.  The information is communicated to the user through the
*  vscu_printf utility.
**/

void nand_info(inst_id)
void *inst_id;
{
   vscu_printf(VERSION);
}

/***************************************************************************
*
*   NAND_CLOSE
*
*  This routine is called by Viewsim when simulation has terminated.  This
*  routine will free any data structures allocated to a particular instance.
**/

void nand_close(inst_id)
void *inst_id;
{
   /* Normally you would want to free the particular instance passed
      into this routine, however for simplicity in this application 
      the entire list can be freed in vsc_terminate. */
	
   return;
}

/****************************************************************************
*
*  VSC_TERMINATE
*
*  This routine is the last routine called to free/cleanup any structures 
*  used globally by the model(s).  Note - this routine MUST be named
*  vsc_terminate.
*
**/

void vsc_terminate()
{
   nand_struct *nand_inst = Head_nand_inst;
   nand_struct *next_nand_inst;

   /* Free the global areas allocated during the initialization
      process */

   while (nand_inst != NULL)
      {
      next_nand_inst = nand_inst -> next_inst;
      vscu_deallocate((void *) nand_inst);
      nand_inst      = next_nand_inst; 
      }
	
   vscu_deallocate((void *) Nand_entries);

   Nand_entries   = NULL;
   Head_nand_inst = NULL;

   /* remove command handler */

   if (!vscu_reset_keyword(SAMPLE_CMD))
      vscu_printf("vscu_reset_keyword removal of \"%s\" failed.\n",
                  SAMPLE_CMD);

   vscu_printf("Terminating C Models    Elapsed time: %ld\n",
               vscu_get_elapsed_time());
}

/****************************************************************************
*
*  CMD_HANDLER   
*
*  This routine is the command handler for the sample command registered  
*  in vsc_initialization.  It is included to show how a new Viewsim command
*  may be created using the interface
*
**/

void cmd_handler(num_args, args)
int num_args;
char *args[];
{
   nand_struct *n;
   int i, num_ios;
   char *signame;
   unsigned long time;
   vsc_pin_values sigvalue;
   extern char *vsc_31state_display[];

   /* print out the command line args */

   vscu_printf("command line args:");
   for (i = 0; i < num_args; ++i)
      vscu_printf(" %s", args[i]);
   vscu_printf("\n");

   /* print out some information about each model instance */

   for (n = Head_nand_inst; n != NULL; n = n->next_inst)
      {
      vscu_printf("For instance %s:", vscu_get_inst_name((void *) n));
      num_ios = vscu_num_pins((void *) n);
      for (i = 0; i < num_ios; ++i)
         if ((signame = vscu_pin_signal((void *) n, i)) == NULL)
            vscu_printf(" NULL");
         else 
            {
            vscu_printf(" %s", signame);
            if (vscu_signal_value(signame, &sigvalue))
               vscu_printf(" (%s)", vsc_31state_display[sigvalue]);
            }    
      vscu_printf("\n\n");
      }

   if (vscu_next_event(&time))
      vscu_printf("time of next event: %s\n", vscu_time_string(time));
   else vscu_printf("there are no pending events.\n");
}

