/*
 * $Revision: 1.10 $
 *
 * This module parses vantage.tms files.  The code is redundant with
 * the vantage internal libraries, and also with vhdllib, a Viewlogic C
 * library that does the same thing.
 *
 * The reason we don't use the Vantage code is that we don't want to
 * link in all their libraries to get that one tiny bit of
 * functionality.  I do believe you'd either have to link in their
 * whole UIMS to get this one function, or you'd have to hack it, and
 * it's fairly cumbersome.
 *
 * The reason we don't use vhdllib is historical -- this was written
 * first, then it became obvious the functionality was required from C.
 * In addition, this functionality is needed from executables that currently
 * don't link in vhdllib, although that obviously could be changed without
 * much overhead.  Either that or we could use Ken Cox's way cool dynamic
 * loading.
 *
 * The format of the file, as I've figured from reading Vantage source code,
 * and from looking at the binary file in Emacs, is the following:
 *
 *      \252\252\252\252                4-byte Signature
 *      version                         4-byte version number, >=1
 *      source time                     4-byte unix time code for source file
 *      bin time                        4-byte time code for .o file
 *      vif time                        4-byte time code for .vif file
 *      num_flags                       4-byte number of flags
 *      data_length                     4-byte number...obsolete, should be 0
 *      [                               repeating sequence...
 *       flag_number                      4-byte flag number, 1..MAX_LBR_FLAG
 *       value                            usually 4-bytes, except for
 *                                        LBR_FLAG_SOURCE_FILE, which contains
 *                                        ascii bytes terminated by \n, so it
 *                                        can be read with fgetline.
 *      ]                               end of repeating sequence (*num_flags).
 */
SOURCE_TIME             := 0;

LBR_FLAG_SOURCE_FILE    := 1;
LBR_FLAG_RCHK           := 2;
LBR_FLAG_DBGLVL         := 3;
LBR_FLAG_C_OPTLVL       := 4;
LBR_FLAG_VSS_VER_MAJOR  := 5;
LBR_FLAG_VSS_VER_MINOR  := 6;
/*
 * LBR_FLAG_VHDL_OPTLVL    := 7;
 */
LBR_FLAG_LINE_INFO      := 7;
LBR_FLAG_LANG           := 8;

LBR_FLAG_START_LINE     := 9;
LBR_FLAG_END_LINE       := 10;

MAX_LBR_FLAG            := 10;

VIF_TIME                := 11;
BIN_TIME                := 12;

NUM_TMS_FIELDS          := 13;

function read_lib_tms(fn) {
    tms := nil;
    if (fexists (fn)) {
        f := fopen (fn, "rb");

        /* The file starts with 4 bytes of magic */
        if ((fgetchar (f) != '\252') ||
            (fgetchar (f) != '\252') ||
            (fgetchar (f) != '\252') ||
            (fgetchar (f) != '\252'))
            error ("Bad magic number in library timestamp file", fn);

        version := fget32 (f, vlIntVT);
        if (version >= 1) {     /* upward-compatible, I hope! */
            tms = make_vector (NUM_TMS_FIELDS, nil);
            tms[SOURCE_TIME] = fget32 (f, vlIntVT);
            tms[BIN_TIME]    = fget32 (f, vlIntVT);
            tms[VIF_TIME]    = fget32 (f, vlIntVT);

            num_flags := fget32 (f, vlIntVT);
            data_length := fget32 (f, vlIntVT);
            declare flag_type;

            if (data_length == 0) {  /* older versions used this field */
                while ((flag_type = fget32 (f, vlIntVT)) > 0) {
                    if (flag_type == LBR_FLAG_SOURCE_FILE)
                        tms[flag_type] = string_trim (fgetline (f));
                    else if (flag_type == LBR_FLAG_LINE_INFO) {
                        tms[flag_type] = fget32 (f, vlIntVT);
                        tms[LBR_FLAG_START_LINE] =
                            quotient (tms[flag_type], 0x10000);
                        tms[LBR_FLAG_END_LINE] = tms[flag_type] & 0xFFFF;

                        /*
                         * Sometimes the Vantage compiler reports
                         * the lines as 0!  If so, just null them
                         * out and pretend Vantage never said anything
                         * about source lines.
                         */
                        if ((tms[LBR_FLAG_START_LINE] == 0) ||
                            (tms[LBR_FLAG_END_LINE] == 0))
                        {
                            tms[LBR_FLAG_START_LINE] = nil;
                            tms[LBR_FLAG_END_LINE] = nil;
                        } /* if */
                    } /* else if */
                    else if ((flag_type > 0) && (flag_type <= MAX_LBR_FLAG))
                        tms[flag_type] = fget32 (f, vlIntVT);
                    else
                        /*
                         * Just silently fail.  The user *really* doesn't care!
                         * error ("invalid flag in tms file " + fn, flag_type)
                         */
                        ;
                } /* while */
            } /* if */
        } /* if */

        fclose (f);
    } /* if */

    tms;
} /* function read_lib_tms */

/*
 * This test function can be enabled and used to test read_lib_tms without
 * any other .vs files -- it should work from standalone Viewscript.  
 */
if (FALSE) {
    function tst(fn) {
        tms := read_lib_tms (fn);
        if (tms != nil) {
            function s(x)
                convert (vlStringVT, x);

            function int_to_date(time) {
                if (time == 0)
                    "?";
                else {
                    declare mm, dd, yyyy, h, m, s;
                    if (vlCvtTime (time, &mm, &dd, &yyyy, &h, &m, &s) == 0)
                        error ("vlCvtTime failed", time);
                    sprintf ("%d/%d/%d %d:%02d:%02d",
                             mm, dd, yyyy, h, m, s);
                } /* else */
            } /* function time_str */


            (" src=" + int_to_date (tms[SOURCE_TIME]) +
             " bin=" + int_to_date (tms[BIN_TIME]) +
             " vif=" + int_to_date (tms[VIF_TIME]) +
             " file=" + s(tms[LBR_FLAG_SOURCE_FILE]) +
             " start=" + s(tms[LBR_FLAG_START_LINE]) +
             " end=" + s(tms[LBR_FLAG_END_LINE]) +
             " range_check=" + s(tms[LBR_FLAG_RCHK]) +
             " c_opt_level=" + s(tms[LBR_FLAG_C_OPTLVL]) +
             " version=" + sprintf ("%d.%d",
                                    tms[LBR_FLAG_VSS_VER_MAJOR],
                                    tms[LBR_FLAG_VSS_VER_MINOR]) +
             " vhdl_opt_level=" + s(tms[LBR_FLAG_VHDL_OPTLVL]));
        } /* if */
        else
            nil;
    } /* function tst */
} /* if */
