/* vapi.vs  Viewdraw API extensions
 *
 * Author:   Onofrio Sozio
 *
 * Copyright (c) 1994 by Viewlogic Systems,Inc.
 */



package vapi 
(
 DeleteObj,                   /* Delete Object in Block */

 LibBlockOpen,                /* Open Block by alias */
 
 LibBlockOpenSheet,           /* Open Block by alias and sheet */
 
 LibBlockExists,              /* Block Exists by alias */
 
 LibBlockSheetExists,         /* Sheet Exists by alias and sheet num */
	      
 LibDataObjExists,            /* Data Object Exists by alias */

 LibDataObjGetPath            /* Get Path for Data Object Exists by alias */

 )
{

  /* DeleteObj (block,obj)
     
     DESCRIPTION:  Delete Object
     
     EXAMPLE: vapi.DeleteObj(block,obj);
     
     block vqBlockVT
     obj  any viewdraw object (label, attribute, text....); 

     RETURNS: void
     */
		 
  function DeleteObj (block,obj)
  {
    vdraw.BlockDeselectAll(block);
    vdraw.BlockSelectObj(block,obj);
    vdraw.BlockDelSelectedAll(block,TRUE);
  }

  /* LibBlockOpen (name,type)
     
     DESCRIPTION:  Open Block by aliased name
     
     EXAMPLE: vapi.LibBlockOpen("mylib:mysym.1","Symbol");
     
     name vlStringVT
     type vqDataTypeVT 

     RETURNS: vqBlockVT
     */
		 
  function LibBlockOpen (name,type)
  {
    lib := vdraw.NameSpecGetLib(name);
    symname := vdraw.NameSpecGetName(name);
    sheet := vdraw.NameSpecGetSheet(name);
    vdraw.BlockOpen(lib,symname,sheet,type);
  }

  /* LibBlockOpenSheet (name,type,sheet) 

     DESCRIPTION:  Open Block by aliased name and explicit sheet number 

     EXAMPLE: vapi.LibBlockOpen("mylib:mysch",3,"Schematic");

     name vlStringVT
     sheet vqSheetNumVT
     type vqDataTypeVT 

     RETURNS: vqBlockVT
     */	 

  function LibBlockOpenSheet(name,sheet,type)
  {
    lib := vdraw.NameSpecGetLib(name);
    symname := vdraw.NameSpecGetName(name);
    vdraw.BlockOpen(lib,symname,sheet,type);
  }

  /* LibBlockExists (name,type) 

     DESCRIPTION: Checks to see if a block exists

     EXAMPLE:  vapi.LibBlockExists("mylib.mysch.3","Schematic");

     name vlStringVT
     type vqDataTypeVT 
     RETURNS: vlTrueFalseVT
     */	 

  function LibBlockExists(name,type)
  {
    lib := vdraw.NameSpecGetLib(name);
    symname := vdraw.NameSpecGetName(name);
    sheet := vdraw.NameSpecGetSheet(name);
    vdraw.BlockExists(lib,symname,sheet,type);
  }

  /* LibDataObjExists (name,class) 

     DESCRIPTION: Checks to see if a Data Object exists

     EXAMPLE:  vapi.LibDataObjExists("mylib.mysch.3","GenericLibrary");

     name vlStringVT
     class vqObjClassVT 
     RETURNS: vlTrueFalseVT
     */	 

  function LibDataObjExists(name,class)
  {

    lib := vdraw.NameSpecGetLib(name);
    symname := vdraw.NameSpecGetName(name);

    if(pos := lindex(name, '.'))
      sheet := substring(name,pos+1, 0);
    else            
      sheet := vdraw.NameSpecGetSheet(name);

    vdraw.DataObjExists(lib,symname,sheet,class);
  }

  /* LibDataObjGetPath (name,class) 

     DESCRIPTION: Get Physical path for a VL Data Object

     EXAMPLE:  vapi.LibDataObjGetPath("mylib.mysch.3","GenericLibrary");

     name vlStringVT
     class vqObjClassVT 
     RETURNS: vlStringVT
     */	 

  function LibDataObjGetPath(name,class)
  {
    lib := vdraw.NameSpecGetLib(name);
    symname := vdraw.NameSpecGetName(name);
    if(pos := lindex(name, '.'))
      sheet := substring(name,pos+1, 0);
    else            
      sheet := vdraw.NameSpecGetSheet(name);
    vdraw.DataObjGetPath(lib,symname,sheet,class);
  }

  /* LibBlockSheetExists (name,type,sheet) 

     DESCRIPTION: Checks to see if a specific sheet of a block exists

     EXAMPLE:  vapi.LibBlockSheetExists("mylib.mysch",3,"Schematic");

     name vlStringVT
     sheet vqSheetNumVT
     type vqDataTypeVT 
     RETURNS: vlTrueFalseVT
     */	 

  function LibBlockSheetExists(name,sheet,type)
  {
    lib := vdraw.NameSpecGetLib(name);
    symname := vdraw.NameSpecGetName(name);
    vdraw.BlockExists(lib,symname,sheet,type);
  }


}
