/*
 * WildcardingVT -- enum describing automatic wildcarding strategy for a
 *                  given vtype.
 *
 *     Leading  -- add leading wildcard when searching
 *     Trailing -- add trailing wildcard when searching
 */
define_mask ("WildcardingVT",
             `("Leading" "Trailing"));

/*
 * vdbOrganizedVT -- enum describing how user wants data organized
 *
 *     ByLibrary   -- components separated by type into different libraries
 *     ByNamedView -- component category determined by some vtype,
 *                    e.g. "CATEGORY"
 */
define_enum ("vdbOrganizedVT",
             `("ByLibrary" "ByNamedView"));

/*
 * vdbModeVT -- enum describing what mode of operation VDB is in.
 *
 *     Idle         -- tool either just started, or just finished
 *                     back-annotation.  Window is blank and things
 *                     are disabled, other than the Search menu and the
 *                     load button.
 *     BackAnnotate -- current search values were loaded from ViewDraw.
 *                     Intention is back-annotation of these values to
 *                     the schematic.
 *     AddGeneric   -- there exists > 1 candidate components.
 *                     Intention is to add a component with the
 *                     applicable common views as attributes.
 *     AddUnique    -- there exists only 1 candidate component.
 *                     Intention is to add the component with all
 *                     the applicable views as attributes.
 */
define_enum ("vdbModeVT",
             list ("Idle", "AddGeneric", "AddUnique", "BackAnnotate"));

/*
 * vdbValueSourceVT -- enum describing the source of a value in a search field,
 * i.e. where the string came from.
 *
 * IMPORTANT: These are listed in increasing order of importance.  That
 * is, higher values are never replaced by lower values, except on
 * search reset or clear.
 *
 *     NotSet        -- nothing has been entered here yet.
 *     CommonView    -- value "fell out" of a search as a common view.
 *     FromVdraw     -- value got loaded directly from a ViewDraw attr.
 *     UserSpecified -- specified by the user (typed directly or picked
 *                      from list etc.)
 */
define_enum ("vdbValueSourceVT",
             list ("NotSet", "CommonView", "FromVdraw", "UserSpecified"));

function enum_lt(a,b)
    convert (vlIntVT, a) < convert (vlIntVT, b);

function enum_le(a,b)
    convert (vlIntVT, a) <= convert (vlIntVT, b);

function enum_gt(a,b)
    convert (vlIntVT, a) > convert (vlIntVT, b);

function enum_ge(a,b)
    convert (vlIntVT, a) >= convert (vlIntVT, b);

vscript.overload ("<",  enum_lt, vdbValueSourceVT, vdbValueSourceVT);
vscript.overload ("<=", enum_le, vdbValueSourceVT, vdbValueSourceVT);
vscript.overload (">",  enum_gt, vdbValueSourceVT, vdbValueSourceVT);
vscript.overload (">=", enum_ge, vdbValueSourceVT, vdbValueSourceVT);

/* like assoc, but uses string_eqi instead of equal
 *
 * sort of complicated because the type of the cars do not have to be
 * vlStringVT, so I have to check whether to use string_equi or equal.
 */
function assoc_string_eqi(s, l)
    if (null (s))
        assoc (s, l);
    else
    {
        bFoundIt := FALSE;
        while (!null (l) && !bFoundIt)
        {
            s2 := caar(l);
            bFoundIt = ((istype (s2, vlStringVT))
                        ? string_eqi (s, s2)
                        : equal (s, s2));
            if (!bFoundIt)
                l = cdr (l);
        }
        if (null (l))
            FALSE;
        else
            car (l);
    }

/* like assoc, but compares vtype (stops compare at '\\' character)
 *
function assoc_no_qualifier(s, l)
{
    foreach (e; vscript.list_iter (l))
    {
        declare s2;
        string_token (car (e), &s2, "\\");
        if (string_eq (s, s2))
            return (e);
    }
    FALSE;
}
*/

/* throw exception */
function vdbThrow(num, args...)
    error2 ("vdb", vlErrSev.Error, num, nil, apply (sprintf, args));

/* report error, but don't throw */
function vdbReport(args...)
{
    if (istype (args[0], vcWindowVT))
    {
        win := car (args);
        args = cdr (args);
    }
    else
        win := nil;
    vcErr.Report (win, nil,
                  vlErr.Create (nil, "vdb", 0, vlErrSev.Error,
                                apply (sprintf, args)));
}

function VtypeToAttr(sVtype)
    if (null (sAttr := vlSym.Fetch (vdbCfg.symVtypeToAttr, sVtype)))
        sVtype;
    else
        sAttr;

/*
 * NB: the list returned does not maintain any sorting it may have had
 * previously.
 */
function VtypeToAttrList(lVtypes)
{
    declare l;
    foreach (e; vscript.list_iter (lVtypes))
        l = cons (VtypeToAttr (e), l);
    l;
}

/*
 * symAttrToCatVtype stores an alist of (sCategory . sVtype) pairs.
 * Only map the attr name if sCategory is found in that alist (or if we find
 * the nil category in the list).
 */
function AttrAndCategoryToVtype(sCategory, sAttr)
    if (null (alCatVtype := vlSym.Fetch (vdbCfg.symAttrToCatVtype, sAttr)))
        sAttr;
    else
    {
        e := assoc_string_eqi (sCategory, alCatVtype);
        if (!e)
            e = assq (nil, alCatVtype);
        if (!e)
            sAttr;
        else
            cdr (e);
    }

/*
 * Returns `sAttr' if `sAttr' has zero or one mapping.
 * Returns FALSE otherwise.  In that case, you have to use
 * AttrAndCategoryToVtype.
 */
function AttrToVtype(sAttr)
    if (null (alCatVtype := vlSym.Fetch (vdbCfg.symAttrToCatVtype, sAttr)))
        sAttr;
    else
    {
        if (length (alCatVtype) == 1)
            cdar (alCatVtype);
        else
            FALSE;
    }

/*
 * Returns TRUE if `sAttr' has a one-to-many relationship with vtypes (that
 * is, more than one vtype is mapped to it).
 * Returns FALSE if it has a one-to-one relationship, or is not mapped.
 */
function AttrHasManyMappings(sAttr)
    if (null (alCatVtype := vlSym.Fetch (vdbCfg.symAttrToCatVtype, sAttr)))
        FALSE;
    else
        length (alCatVtype) > 1;

/*
 * Returns TRUE if `sAttr' has any mappings.
function AttrHasAnyMappings(sAttr)
    !null (vlSym.Fetch (vdbCfg.symAttrToCatVtype, sAttr));
 */

function AttrToVtypeList(sCategory, lAttrs)
{
    declare l;
    foreach (e; vscript.list_iter (lAttrs))
        l = cons (AttrAndCategoryToVtype (sCategory, e), l);
    l = reverse_destructive (l);
    l;
}
