/*
 *  $Revision: 1.6.5.1 $
 *  
 *  Analog clock demonstration.  This illustrates the use of timers, line
 *  thicknesses, trigonometry (to calculate the position of the clock hands
 *  and ticks), canvases, and using XOR mode to animate graphics.
 *
 *  Copyright (c) 1992 by Viewlogic Systems, Inc.
 */

background_color    := vgColor.Black;
ticks_color         := vgColor.Green;
hands_color         := vgColor.Yellow;


function make_clock() {
 /*
  * Create a static text area to show the current time.
  */
 r := vcRegion.Create ("Analog ViewScript Clock", vcCanvas.fac, nil);
 vp := nil;
 vcWin.Create (r, vcWin.NoBindings + vcWin.NoStatus);

 function angle(min)  ((15 - min) / 30) * 3.141592654;

 function ensure_vp() {
     if (vp == nil) {
         vp = vcRegion.Viewport (r, vgRect.Create (0, 0, 1000, 1000),
                                 vgU.MatchAspectRatio);
         vgU.StkSize (vp, 50, 50);
         vgU.Fillstyle (vp, vgFill.Solid);
     }
 }

 function refresh(event) {
     ensure_vp ();

     /*
      * Erase the whole region with the background color
      */
     r := vcEvent.Rect (event);
     vgU.PixelToUserRect (vp, r);
     vgU.Color (vp, background_color);
     vgU.Rasterop (vp, vgRasterop.Replace);
     vgU.Box (vp, r);

     /* Clock perimeter */
     vgU.Color (vp, ticks_color);
     vgU.Circle (vp, 500, 500, 400);
     vgU.Color (vp, background_color);
     vgU.Circle (vp, 500, 500, 390);
     vgU.Color (vp, ticks_color);

     declare cos_theta, sin_theta, theta;

     /* Hourly Tick marks, numbers */
     for (hour := 1; hour <= 12; ++hour) {
         theta = angle (hour*5);    /* 5 minutes per hour on dial */
         cos_theta = cos (theta);
         sin_theta = sin (theta);
         vgU.StkDraw (vp, hour, vgAlign.MiddleCenter,
                      500 + 320 * cos_theta,
                      500 + 320 * sin_theta);
         vgU.Line (vp,
                   500 + 400 * cos_theta,
                   500 + 400 * sin_theta,
                   500 + 370 * cos_theta,
                   500 + 370 * sin_theta);
     } /* for */

     /* Minute tick marks, numbers */
     for (min := 1; min <= 60; ++min) {
         theta = angle (min);
         cos_theta = cos (theta);
         sin_theta = sin (theta);
         vgU.Circle (vp,
                     500 + 380 * cos_theta,
                     500 + 380 * sin_theta,
                     5);
     } /* for */

     draw_3_hands ();
 } /* function refresh */

 /*
  * Define hour, minute, and second storage.  Legal ranges for all three hands
  * range from 0 to 59.  Yes, the resolution of the hour hand is same as
  * others.  Actually hour can spill slightly over 59 between 12 and 1
  * o'clock.  Think of hour as incrementing every 12 minutes...
  *   Examples:
  *              23:11:50  =>  (hour=55, minute=11, second=50)
  *              12:59:50  =>  (hour=64, minute=59, second=50)
  */
 declare hour, minute, second;

 /*
  * Define storage for old values of hour, minute, and second.  Old values
  * have two purposes, to control when to update the hands and how to erase
  * old hand position upon update.
  */
 declare old_hour, old_minute, old_second;

 function draw_3_hands() {
     function draw_hand(radius, old, new) {
         function hand(theta)
             vgU.Line (vp, 500, 500,
                       500 + radius * cos (theta),
                       500 + radius * sin (theta));
         if (old != new) {
             if (old != nil) hand(angle(old));
             if (new != nil) hand(angle(new));
         } /* if */
      } /* function draw_hand */

     vgU.Color (vp, hands_color);
     vgU.Rasterop (vp, vgRasterop.Xor);
     vgU.LineThickness (vp, 20);
     draw_hand (200, old_hour, hour);
     draw_hand (350, old_minute, minute);
     vgU.LineThickness (vp, 0);
     draw_hand (350, old_second, second);
     vgU.Rasterop (vp, vgRasterop.Replace);
 }

 vcCanvas.RefreshProc (r, refresh);

 /*
  * Define a function to update the time displayed in the static text region.
  */
 function update() {
     declare month, day, year;

     ensure_vp ();

     t := vcWv.Time ();
     t = string_token (t, &month, "/");
     t = string_token (t, &day, "/");
     t = string_token (t, &year, " ");
     t = string_token (t, &hour, ":");   hour = convert (integer, hour);
     t = string_token (t, &minute, ":"); minute = convert (integer, minute);
     t = string_token (t, &second, ":"); second = convert (integer, second);

     /* fudge hour to be fifth's of hours (every 12 minutes hour hand moves) */
     hour = (hour*5) + (minute/12);

     draw_3_hands ();

     old_hour = hour;           hour = nil;
     old_minute = minute;       minute = nil;
     old_second = second;       second = nil;

 } /* function update */

 /*
  * Create a timer to call the update function every second.
  */
 timer := vcWv.Timer (wv, 1000, update, TRUE);

 /*
  * If the user kills the window, be sure to cancel the timer, otherwise
  * errors will occur when the update function is called on the invalid
  * region.
  */
 vcRegion.DestroyNotify (r, func(r) vcWv.CancelTimer (timer););
} /* function make_clock */

make_clock ();
