/*
 *  $Revision: 1.6.6.1 $
 *
 *  Copyright (c) 1992 by Viewlogic Systems, Inc.
 *
 */

/*
 * Display a histogram representing the alphabetic character count
 * in a file.
 */

/*
 * This function takes a filename and an array of 27 elements.  It puts
 * the filename in array27[0], and it puts the frequence of letters in
 * the alphabet in the remaining 26 elements.
 */
function count_chars(filename, array27) {
    f := fopen (filename, "r");
    a := make_vector (256, 0);
    function record_char(c) {c=char2int (c); a[c] = a[c] + 1;}
    function record_line(s) {
        for (l := string2list (s); l != nil; l = cdr (l))
            record_char (car (l));
    }
    while ((s := fgetline (f)) != nil)
        record_line (s);

    array27[0] = filename;
    for (i := 1; i < 27; ++i)
        array27[i] = a[i + char2int ('A')] + a[i + char2int('a')];
} /* function count_chars */

/*
 * This routine is used to get a window ID out of thin air.  It
 * loops through all the open windows in the process, and saves the last
 * one created as "w", then returns it.
 */
function get_window() {
    w := nil;
    foreach (ww; vcWin.Loop ()) w = ww;
    w;
} /* function get_window */

/*
 * Here is the _Create method for the facility.  The _Create method
 * is used whenever someone calls vcRegion.Create for this facility.
 * It's first argument is the data to edit -- in this case that's
 * the initial filename to show the character counts.  The _Create
 * method is responsible for setting up the data for the facility,
 * which in this case is a 27-element array, and for returning that
 * data structure.
 */ 
function _Create(filename, region) {
    a := make_vector (27, 0);
    vcWin.WithBusy (get_window (), func() count_chars (filename, a););
    a;
} /* function _Create */

/*
 * Walk through the array and determine the largest number of characters.
 */
function largest(a) {
    l := 0;
    for (i := 1; i < 27; ++i)
        if (a[i] > l) l = a[i];
    l;
} /* function largest */

/*
 * Draw the histogram on the screen.
 */
function _Refresh(a, event) {
    /*
     * Figure out the desired bounding box of the data and get
     * a viewport to do some drawing.
     */
    area := vgRect.Create (0, 0, 260, largest (a));
    vp := vcRegion.Viewport (vcEvent.Region (event), area, 0);

    /*
     * Erase the drawing area.
     */
    vgU.Fillstyle (vp, vgFill.Solid);
    vgU.Rasterop (vp, vgRasterop.Replace);
    vgU.Color (vp, vgColor.Black);
    vgU.Box (vp, area);

    /*
     * Draw a header identifying the filename, which is stored in a[0].
     */
    vgU.Color (vp, vgColor.White);
    vgU.StkSize (vp, 7, 7);
    vgU.StkDraw (vp, a[0], vgAlign.UpperLeft,
                 vgRect.Xmin (area), vgRect.Ymax (area));

    /*
     * Draw the rectangles for the histogram.
     */
    vgU.Color (vp, vgColor.Red);
    for (i := 0; i < 26; ++i)
        vgU.Box (vp, vgRect.Create (10*i, 0, 10*i+9, a[i+1]));

    /*
     * Draw the letter names at the bottom of each rectangle
     */
    vgU.Color (vp,vgColor.Green);
    function i2str(c) sprintf("%c", int2char (c+65));
    vgU.StkSize (vp, 10, 10);
    for (i = 0; i < 26; ++i)
        vgU.StkDraw (vp, i2str(i), vgAlign.LowerLeft, 10*i, 0);

    /*
     * Draw the number of characters at the top of each rectangle.
     */
    vgU.StkSize (vp, 5, 5);
    vgU.Color (vp, vgColor.Blue);
    for (i = 0; i < 26; ++i)
        vgU.StkDraw (vp, a[i+1], vgAlign.UpperLeft, 10*i, a[i+1]);
} /* function _Refresh */

/*
 * Compute the desired size of the region, taking into account the
 * aspect ratio of the histogram.
 */
function _Size(a, size, region, wv) {
    w_to_h := largest (a) / 260;
    vgSize.Set (size,
                500, 500*w_to_h,
                200, 200*w_to_h,
                1000, 1000*w_to_h);
} /* function _Size */

/*
 * Function that is called when user runs the "new" command.
 */
function new_file(a, event, filename) {
    vcWin.WithBusy (vcEvent.Window (event),
                    func() count_chars (filename, a););
    r := vcEvent.Region (event);
    vcRegion.RequestSize (r);
    vcRegion.Refresh (r, vcRegion.Area (r));
} /* function new_file */

new_file_cmd :=
    vcWv.Cmd (fac, new_file, "new", "Count characters in file",
              vcWv.Param ("Filename", vlStringVT, vcWv.Sticky, "",
                          vcFileSelect.fac));
vcCmd.EnableApply (new_file_cmd, vlOnOff.On);

my_menu := vcWv.MakeMenu ("histogram",
                          vcWin.RedSquareEntry,
                          list ("New File", new_file_cmd));
vcFac.SetMenus (fac, my_menu);
